Package: REMP
Type: Package
Title: Repetitive Element Methylation Prediction
Version: 1.37.0
Authors@R: c(person("Yinan","Zheng", role = c("aut","cre"), email = "y-zheng@northwestern.edu"),person("Lei","Liu", role = c("aut"), email = "lei.liu@northwestern.edu "), person("Wei","Zhang", role = c("aut"), email = "wei.zhang1@northwestern.edu"), person("Warren","Kibbe", role = c("aut"), email = "warren.kibbe@nih.gov"), person("Lifang","Hou", role = c("aut","cph"), email = "l-hou@northwestern.edu"))
Description: Machine learning-based tools to predict DNA methylation of locus-specific repetitive elements (RE) by learning surrounding genetic and epigenetic information. These tools provide genomewide and single-base resolution of DNA methylation prediction on RE that are difficult to measure using array-based or sequencing-based platforms, which enables epigenome-wide association study (EWAS) and differentially methylated region (DMR) analysis on RE. 
License: GPL-3
Depends: R (>= 3.6), SummarizedExperiment(>= 1.1.6), minfi (>= 1.22.0)
Imports: readr, rtracklayer, graphics, stats, utils, methods, settings,
        BiocGenerics, S4Vectors, Biostrings, GenomicRanges, IRanges,
        Seqinfo, BiocParallel, doParallel, parallel, foreach, caret,
        kernlab, ranger, BSgenome, AnnotationHub, org.Hs.eg.db, impute,
        iterators
Suggests: IlluminaHumanMethylation450kanno.ilmn12.hg19,
        IlluminaHumanMethylationEPICanno.ilm10b2.hg19,
        BSgenome.Hsapiens.UCSC.hg19, BSgenome.Hsapiens.UCSC.hg38,
        knitr, rmarkdown, minfiDataEPIC
Collate: REMParcel-class.R REMProduct-class.R generics.R
        REMParcel-methods.R REMProduct-methods.R options.R utils.R
        REMPtools.R getGM12878.R grooMethy.R initREMP.R rempTemplate.R
        remp.R remprofile.R DemoData.R onAttach.R REMP-package.R
URL: https://github.com/YinanZheng/REMP
BugReports: https://github.com/YinanZheng/REMP/issues
Encoding: UTF-8
LazyData: true
biocViews: DNAMethylation, Microarray, MethylationArray, Sequencing,
        GenomeWideAssociation, Epigenetics, Preprocessing,
        MultiChannel, TwoChannel, DifferentialMethylation,
        QualityControl, DataImport
RoxygenNote: 7.1.2
git_url: https://git.bioconductor.org/packages/REMP
git_branch: devel
git_last_commit: 3f67cb8
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.24
Date/Publication: 2026-04-29
NeedsCompilation: no
Packaged: 2026-04-29 23:56:13 UTC; biocbuild
Author: Yinan Zheng [aut, cre],
  Lei Liu [aut],
  Wei Zhang [aut],
  Warren Kibbe [aut],
  Lifang Hou [aut, cph]
Maintainer: Yinan Zheng <y-zheng@northwestern.edu>

# Generated by roxygen2: do not edit by hand

S3method("[",template)
export(GRannot)
export(fetchRMSK)
export(fetchRefSeqGene)
export(findRECpG)
export(getBackend)
export(getGM12878)
export(grooMethy)
export(initREMP)
export(remp)
export(rempTemplate)
export(remp_options)
export(remp_reset)
export(remprofile)
exportClasses(REMParcel)
exportClasses(REMProduct)
exportMethods(decodeAnnot)
exportMethods(details)
exportMethods(getILMN)
exportMethods(getParcelInfo)
exportMethods(getRE)
exportMethods(getRECpG)
exportMethods(getRefGene)
exportMethods(rempAggregate)
exportMethods(rempAnnot)
exportMethods(rempB)
exportMethods(rempCombine)
exportMethods(rempImp)
exportMethods(rempM)
exportMethods(rempQC)
exportMethods(rempStats)
exportMethods(rempTrim)
exportMethods(remplot)
exportMethods(saveParcel)
import(BiocParallel)
import(S4Vectors)
import(caret)
import(methods)
importFrom(AnnotationHub,AnnotationHub)
importFrom(AnnotationHub,getAnnotationHubOption)
importFrom(AnnotationHub,setAnnotationHubOption)
importFrom(BSgenome,getSeq)
importFrom(BiocGenerics,mget)
importFrom(Biostrings,DNAString)
importFrom(Biostrings,vmatchPattern)
importFrom(GenomicRanges,"end<-")
importFrom(GenomicRanges,"start<-")
importFrom(GenomicRanges,GRanges)
importFrom(GenomicRanges,GRangesList)
importFrom(GenomicRanges,end)
importFrom(GenomicRanges,findOverlaps)
importFrom(GenomicRanges,granges)
importFrom(GenomicRanges,makeGRangesListFromFeatureFragments)
importFrom(GenomicRanges,promoters)
importFrom(GenomicRanges,ranges)
importFrom(GenomicRanges,seqnames)
importFrom(GenomicRanges,shift)
importFrom(GenomicRanges,start)
importFrom(GenomicRanges,strand)
importFrom(IRanges,CharacterList)
importFrom(IRanges,IRanges)
importFrom(IRanges,IRangesList)
importFrom(IRanges,NumericList)
importFrom(IRanges,subsetByOverlaps)
importFrom(Seqinfo,"seqlevels<-")
importFrom(SummarizedExperiment,SummarizedExperiment)
importFrom(SummarizedExperiment,assays)
importFrom(SummarizedExperiment,colData)
importFrom(SummarizedExperiment,rowData)
importFrom(SummarizedExperiment,rowRanges)
importFrom(doParallel,registerDoParallel)
importFrom(foreach,registerDoSEQ)
importFrom(graphics,lines)
importFrom(graphics,plot)
importFrom(impute,impute.knn)
importFrom(iterators,idiv)
importFrom(iterators,nextElem)
importFrom(kernlab,predict)
importFrom(minfi,GenomicRatioSet)
importFrom(minfi,IlluminaMethylationAnnotation)
importFrom(minfi,MethylSet)
importFrom(minfi,RatioSet)
importFrom(minfi,getAnnotation)
importFrom(minfi,getBeta)
importFrom(minfi,getM)
importFrom(org.Hs.eg.db,org.Hs.egREFSEQ2EG)
importFrom(org.Hs.eg.db,org.Hs.egSYMBOL)
importFrom(parallel,detectCores)
importFrom(ranger,importance)
importFrom(ranger,ranger)
importFrom(readr,col_character)
importFrom(readr,col_integer)
importFrom(readr,cols_only)
importFrom(readr,read_delim)
importFrom(rtracklayer,liftOver)
importFrom(settings,options_manager)
importFrom(settings,reset)
importFrom(settings,stop_if_reserved)
importFrom(stats,density)
importFrom(stats,na.pass)
importFrom(stats,setNames)
importFrom(utils,data)
importFrom(utils,download.file)
importFrom(utils,packageVersion)
importFrom(utils,read.table)

## REMP: Repetitive Element Methylation Prediction

*REMP* provides machine learning-based tools to predict DNA methylation of locus-specific repetitive elements (RE) by learning surrounding genetic and epigenetic information. These tools provide genomewide and single-base resolution of DNA methylation prediction on RE that are difficult to measure directly using array-based or sequencing-based platforms, which enables epigenome-wide association study (EWAS) and differentially methylated region (DMR) analysis on RE. 

Supporting methylation platforms:
+ Illumina 450k BeadChip Array
+ Illumina EPIC BeadChip Array
+ Illumina methylation sequencing platforms (e.g., TruSeq Methyl Capture EPIC)

## Installation 

REMP is available in Bioconductor repository
```r
## try http:// if https:// URLs are not supported
if (!requireNamespace("BiocManager", quietly=TRUE))
    install.packages("BiocManager")

## To get the latest version of REMP, please make sure you are using the latest Bioconductor

## Install REMP
BiocManager::install("REMP")
```

## Citation:

Zheng Y, Joyce BT, Liu L, Zhang Z, Kibbe WA, Zhang W, Hou L. Prediction of genome-wide DNA methylation in repetitive elements. Nucleic Acids Research. 2017;45(15):8697-711. [PubMed PMID: 28911103](https://www.ncbi.nlm.nih.gov/pubmed/28911103); [PMCID: PMC5587781](https://www.ncbi.nlm.nih.gov/pmc/articles/PMC5587781/). [http://dx.doi.org/10.1093/nar/gkx587](http://dx.doi.org/10.1093/nar/gkx587).

## Contact package maintainer:
Yinan Zheng 

Email: y-zheng@northwestern.edu

## Generate REMP demo data:

library(REMP)
library(devtools)

if(!isNamespaceLoaded("IlluminaHumanMethylation450kanno.ilmn12.hg19")) attachNamespace("IlluminaHumanMethylation450kanno.ilmn12.hg19")
if(!isNamespaceLoaded("IlluminaHumanMethylationEPICanno.ilm10b2.hg19")) attachNamespace("IlluminaHumanMethylationEPICanno.ilm10b2.hg19")

annot450k = minfi::getAnnotation(IlluminaHumanMethylation450kanno.ilmn12.hg19::IlluminaHumanMethylation450kanno.ilmn12.hg19)
annotEPIC = minfi::getAnnotation(IlluminaHumanMethylationEPICanno.ilm10b2.hg19::IlluminaHumanMethylationEPICanno.ilm10b2.hg19)

## Remove "ch" probes
annot450k = subset(annot450k, substring(Name, 1,2) != "ch")
annotEPIC = subset(annotEPIC, substring(Name, 1,2) != "ch")

## Remove sex probes
annot450k = subset(annot450k, !chr %in% c("chrX","chrY"))
annotEPIC = subset(annotEPIC, !chr %in% c("chrX","chrY"))

ILMN450k.GR = GenomicRanges::GRanges(seqnames = annot450k$chr, 
                                     ranges = IRanges::IRanges(start = annot450k$pos, end = annot450k$pos), 
                                     strand = annot450k$strand, 
                                     ILMNID = rownames(annot450k))

ILMNEPIC.GR = GenomicRanges::GRanges(seqnames = annotEPIC$chr,
                                     ranges = IRanges::IRanges(start = annotEPIC$pos, end = annotEPIC$pos),
                                     strand = annotEPIC$strand,
                                     ILMNID = rownames(annotEPIC))

ILMN450kEPIC.GR = subsetByOverlaps(ILMN450k.GR, ILMNEPIC.GR, ignore.strand = TRUE)
ILMN450kEPIC.GR_38 = REMP::.liftOver_Hg19toHg38(ILMN450kEPIC.GR)

### Alu demo
Alu.GR <- fetchRMSK(REtype = "Alu", genome = "hg19", verbose = TRUE)
Alu38.GR <- fetchRMSK(REtype = "Alu", genome = "hg38", verbose = TRUE)

# Keep Alu only with ILMN profiled CpG
Alu_ILMN.GR = subsetByOverlaps(Alu.GR, ILMN450kEPIC.GR, ignore.strand = TRUE)
Alu38_ILMN.GR = subsetByOverlaps(Alu38.GR, ILMN450kEPIC.GR_38, ignore.strand = TRUE)

# 
RE_CpG <- findRECpG(RE = Alu_ILMN.GR, REtype = "Alu", genome = "hg19", verbose = TRUE)
RE38_CpG <- findRECpG(RE = Alu38_ILMN.GR, REtype = "Alu", genome = "hg38", verbose = TRUE)

RE_CpG_flanking <- .twoWayFlank(granges(RE_CpG), 900)
RE38_CpG_flanking <- .twoWayFlank(granges(RE38_CpG), 900)

HITS <- findOverlaps(RE_CpG_flanking, ILMN450kEPIC.GR, ignore.strand = TRUE)
HITS_38 <- findOverlaps(RE38_CpG_flanking, ILMN450kEPIC.GR_38, ignore.strand = TRUE)

## Part 1: RE CpG
RE_NeibCpG <- RE_CpG[queryHits(HITS), ]
RE38_NeibCpG <- RE38_CpG[queryHits(HITS_38), ]

## Total RE that can be predicted (contains neighboring CpGs within
## given window)

## Part 2: Neighboring ILMN CpG
RE_NeibCpG_ILMN <- ILMN450kEPIC.GR[subjectHits(HITS), ]
RE_NeibCpG_ILMN <- DataFrame(RE_NeibCpG_ILMN.GR = granges(RE_NeibCpG_ILMN), 
                             mcols(RE_NeibCpG_ILMN))
mcols(RE_NeibCpG) <- DataFrame(mcols(RE_NeibCpG), RE_NeibCpG_ILMN)

#
RE38_NeibCpG_ILMN <- ILMN450kEPIC.GR_38[subjectHits(HITS_38), ]
RE38_NeibCpG_ILMN <- DataFrame(RE_NeibCpG_ILMN.GR = granges(RE38_NeibCpG_ILMN), 
                             mcols(RE38_NeibCpG_ILMN))
mcols(RE38_NeibCpG) <- DataFrame(mcols(RE38_NeibCpG), RE38_NeibCpG_ILMN)

## Remove singleton (RE CpGs with only one neighboring ILMN CpG)
RE_NeibCpG$distance <- abs(start(RE_NeibCpG$RE_NeibCpG_ILMN.GR) - start(RE_NeibCpG))
RE_NeibCpG <- RE_NeibCpG[RE_NeibCpG$distance > 1,]
RE_NeibCpG_GR <- granges(RE_NeibCpG)
RE_NeibCpG <- subsetByOverlaps(RE_NeibCpG, unique(RE_NeibCpG_GR[duplicated(RE_NeibCpG_GR)]))

Alu_ILMN.GR <- Alu_ILMN.GR[Alu_ILMN.GR$Index %in% unique(RE_NeibCpG$Index)]

#
RE38_NeibCpG$distance <- abs(start(RE38_NeibCpG$RE_NeibCpG_ILMN.GR) - start(RE38_NeibCpG))
RE38_NeibCpG <- RE38_NeibCpG[RE38_NeibCpG$distance > 1,]
RE38_NeibCpG_GR <- granges(RE38_NeibCpG)
RE38_NeibCpG <- subsetByOverlaps(RE38_NeibCpG, unique(RE38_NeibCpG_GR[duplicated(RE38_NeibCpG_GR)]))

Alu38_ILMN.GR <- Alu38_ILMN.GR[Alu38_ILMN.GR$Index %in% unique(RE38_NeibCpG$Index)]


## Find same Alu between two build:
Alu_ILMN.GR$id <- paste0(seqnames(Alu_ILMN.GR), strand(Alu_ILMN.GR), Alu_ILMN.GR$repName, Alu_ILMN.GR$swScore)
Alu38_ILMN.GR$id <- paste0(seqnames(Alu38_ILMN.GR), strand(Alu38_ILMN.GR), Alu38_ILMN.GR$repName, Alu38_ILMN.GR$swScore)

commonid <- intersect(Alu_ILMN.GR$id, Alu38_ILMN.GR$id )

Alu_ILMN.GR_common <- Alu_ILMN.GR[match(commonid, Alu_ILMN.GR$id)]
Alu38_ILMN.GR_common <- Alu38_ILMN.GR[match(commonid, Alu38_ILMN.GR$id)]

## 
Alu38_ILMN.GR_common_LO <- REMP::.liftOver_Hg38toHg19(Alu38_ILMN.GR_common)
Alu_ILMN.GR_common_true <- subsetByOverlaps(Alu_ILMN.GR_common, Alu38_ILMN.GR_common_LO, type = "equal")

Alu_ILMN.GR_common_LO <- REMP::.liftOver_Hg19toHg38(Alu_ILMN.GR_common)
Alu38_ILMN.GR_common_true <- subsetByOverlaps(Alu38_ILMN.GR_common, Alu_ILMN.GR_common_LO, type = "equal")

identical(Alu_ILMN.GR_common_true$id, Alu38_ILMN.GR_common_true$id)
identical(Alu_ILMN.GR_common_true$repName, Alu38_ILMN.GR_common_true$repName)
identical(Alu_ILMN.GR_common_true$swScore, Alu38_ILMN.GR_common_true$swScore)


#
set.seed(2022)
ind <- sample(seq_len(length(Alu_ILMN.GR_common_true)), 500)

Alu.hg19.demo = Alu_ILMN.GR_common_true[ind, ]
Alu.hg19.demo$id <- NULL

Alu.hg38.demo = Alu38_ILMN.GR_common_true[ind, ]
Alu.hg38.demo$id <- NULL

Alu.hg19.demo <- sort(Alu.hg19.demo)
Alu.hg38.demo <- sort(Alu.hg38.demo)

Alu.hg38.demo_LO <- REMP::.liftOver_Hg38toHg19(Alu.hg38.demo)
Alu.hg19.demo_LO <- REMP::.liftOver_Hg19toHg38(Alu.hg19.demo)

identical(granges(Alu.hg38.demo_LO), granges(Alu.hg19.demo))
identical(granges(Alu.hg19.demo_LO), granges(Alu.hg38.demo))

#####################################
Alu.hg19.demo
Alu.hg38.demo

seqnames(Alu.hg19.demo)
seqnames(Alu.hg38.demo)

## Index must be Rle
Alu.hg19.demo$Index
Alu.hg38.demo$Index

use_data(Alu.hg19.demo, overwrite = TRUE)
use_data(Alu.hg38.demo, overwrite = TRUE)

#' @title Repetitive element methylation prediction
#'
#' @description
#' \code{remp} is used to predict genomewide methylation levels of locus-specific repetitive elements (RE).
#' Two major RE types in human, Alu element (Alu) and LINE-1 (L1) are available.
#'
#' @param methyDat A \code{\link{RatioSet}}, \code{\link{GenomicRatioSet}}, \code{\link{DataFrame}},
#' \code{data.table}, \code{data.frame}, or \code{matrix} of Illumina BeadChip methylation data
#' (450k or EPIC array) or Illumina methylation percentage estimates by sequencing. See Details. 
#' Alternatively, user can also specify a pre-built data template (see \code{\link{rempTemplate}}).
#' \code{remp} to carry out the prediction. See \code{\link{rempTemplate}}. With template specified, \code{methyDat},
#' \code{REtype}, \code{parcel}, and \code{work.dir} can be skipped.
#' @param REtype Type of RE. Currently \code{"Alu"}, \code{"L1"}, and \code{"ERV"} are supported. If \code{NULL}, 
#' the type of RE will be extracted from \code{parcel}.
#' @param Seq.GR A \code{\link{GRanges}} object containing genomic locations of the CpGs profiled by sequencing
#' platforms. This parameter should not be \code{NULL} if the input methylation data \code{methyDat} are
#' obtained by sequencing. Note that the genomic location can be in either hg19 or hg38 build. 
#' See details in \code{\link{initREMP}}.
#' @param parcel An \code{\link{REMParcel}} object containing necessary data to carry out the
#' prediction. If \code{NULL}, \code{REtype} must specify a type of RE so that the function can search the
#' \code{.rds} data file in \code{work.dir} exported by \code{\link{initREMP}} (with \code{export = TRUE})
#' or \code{\link{saveParcel}}.
#' @param work.dir Path to the directory where the annotation data generated by \code{\link{initREMP}}
#' are saved. Valid when the argument \code{parcel} is missing. If not specified, temporary directory
#' \code{tempdir()} will be used. If specified, the directory path has to be the same as the
#' one specified in \code{\link{initREMP}} or in \code{\link{saveParcel}}.
#' @param win An integer specifying window size to confine the upstream and downstream flanking
#' region centered on the predicted CpG in RE for prediction. Default = \code{1000}. See Details.
#' @param method Name of model/approach for prediction. Currently \code{"rf"} (Random Forest),
#' \code{"xgbTree"} (Extreme Gradient Boosting), \code{"svmLinear"} (SVM with linear kernel), \code{"svmRadial"}
#' (SVM with radial kernel), and \code{"naive"} (carrying over methylation values of the closest
#' CpG site) are available. Default = \code{"rf"} (Random Forest). See Details.
#' @param autoTune Logical parameter. If \code{TRUE}, a 3-time repeated 5-fold cross validation
#' will be performed to determine the best model parameter. If \code{FALSE}, the \code{param} option
#' (see below) must be specified. Default = \code{TRUE}. Auto-tune will be disabled using Random Forest.
#' See Details.
#' @param param A list specifying fixed model tuning parameter(s) (not applicable for Random Forest, see Details).
#' For Extreme Gradient Boosting, \code{param} list must contain '\code{$nrounds}', '\code{$max_depth}',
#' '\code{$eta}', '\code{$gamma}', '\code{$colsample_bytree}', '\code{$min_child_weight}', and '\code{$subsample}'.
#' See \code{xgbTree} in package \code{caret}. For SVM, \code{param} list must contain '\code{$C}' (cost) for linear kernel
#' or '\code{$sigma}' and '\code{$C}' for radial basis function kernel. This parameter is valid only
#' when \code{autoTune = FALSE}.
#' @param seed Random seed for Random Forest model for reproducible prediction results.
#' Default is \code{NULL}, which generates a seed.
#' @param ncore Number of cores used for parallel computing. By default, max number of cores available
#' in the machine will be utilized. If \code{ncore = 1}, no parallel computing is allowed.
#' @param BPPARAM An optional \code{\link{BiocParallelParam}} instance determining the parallel back-end to
#' be used during evaluation. If not specified, default back-end in the machine will be used.
#' @param verbose Logical parameter. Should the function be verbose?
#'
#' @details
#' Before running \code{remp}, user should make sure the methylation data have gone through
#' proper quality control, background correction, and normalization procedures. Both beta value
#' and M value are allowed. Rows represents probes and columns represents samples. For array data,
#' please make sure to have row names that specify the Illumina probe ID (i.e. cg00000029). For sequencing
#' data, please provide the genomic location of CpGs in a \code{\link{GRanges}} obejct and
#' specify it using \code{Seq.GR} parameter. \code{win = 1000} is based on previous findings showing that
#' neighboring CpGs are more likely to be co-modified within 1000 bp. User can specify narrower window size
#' for slight improvement of prediction accuracy at the cost of less predicted RE. Window size greater than 1000 is not
#' recommended as the machine learning models would not be able to learn much userful information
#' for prediction but introduce noise. Random Forest model (\code{method = "rf"}) is recommented
#' as it offers more accurate prediction and it also enables prediction reliability functionality.
#' Prediction reliability is estimated by conditional standard deviation using Quantile Regression Forest.
#' Please note that if parallel computing is allowed, parallel Random Forest
#' (powered by package \code{\link{ranger}}) will be used automatically. The performance of
#' Random Forest model is often relatively insensitive to the choice of \code{mtry}.
#' Therefore, auto-tune will be turned off using Random Forest and \code{mtry} will be set to one third
#' of the total number of predictors. For SVM, if \code{autoTune = TRUE}, preset tuning parameter
#' search grid can be access and modified using \code{\link{remp_options}}.
#'
#' @return A \code{\link{REMProduct}} object containing predicted RE methylation results.
#'
#' @seealso See \code{\link{initREMP}} to prepare necessary annotation database before running \code{remp}.
#'
#' @examples
#' # Obtain example Illumina example data (450k)
#' if (!exists("GM12878_450k")) 
#'   GM12878_450k <- getGM12878("450k")
#' 
#' # Make sure you have run 'initREMP' first. See ?initREMP.
#' if (!exists("remparcel")) {
#'   data(Alu.hg19.demo)
#'   remparcel <- initREMP(arrayType = "450k",
#'                         REtype = "Alu",
#'                         annotation.source = "AH",
#'                         genome = "hg19",
#'                         RE = Alu.hg19.demo,
#'                         ncore = 1,
#'                         verbose = TRUE)
#' }
#' 
#' # With data template pre-built. See ?rempTemplate.
#' if (!exists("template")) 
#'   template <- rempTemplate(GM12878_450k, 
#'                            parcel = remparcel, 
#'                            win = 1000, 
#'                            verbose = TRUE)
#' 
#' # Run remp with pre-built template:
#' remp.res <- remp(template, ncore = 1)
#' 
#' # Or run remp without pre-built template (identical results):
#' \dontrun{
#'   remp.res <- remp(GM12878_450k, 
#'                    REtype = "Alu", 
#'                    parcel = remparcel, 
#'                    ncore = 1,
#'                    verbose = TRUE)
#' }
#' 
#' remp.res
#' details(remp.res)
#' rempB(remp.res) # Methylation data (beta value)
#' 
#' # Extract CpG location information. 
#' # This accessor is inherit from class 'RangedSummarizedExperiment')
#' rowRanges(remp.res)
#' 
#' # RE annotation information
#' rempAnnot(remp.res)
#' 
#' # Add gene annotation
#' remp.res <- decodeAnnot(remp.res, type = "symbol")
#' rempAnnot(remp.res)
#' 
#' # (Recommended) Trim off less reliable prediction
#' remp.res <- rempTrim(remp.res)
#' 
#' # Obtain RE-level methylation (aggregate by mean)
#' remp.res <- rempAggregate(remp.res)
#' rempB(remp.res) # Methylation data (beta value)
#' 
#' # Extract RE location information
#' rowRanges(remp.res)
#' 
#' # Density plot across predicted RE
#' remplot(remp.res)
#' 
#' @export
remp <- function(methyDat = NULL, 
                 REtype = c("Alu", "L1", "ERV"), 
                 Seq.GR = NULL,
                 parcel = NULL, 
                 work.dir = tempdir(), 
                 win = 1000,
                 method = c("rf", "xgbTree", "svmLinear", "svmRadial", "naive"),
                 autoTune = TRUE, 
                 param = NULL, 
                 seed = NULL, 
                 ncore = NULL, 
                 BPPARAM = NULL,
                 verbose = FALSE) {
  currenT <- Sys.time()

  if (is.null(methyDat)) stop("Methylation dataset (methyDat) is missing.")
  if (!is.null(Seq.GR)) {
    .isGROrStop(Seq.GR)
    names(Seq.GR) <- NULL
  }
  if (!is.null(param) & !is(param, "list")) stop("Tuning parameter(s) (param) must be a list object.")

  method <- match.arg(method)
  if (method != "naive" & !autoTune & is.null(param)) {
    message("Tuning parameter for method = '", method, "' is missing!")
    stop("You turned off auto-tune. Please specify tuning parameter(s).")
  }

  if (method %in% c("rf", "xgbTree")) {
    if (is.null(seed)) {
      seed <- sample(.Random.seed, 1)
    }
    message("Using random seed = ", seed)

    if (method == "rf") {
      autoTune <- FALSE # No need to tune RF mtry
      param <- list(mtry = round(length(remp_options(".default.predictors")) / 3)) # one third of number of predictors
      if (verbose) {
        message("Note: Auto-tune is disabled for Random Forest and mtry is set to 
                one third of number of predictors (", param$mtry, ").")
      }
    }
  }

  if (is.null(ncore)) {
    ncore <- 1
  }

  if (!is(methyDat, "template")) {
    if (win > remp_options(".default.max.flankWindow")) {
      stop(
        "Flanking window size cannot be greater than ",
        remp_options(".default.max.flankWindow"),
        ". Please see ?remp_options for details."
      )
    }

    ## Groom methylation data
    methyDat <- grooMethy(methyDat, Seq.GR, verbose = verbose)
    arrayType <- gsub("IlluminaHumanMethylation", "", methyDat@annotation["array"])
    
    methyDat <- minfi::getM(methyDat)
    
    ## Get the REMParcel object ready
    if (is.null(parcel)) {
      if (is.null(REtype)) {
        stop("REtype and parcel parameters cannot be both empty!")
      } else {
        REtype <- match.arg(REtype)
      }

      subDirName <- paste0("REMP.data.", arrayType)
      work.dir <- .forwardSlashPath(work.dir)
      data.dir <- .forwardSlashPath(file.path(work.dir, subDirName))

      path_to_parcel <- file.path(
        data.dir,
        paste0(
          "REMParcel_",
          REtype, "_",
          arrayType, "_",
          remp_options(".default.max.flankWindow"),
          ".rds"
        )
      )

      if (file.exists(path_to_parcel)) {
        remparcel <- readRDS(path_to_parcel)
      } else {
        stop("Necessary annotation data files are missing. Please make sure:
             1) you have run initREMP() first to generate the annotation data. See ?initREMP for details; 
             2) the RE type (Alu/L1) and/or platform (450k/EPIC/Sequencing) match the annotation data generated by initREMP().
             3) parameter 'parcel' is speficied, if the generated annotation data were not exported to local machine;
             4) if exported, your working directory specified is the same as you did in initREMP().")
      }
    } else {
      .isREMParcelOrStop(parcel)
      if (getParcelInfo(parcel)$platform != arrayType) {
        stop(paste0(
          "The methylation data type: ", arrayType,
          " does not match the REMParcel type: ",
          getParcelInfo(parcel)$platform
        ))
      }
      remparcel <- parcel
    }

    TEMPLATE <- rempTemplate(methyDat, Seq.GR, remparcel, win, FALSE)
  } else {
    TEMPLATE <- methyDat
  }

  ## Setup backend for paralell computing
  be <- getBackend(ncore, BPPARAM, verbose)
  ncore <- BiocParallel::bpnworkers(be)

  message(
    "Start RE methylation prediction with ",
    BiocParallel::bpnworkers(be), " core(s) ..."
  )

  REtype <- TEMPLATE$REtype # make sure REtype is consistent with REMParcel
  genome <- TEMPLATE$genome
  arrayType <- TEMPLATE$arrayType
  methyDat <- TEMPLATE$NBCpG_methyDat
  RE_NeibCpG <- TEMPLATE$NBCpG_GR
  RE_CpG_ILMN_DATA <- TEMPLATE$RECpG_methyDat
  RE_CpG_ILMN <- TEMPLATE$RECpG_GR

  refgene_main <- TEMPLATE$refgene
  RE_refGene.original <- TEMPLATE$RE

  samplenames <- colnames(methyDat)
  sampleN <- length(samplenames)

  #######################################

  if ("naive" == method) {
    BEST_TUNE <- DataFrame(matrix(NA, nrow = sampleN, ncol = 1))
    colnames(BEST_TUNE) <- "Not_Applicable"
    libToLoad <- NULL
    method_text <- "Naive (nearest CpG)"
    QCname_text <- "N/A"
  }
  if ("rf" == method) {
    BEST_TUNE <- DataFrame(matrix(NA, nrow = sampleN, ncol = 1))
    colnames(BEST_TUNE) <- "mtry"
    libToLoad <- NULL
    method_text <- "Random Forest"
    QCname_text <- "Quantile Regression Forest"
  }
  if ("xgbTree" == method) {
    BEST_TUNE <- DataFrame(matrix(NA, nrow = sampleN, ncol = 7))
    colnames(BEST_TUNE) <- c("nrounds", "max_depth", "eta", "gamma", "colsample_bytree", "min_child_weight", "subsample")
    libToLoad <- NULL
    method_text <- "Extreme Gradient Boosting"
    QCname_text <- "N/A"
  }
  if ("svmLinear" == method) {
    BEST_TUNE <- DataFrame(matrix(NA, nrow = sampleN, ncol = 1))
    colnames(BEST_TUNE) <- "cost"
    libToLoad <- c("kernlab")
    method_text <- "SVM(Linear)"
    QCname_text <- "N/A"
  }
  if ("svmRadial" == method) {
    BEST_TUNE <- DataFrame(matrix(NA, nrow = sampleN, ncol = 2))
    colnames(BEST_TUNE) <- c("sigma", "cost")
    libToLoad <- c("kernlab")
    method_text <- "SVM(Radial)"
    QCname_text <- "N/A"
  }
  rownames(BEST_TUNE) <- samplenames

  if (verbose) {
    message("Prediction model: ", method_text, "\nQC model: ", QCname_text)
  }

  cpgRanges <- unique(RE_NeibCpG[, "RE.Index"])
  RE_annotation <- subsetByOverlaps(RE_refGene.original, cpgRanges)
  # Note: Some REs have overlapping regions thus there can be more than one REs mapped to one RE-CpG.
  RE_annotation_name <- colnames(mcols(RE_annotation))
  regionCode <- mcols(RE_annotation)[remp_options(".default.genomicRegionColNames")]
  RE_annotation <- RE_annotation[, RE_annotation_name[!RE_annotation_name %in%
    remp_options(".default.genomicRegionColNames")]]

  # Initiate container
  REMP_PREDICT_CpG <- matrix(NA, nrow = length(cpgRanges), ncol = sampleN)
  REMP_PREDICT_QC <- matrix(NA, nrow = length(cpgRanges), ncol = sampleN)
  REMP_PREDICT_IMP <- setNames(DataFrame(matrix(NA,
    nrow = length(remp_options(".default.predictors")),
    ncol = sampleN
  )), samplenames)
  rownames(REMP_PREDICT_IMP) <- remp_options(".default.predictors")

  ## RE coverage
  RE_COVERAGE <- .coverageStats_RE(RE_annotation, regionCode, cpgRanges, RE_CpG_ILMN,
    REtype,
    indent = "    ", verbose
  )

  # Gene coverage
  GENE_COVERAGE <- .coverageStats_GENE(regionCode, refgene_main,
    REtype,
    indent = "    ", verbose
  )

  ## Prediction
  tRun <- .timeTrace(currenT)
  currenT <- tRun$t

  for (i in seq_len(sampleN)) {
    message("Predicting sample ", samplenames[i], " ...")
    RE_CpG_ILMN_OneMethyDat <- RE_CpG_ILMN_DATA[, samplenames[i]]
    # identical(RE_CpG_ILMN$Index, names(RE_CpG_ILMN_OneMethyDat))

    ### To be predicted
    if(verbose) message("    Loading methylation data for prediction...")
    RE_unprf_neib <- .loadMethy(methyDat[, i, drop = FALSE], RE_NeibCpG)
    # summary(RE_unprf_neib$distance)

    ### For model training
    RE_unprf_neib <- unique(RE_unprf_neib[, c("RE.Index", remp_options(".default.predictors"))])
    HITS <- findOverlaps(RE_unprf_neib, RE_CpG_ILMN, ignore.strand = TRUE)
    RE_prf_neib <- RE_unprf_neib[queryHits(HITS), ]
    RE_prf_neib$Methy <- RE_CpG_ILMN_OneMethyDat[subjectHits(HITS)]
    # summary(RE_prf_neib$distance.min2)
    # cor(RE_prf_neib$Methy.min, RE_prf_neib$Methy)

    P_basic <- .modelTrain(
      RE_prf_neib, remp_options(".default.predictors"), method,
      autoTune, param, be, seed, verbose
    )

    best_tune <- as.numeric(P_basic$bestTune)

    if (method != "naive") {
      REMP_PREDICT_IMP[, i] <- P_basic$importance[remp_options(".default.predictors"), "Overall"]
      for (k in seq_len(length(best_tune)))
      {
        BEST_TUNE[i, k] <- best_tune[k]
      }
    }

    ##############################################################################################

    newdata <- as.matrix(mcols(RE_unprf_neib)[, remp_options(".default.predictors")])

    if (method != "rf") {
      if (ncore > 1) {
        BiocParallel::bpstart(be)
        .bploadLibraryQuiet(libToLoad, be)
        ITER <- .iblkrow(newdata, chunks = ncore)
        REMP_PREDICT_CpG[, i] <- BiocParallel::bpiterate(ITER, .predictREMP,
          model = P_basic$model,
          BPPARAM = be,
          REDUCE = c,
          reduce.in.order = TRUE
        )
        BiocParallel::bpstop(be)
      } else {
        REMP_PREDICT_CpG[, i] <- .predictREMP(newdata, P_basic$model)
      }
      REMP_PREDICT_QC <- NULL
    } else {
      REMP_PREDICT_CpG[, i] <- predict(P_basic$model, newdata,
        type = "response",
        num.threads = ncore
      )$predictions
      REMP_PREDICT_QC[, i] <- .QTF(
        rangerObj = P_basic$model,
        x = as.matrix(mcols(RE_prf_neib)[, remp_options(".default.predictors")]),
        y = as.numeric(mcols(RE_prf_neib)[, "Methy"]),
        newX = newdata,
        seed = seed,
        num.threads = ncore
      )
    }

    tRun <- .timeTrace(currenT)
    currenT <- tRun$t
    message(
      "    ", samplenames[i], " completed! ", sampleN - i,
      " sample(s) left ...", tRun$t_text
    )
  }

  message("Done.")

  remproduct <- REMProduct(
    REtype = REtype, genome = genome, 
    platform = arrayType, win = as.character(win),
    predictModel = method_text, QCModel = QCname_text,
    rempM = REMP_PREDICT_CpG, rempQC = REMP_PREDICT_QC,
    cpgRanges = cpgRanges, sampleInfo = BEST_TUNE,
    REannotation = RE_annotation,
    RECpG = RE_CpG_ILMN,
    regionCode = regionCode,
    refGene = refgene_main,
    varImp = REMP_PREDICT_IMP,
    REStats = RE_COVERAGE, GeneStats = GENE_COVERAGE,
    Seed = seed
  )
  return(remproduct)
} ## End of remp



## --------------------------------------
## Internal functions

.toIndicator <- function(object) {
  genomicRegionInd <- matrix(0L,
    nrow = length(object),
    ncol = length(remp_options(".default.genomicRegionColNames"))
  )
  genomicRegionInd[!is.na(mcols(object)[, remp_options(".default.genomicRegionColNames")])] <- 1L
  genomicRegionInd <- DataFrame(genomicRegionInd)
  colnames(genomicRegionInd) <- remp_options(".default.genomicRegionColNames")
  mcols(object)[, remp_options(".default.genomicRegionColNames")] <- genomicRegionInd
  return(object)
}

.loadMethy <- function(methyDatOne, RE_NeibCpG) {
  RE_NeibCpG$Methy.all <- methyDatOne[RE_NeibCpG$Methy.ptr, ]
  RE_NeibCpG$Methy.min <- methyDatOne[RE_NeibCpG$Methy.ptr.min, ]
  RE_NeibCpG$Methy.min2 <- methyDatOne[RE_NeibCpG$Methy.ptr.min2, ]

  RE_NeibCpG.DF <- mcols(RE_NeibCpG)[, c("RE.CpG.ID", "distance_cat", "Methy.all")]
  catList <- sort(unique(RE_NeibCpG.DF$distance_cat))

  Methy_all_agg <- .aggregateNeib(
    Methy.all ~ RE.CpG.ID, RE_NeibCpG.DF,
    function(x) sd(x),
    c("RE.CpG.ID", "Methy.std")
  )

  ## Moving average
  cati <- NULL
  for (i in catList)
  {
    cati <- c(cati, i)
    sub_RE_NeibCpG.DF <- RE_NeibCpG.DF[RE_NeibCpG.DF$distance_cat %in% cati, ]
    sub_Methy_all_agg <- .aggregateNeib(
      Methy.all ~ RE.CpG.ID,
      sub_RE_NeibCpG.DF,
      function(x) mean(x),
      c("RE.CpG.ID", paste0("Methy.mean.mov", i))
    )
    Methy_all_agg <- merge(Methy_all_agg, sub_Methy_all_agg,
      by = "RE.CpG.ID", all.x = TRUE
    )
  }

  ## carry over missing
  for (j in catList + 2)
  {
    current <- Methy_all_agg[, j]
    k <- 1
    while (any(is.na(current)) & j + k <= 6) {
      current[is.na(current)] <- Methy_all_agg[is.na(current), j + k]
      k <- k + 1
    }
    Methy_all_agg[, j] <- current
  }

  RE_NeibCpG_meta <- mcols(RE_NeibCpG)
  mcols(RE_NeibCpG) <- cbind(
    RE_NeibCpG_meta,
    Methy_all_agg[base::match(
      RE_NeibCpG_meta$RE.CpG.ID,
      Methy_all_agg$RE.CpG.ID
    ), -1]
  )
  return(RE_NeibCpG)
}

.aggregateNeib <- function(relation, d, FUN, newColNames) {
  d_agg <- aggregate(relation, data = d, FUN)
  d_agg <- DataFrame(d_agg[, 1], DataFrame(d_agg[, -1]))
  return(setNames(d_agg, newColNames))
}

.modelTrain <- function(d, varname, method, autoTune, param,
                        be, seed, verbose) {
  ncore <- BiocParallel::bpnworkers(be)

  d <- as.data.frame(mcols(d)[, c("Methy", varname)])

  if (autoTune) {
    trC.method <- "repeatedcv"
    trC.number <- 5
    trC.repeats <- 3
  } else {
    trC.method <- "none"
    trC.number <- 1
    trC.repeats <- NA
  }

  if (method == "naive") {
    modelFit <- list(
      model = NULL, QC = NULL, importance = NULL, bestTune = NULL,
      name = "naive"
    )
  } else {

    # set up the tuning grid
    if (method == "rf") {
      param_text <- "mtry"
    }

    if (method == "svmLinear") {
      if (autoTune) param <- remp_options(".default.svmLinear.tune")
      grid <- expand.grid(C = param$C)
      param_text <- "cost"
    }

    if (method == "svmRadial") {
      if (autoTune) param <- remp_options(".default.svmRadial.tune")
      grid <- expand.grid(sigma = param$sigma, C = param$C)
      param_text <- c("sigma", "cost")
    }

    if (method == "xgbTree") {
      if (autoTune) param <- remp_options(".default.xgbTree.tune")
      grid <- expand.grid(
        nrounds = param$nrounds, max_depth = param$max_depth, eta = param$eta,
        gamma = param$gamma, colsample_bytree = param$colsample_bytree,
        min_child_weight = param$min_child_weight, subsample = param$subsample
      )
      param_text <- c("nrounds", "max_depth", "eta", "gamma", "colsample_bytree", "min_child_weight", "subsample")
    }

    if (verbose & !autoTune) {
      message(
        "    Pre-specified tuning parameter: ",
        paste(paste(param_text, unlist(param), sep = " = "), collapse = ", ")
      )
    }

    if (method == "rf") {
      doQC <- TRUE
      model.tune <- ranger::ranger(Methy ~ .,
        data = d, mtry = param$mtry, num.threads = ncore,
        importance = "permutation", keep.inbag = FALSE,
        scale.permutation.importance = TRUE, seed = seed
      )
      var_imp <- ranger::importance(model.tune)
      var_imp <- data.frame(
        Predictor = names(var_imp),
        Overall = var_imp
      )
      best_param <- model.tune$mtry
      model <- model.tune
    } else {
      set.seed(seed)
      if (ncore > 1) {
        trC <- caret::trainControl(
          method = trC.method, number = trC.number,
          repeats = trC.repeats, allowParallel = TRUE
        )
        cluster <- parallel::makeCluster(ncore)
        doParallel::registerDoParallel(cluster)
        model.tune <- caret::train(Methy ~ .,
          data = d, method = method,
          tuneGrid = grid, trControl = trC, importance = TRUE
        )
        parallel::stopCluster(cluster)
        registerDoSEQ()
      } else {
        trC <- caret::trainControl(
          method = trC.method, number = trC.number,
          repeats = trC.repeats, allowParallel = FALSE
        )
        model.tune <- caret::train(Methy ~ .,
          data = d, method = method,
          tuneGrid = grid, trControl = trC, importance = TRUE
        )
      }

      var_imp <- caret::varImp(model.tune)
      var_imp <- data.frame(
        Predictor = rownames(var_imp$importance),
        Overall = var_imp$importance
      )
      best_param <- model.tune$bestTune
      model <- model.tune$finalModel
    }

    if (verbose & autoTune) {
      message(
        "    Best tuning parameter found: ",
        paste(paste(param_text, unlist(best_param), sep = " = "), collapse = ", ")
      )
    }

    modelFit <- list(
      model = model,
      importance = var_imp,
      bestTune = best_param,
      name = method
    )
  }
  return(modelFit)
}

# Prediction reliability based on Quantile random forest
.QTF <- function(rangerObj, x, y, newX, seed, num.threads) {
  nodesX <- predict(rangerObj, x,
    num.threads = num.threads,
    type = "terminalNodes"
  )$predictions
  nnodes <- max(nodesX)
  ntree <- ncol(nodesX)
  n <- nrow(x)
  valuesNodes <- matrix(nrow = nnodes, ncol = ntree)

  set.seed(seed)
  for (tree in seq_len(ntree)) {
    shuffledNodes <- nodesX[rank(ind <- sample(seq_len(n), n)), tree]
    useNodes <- sort(unique(as.numeric(shuffledNodes)))
    valuesNodes[useNodes, tree] <- y[ind[base::match(useNodes, shuffledNodes)]]
  }

  predictNodes <- predict(rangerObj, newX, num.threads,
    type = "terminalNodes"
  )$predictions
  valuesPredict <- 0 * predictNodes
  for (tree in seq_len(ntree)) {
    valuesPredict[, tree] <- valuesNodes[ predictNodes[, tree], tree]
  }
  result <- apply(valuesPredict, 1, sd)
  return(result)
}

.predictREMP <- function(newdata, model) {
  if (is.null(model)) {
    data.frame(newdata)$Methy.min # newdata is matrix
  } else {
    if (any(class(model) %in% c("ksvm", "vm"))) {
      kernlab::predict(model, newdata)
    } # kernlab has its own predict function
    else {
      predict(model, newdata)
    }
  }
}

# Count the number of RE by gene type (NM/NR/Total), used by
# .coverageStats_RE()
.countREByGene <- function(RE_annotation, RE_list) {
  Total <- length(RE_annotation)
  RE_annotation_sub <- RE_annotation[runValue(RE_annotation$Index) %in% RE_list]
  NM <- sum(!is.na(RE_annotation_sub$InNM))
  NR <- sum(!is.na(RE_annotation_sub$InNR))
  Gene <- sum(!is.na(RE_annotation_sub$InNM) | !is.na(RE_annotation_sub$InNR))
  Intergenic <- sum(is.na(RE_annotation_sub$InNM) & is.na(RE_annotation_sub$InNR))

  return(c(Total = Total, NM = NM, NR = NR, Gene = Gene, Intergenic = Intergenic))
}

.coverageStats_RE <- function(RE_annotation, regionCode, cpgRanges, RE_CpG_ILMN,
                              REtype, indent, verbose) {
  mcols(RE_annotation) <- cbind(mcols(RE_annotation), regionCode)
  mcols(cpgRanges) <- mcols(RE_annotation[base::match(as.character(cpgRanges$RE.Index), runValue(RE_annotation$Index))])

  RE_annotation_prf <- subsetByOverlaps(RE_annotation, RE_CpG_ILMN, ignore.strand = TRUE)
  cpgRanges_prf <- subsetByOverlaps(cpgRanges, RE_CpG_ILMN, ignore.strand = TRUE)

  # Profiled RE for prediction
  cvr_RE_win_list <- runValue(RE_annotation_prf$Index)

  # Profiled RE for prediction + predicted unprofiled RE
  cvr_unRE_win_list <- runValue(RE_annotation$Index)

  REStats <- DataFrame(cbind(
    .countREByGene(RE_annotation_prf, cvr_RE_win_list),
    .countREByGene(cpgRanges_prf, cvr_RE_win_list),
    .countREByGene(RE_annotation, cvr_unRE_win_list),
    .countREByGene(cpgRanges, cvr_unRE_win_list)
  ))

  colnames(REStats) <- c("Trained_RE", "Trained_RECpG", "Predict_RE", "Predict_RECpG")

  if (verbose) {
    .showREStats(REStats, REtype, "message", indent, TRUE)
  }

  return(REStats)
}

# Count the number of genes covered by RE, used by
# .coverageStats_GENE()
.countCoveredGene <- function(regionCode, refgene_main) {
  InNM <- na.omit(regionCode$InNM)
  InNR <- na.omit(regionCode$InNR)

  NM_ind <- as.numeric(unique(unlist(strsplit(InNM, "[|]"))))
  NR_ind <- as.numeric(unique(unlist(strsplit(InNR, "[|]"))))
  gene_ind <- c(NM_ind, NR_ind)

  NM <- length(unique(refgene_main$GeneSymbol[NM_ind]))
  NR <- length(unique(refgene_main$GeneSymbol[NR_ind]))
  Gene <- length(unique(refgene_main$GeneSymbol[gene_ind]))

  return(c(NM = NM, NR = NR, Gene = Gene))
}

# indent = "    "
.coverageStats_GENE <- function(regionCode, refgene_main,
                                REtype, indent, verbose) {
  ## Count the total number of gene (protein coding gene and noncoding RNA
  ## gene)
  totgene <- length(unique(refgene_main$GeneSymbol))
  totgene_NM <- length(unique(refgene_main[refgene_main$type == "NM", ]$GeneSymbol))
  totgene_NR <- length(unique(refgene_main[refgene_main$type == "NR", ]$GeneSymbol))

  GeneStats <- DataFrame(rbind(
    c(totgene_NM, totgene_NR, totgene),
    .countCoveredGene(regionCode, refgene_main)
  ))

  rownames(GeneStats) <- c("Total", "Predicted RE")

  if (verbose) {
    .showGeneStats(GeneStats, REtype, "message", indent)
  }

  return(GeneStats)
}

# Variable, global to REMP's namespace.
# This function is not exported to user space and does not need to be documented.
REMPOPTIONS <- settings::options_manager(
  .default.AluFamily.grep = "^Alu$",
  
  .default.L1Family.grep = "^L1$",

  .default.ERVFamily.grep = "^ERV1$|^ERVK$|^ERVL$|^ERVL-MaLR$",
  
  .default.chr = paste0("chr", c(seq_len(22), "X", "Y")),
  
  .default.GM12878.450k.URL = "http://hgdownload.cse.ucsc.edu/goldenPath/hg19/encodeDCC/wgEncodeHaibMethyl450/wgEncodeHaibMethyl450Gm12878SitesRep1.bed.gz",

  .default.RMSK.hg19.URL = "http://hgdownload.cse.ucsc.edu/goldenpath/hg19/database/rmsk.txt.gz",
  
  .default.RMSK.hg38.URL = "http://hgdownload.cse.ucsc.edu/goldenpath/hg38/database/rmsk.txt.gz",
  
  .default.refGene.hg19.URL = "https://hgdownload.soe.ucsc.edu/goldenPath/hg19/database/refGene.txt.gz",
  
  .default.refGene.hg38.URL = "https://hgdownload.soe.ucsc.edu/goldenPath/hg38/database/refGene.txt.gz",
  
  .default.AH.repeatmasker.hg19 = "AH99002",
  
  .default.AH.repeatmasker.hg38 = "AH99003",
  
  .default.AH.refgene.hg19 = "AH5040",

  .default.AH.hg38ToHg19.over.chain = "AH14108",
  
  .default.AH.hg19ToHg38.over.chain = "AH14150",
  
  .default.TSS.upstream = 2000,

  .default.TSS.downstream = 0,

  .default.max.flankWindow = 1200,

  .default.27k.total.probes = 27578,

  .default.450k.total.probes = 485577,

  .default.epic.total.probes = 866895,

  .default.450k.annotation = "ilmn12.hg19",

  .default.epic.annotation = "ilm10b2.hg19",

  .default.genomicRegionColNames = c(
    "InNM", "InNR", "InTSS", "In5UTR",
    "InCDS", "InExon", "In3UTR"
  ),

  .default.predictors = c(
    "RE.swScore", "RE.Length", "RE.CpG.density",
    "RE.InTSS", "RE.In5UTR", "RE.InCDS",
    "N.nbr", "distance.mean", "distance.std", "distance.min", "distance.min2",
    "Methy.min", "Methy.min2", "Methy.mean.mov1", "Methy.mean.mov2",
    "Methy.mean.mov3", "Methy.mean.mov4", "Methy.std"
  ),

  .default.svmLinear.tune = list(C = 2^seq(-15, 3, 2)),

  .default.svmRadial.tune = list(
    C = 2^seq(-7, 7, 2),
    sigma = 2^seq(-9, 1, 2)
  ),

  .default.xgbTree.tune = list(
    nrounds = 500,
    eta = 0.01,
    max_depth = 5,
    gamma = seq(0, 1, 0.2),
    colsample_bytree = 0.8,
    min_child_weight = 4,
    subsample = 0.8
  )
)


#' @title Set or get options for REMP package
#'
#' @description Tools to manage global setting options for REMP package.
#'
#' @param ... Option names to retrieve option values or \code{[key]=[value]} pairs to set options.
#'
#' @section Supported options:
#' The following options are supported
#' \describe{
#'  \item{\code{.default.AluFamily.grep}}{Regular expression for 'grep' to extract Alu family to be included in the prediction.}
#'  \item{\code{.default.L1Family.grep}}{Regular expression for 'grep' to extract L1 family to be included in the prediction.}
#'  \item{\code{.default.ERVFamily.grep}}{Regular expression for 'grep' to extract ERV family to be included in the prediction.}
#'  \item{\code{.default.chr}}{List of human chromosome.}
#'  \item{\code{.default.GM12878.450k.URL}}{URL to download GM12878 450k methylation profiling data.}
#'  \item{\code{.default.RMSK.hg19.URL}}{URL to download RepeatMasker database in hg19 genome.}
#'  \item{\code{.default.RMSK.hg38.URL}}{URL to download RepeatMasker database in hg38 genome.}
#'  \item{\code{.default.refGene.hg19.URL}}{URL to download refSeq gene database in hg19 genome.}
#'  \item{\code{.default.refGene.hg38.URL}}{URL to download refSeq gene database in hg38 genome.}
#'  \item{\code{.default.AH.repeatmasker.hg19}}{\code{AnnotationHub} data ID linked to RepeatMasker
#'  annotation database (Mar 2020, build hg19).}
#'  \item{\code{.default.AH.repeatmasker.hg38}}{\code{AnnotationHub} data ID linked to RepeatMasker
#'  annotation database (Sep 2021, build hg38).}
#'  \item{\code{.default.AH.refgene.hg19}}{\code{AnnotationHub} data ID linked to refSeq gene database
#'  (build hg19)}
#'  \item{\code{.default.AH.hg38ToHg19.over.chain}}{\code{AnnotationHub} hg38 to hg19 liftover chain data ID.}
#'  \item{\code{.default.AH.hg19ToHg38.over.chain}}{\code{AnnotationHub} hg19 to hg38 liftover chain data ID.}
#'  \item{\code{.default.TSS.upstream}}{Define the upstream range of transcription start site region.}
#'  \item{\code{.default.TSS.downstream}}{Define the downstream range of transcription start site region.}
#'  \item{\code{.default.max.flankWindow}}{Define the max size of the flanking window surrounding the
#'  predicted RE-CpG.}
#'  \item{\code{.default.27k.total.probes}}{Total number of probes designed in Illumina 27k array.}
#'  \item{\code{.default.450k.total.probes}}{Total number of probes designed in Illumina 450k array.}
#'  \item{\code{.default.epic.total.probes}}{Total number of probes designed in Illumina EPIC array.}
#'  \item{\code{.default.450k.annotation}}{A character string associated with the Illumina 450k
#'  array annotation dataset.}
#'  \item{\code{.default.epic.annotation}}{A character string associated with the Illumina EPIC
#'  array annotation dataset.}
#'  \item{\code{.default.genomicRegionColNames}}{Define the names of the genomic regions for prediction.}
#'  \item{\code{.default.predictors}}{Define the names of predictors for RE methylation prediction.}
#'  \item{\code{.default.svmLinear.tune}}{Define the default \code{C} (Cost) parameter for Support
#'  Vector Machine (SVM) using linear kernel.}
#'  \item{\code{.default.svmRadial.tune}}{Define the default parameters (\code{C} and \code{sigma}) for SVM
#'  using Radial basis function kernel.}
#'  \item{\code{.default.xgbTree.tune}}{Define the default parameters (\code{nrounds}, \code{eta}, \code{max_depth},
#'  \code{gamma}, \code{colsample_bytree}, \code{min_child_weight}, and \code{subsample}) for Extreme Gradient
#'  Boosting.}
#' }
#'
#' @return \code{NULL}
#'
#' @examples
#' # Display all default settings
#' remp_options()
#' 
#' # Display a specified setting
#' remp_options(".default.max.flankWindow")
#' 
#' # Change default maximum flanking window size to 2000
#' remp_options(.default.max.flankWindow = 2000)
#' 
#' # Reset all options
#' remp_reset()
#' @rdname options
#' @export
remp_options <- function(...) {
  # protect against the use of reserved words.
  settings::stop_if_reserved(...)
  REMPOPTIONS(...)
}

#' @rdname options
#' @export
remp_reset <- function() {
  settings::reset(REMPOPTIONS)
}

#' @title Get methylation data of HapMap LCL sample GM12878 profiled by
#' Illumina 450k array or EPIC array
#'
#' @description
#' \code{getGM12878} is used to obtain public available methylation profiling
#' data of HapMap LCL sample GM12878.
#'
#' @param arrayType Illumina methylation array type. Currently \code{"450k"}
#' and \code{"EPIC"} are supported. Default = \code{"450k"}.
#' @param mapGenome Logical parameter. If \code{TRUE}, function will return
#' a \code{\link{GenomicRatioSet}} object instead of a \code{link{RatioSet}}
#' object.

#' @details
#' Illumina 450k data were sourced and curated from ENCODE
#' \url{http://hgdownload.cse.ucsc.edu/goldenPath/hg19/encodeDCC/wgEncodeHaibMethyl450/wgEncodeHaibMethyl450Gm12878SitesRep1.bed.gz}.
#' Illumina EPIC data were obtained from data package \code{minfiDataEPIC}.
#'
#' @return A \code{\link{RatioSet}} or \code{\link{GenomicRatioSet}} containing
#' beta value and M value of the methylation data.
#'
#' @examples
#' \dontrun{
#' # Get GM12878 methylation data (450k array)
#' if (!exists("GM12878_450k")) GM12878_450k <- getGM12878("450k")
#' GM12878_450k
#' }
#' 
#' # Get GM12878 methylation data (EPIC array)
#' if (!exists("GM12878_EPIC")) GM12878_EPIC <- getGM12878("EPIC")
#' GM12878_EPIC
#' @export
getGM12878 <- function(arrayType = c("450k", "EPIC"), mapGenome = FALSE) {
  arrayType <- match.arg(arrayType)

  if (toupper(arrayType) == "450K") {
    ### GM12878-450k
    GM12878_450k <- .webDownload(url = remp_options(".default.GM12878.450k.URL"),
                                 tag = "remp.GM12878.450k.file", 
                                 col_types = readr::cols_only(X4 = readr::col_character(),
                                                              X5 = readr::col_integer()))
    GM12878_450k.d <- as.matrix(GM12878_450k$X5 / 1000, ncol = 1)
    colnames(GM12878_450k.d) <- "GM12878"
    rownames(GM12878_450k.d) <- GM12878_450k$X4
    GM12878.RatioSet <- minfi::RatioSet(
      Beta = GM12878_450k.d, M = .toM(GM12878_450k.d),
      annotation = c(
        array = "IlluminaHumanMethylation450k",
        annotation = remp_options(".default.450k.annotation")
      )
    )
    if (mapGenome) {
      GM12878.RatioSet <- minfi::mapToGenome(GM12878.RatioSet)
    }
  }

  if (toupper(arrayType) == "EPIC") {
    ### GM12878-850k
    requireNamespace("minfiDataEPIC", quietly = TRUE)
    GM12878_EPIC.beta <- minfi::getBeta(minfiDataEPIC::MsetEPIC)

    GM12878_EPIC.beta <- rowMeans(GM12878_EPIC.beta, na.rm = TRUE)
    GM12878_EPIC.M <- .toM(GM12878_EPIC.beta)

    GM12878_EPIC.beta.d <- matrix(GM12878_EPIC.beta, ncol = 1)
    GM12878_EPIC.M.d <- matrix(GM12878_EPIC.M, ncol = 1)

    colnames(GM12878_EPIC.beta.d) <- "GM12878"
    colnames(GM12878_EPIC.M.d) <- "GM12878"

    rownames(GM12878_EPIC.beta.d) <- names(GM12878_EPIC.beta)
    rownames(GM12878_EPIC.M.d) <- names(GM12878_EPIC.M)

    GM12878.RatioSet <- minfi::RatioSet(
      Beta = GM12878_EPIC.beta.d,
      M = GM12878_EPIC.M.d,
      annotation = c(
        array = "IlluminaHumanMethylationEPIC",
        annotation = remp_options(".default.epic.annotation")
      )
    )
    if (mapGenome) {
      GM12878.RatioSet <- minfi::mapToGenome(GM12878.RatioSet)
    }
  }

  return(GM12878.RatioSet)
} ## End of getGM12878

## REMProduct methods

#### Accessors
#' @rdname REMProduct-class
setMethod("rempM", signature(object = "REMProduct"), function(object) {
  M <- assays(object)[["rempM"]]
  return(DataFrame(M, check.names = FALSE))
})
# rempM(object)

#' @rdname REMProduct-class
setMethod("rempB", signature(object = "REMProduct"), function(object) {
  B <- assays(object)[["rempB"]]
  return(DataFrame(B, check.names = FALSE))
})
# rempB(object)

#' @rdname REMProduct-class
setMethod("rempQC", signature(object = "REMProduct"), function(object) {
  QC <- assays(object)[["rempQC"]]
  return(DataFrame(QC, check.names = FALSE))
})
# rempQC(object)

#' @rdname REMProduct-class
setMethod("rempImp", signature(object = "REMProduct"), function(object) {
  return(metadata(object)$varImp)
})
# rempImp(object)

#' @rdname REMProduct-class
setMethod("rempAnnot", signature(object = "REMProduct"), function(object) {
  return(metadata(object)$REannotation)
})
# rempAnnot(object)

#' @rdname REMProduct-class
setMethod("rempStats", signature(object = "REMProduct"), function(object) {
  return(list(
    RE_Statistics = metadata(object)$REStats,
    Gene_Statistics = metadata(object)$GeneStats
  ))
})
# rempStats(object)


#### Utilities
#' @rdname REMProduct-class
setMethod(
  "remplot", signature(object = "REMProduct"),
  function(object, type = c("individual", "overall"), ...) {
    type <- match.arg(type)
    default_main <- ""
    default_xlab <- "Methylation value (beta)"
    default_xlim <- c(0, 1)
    dotdotdot <- list(...)
    if (!hasArg("main")) {
      dotdotdot$main <- default_main
    }
    if (!hasArg("xlab")) {
      dotdotdot$xlab <- default_xlab
    }
    if (!hasArg("xlim")) {
      dotdotdot$xlim <- default_xlim
    }

    Bval <- as.matrix(rempB(object))

    if (type == "individual") {
      do.call(plot, c(list(x = density(na.omit(Bval[, 1]))), dotdotdot))
      if (ncol(Bval) >= 2) {
        for (i in seq(2, ncol(Bval)))
        {
          do.call(lines, c(list(x = density(na.omit(Bval[, i]))), dotdotdot))
        }
      }
    }

    if (type == "overall") {
      Bval.mean <- rowMeans(Bval, na.rm = TRUE)
      do.call(plot, c(list(x = density(na.omit(Bval.mean))), dotdotdot))
    }
  }
)

#' @rdname REMProduct-class
setMethod("details", signature(object = "REMProduct"), function(object) {
  REtype <- object@REMPInfo[["REtype"]]
  notAggregated <- TRUE
  if (grepl("aggregated", REtype)) notAggregated <- FALSE

  method <- object@REMPInfo[["predictModel"]]
  notProfiled <- TRUE
  notTrimmed <- TRUE
  notNaive <- TRUE
  if (grepl("Profiled", method)) notProfiled <- FALSE
  if (grepl("trimmed", method)) notTrimmed <- FALSE
  if (grepl("Naive", method)) notNaive <- FALSE

  .showREMPinfo(object)
  cat("\n")

  .showCpGcountbyChr(object)
  cat("\n")

  if (notProfiled & notAggregated & notTrimmed & notNaive) {
    cat("Training information:\n")
    .showTrainingStats(metadata(object)$REStats, object@REMPInfo[["REtype"]], "cat", "  ")
    cat("\n")
  }

  cat("Coverage information:\n")
  .showREStats(metadata(object)$REStats, object@REMPInfo[["REtype"]], "cat", "  ", notAggregated)
  .showGeneStats(metadata(object)$GeneStats, object@REMPInfo[["REtype"]], "cat", "  ")
  cat("\n")

  .showPredictionSummary(object)
  if (object@REMPInfo[["QCModel"]] != "N/A") {
    cat("\n")
    .showQCSummary(object)
  }
})

#' @rdname REMProduct-class
setMethod(
  "decodeAnnot", signature(object = "REMProduct"),
  function(object, type = c("symbol", "entrez"), ncore = 1, BPPARAM = NULL) {
    .isREMProductOrStop(object)

    be <- getBackend(ncore, BPPARAM, TRUE)

    skip <- TRUE
    type <- match.arg(type)
    message("Decoding ", object@REMPInfo[["REtype"]], " annotation to ", type, " ...")

    annot <- metadata(object)$REannotation
    code <- metadata(object)$regionCode
    refgene <- as.data.frame(metadata(object)$refGene)
    containSymbol <- any(grepl("symbol", colnames(mcols(annot))))
    containEntrez <- any(grepl("entrez", colnames(mcols(annot))))

    if (type == "symbol" & !containSymbol) {
      skip <- FALSE
      decodeFun <- .decodeSymbol
    }
    if (type == "entrez" & !containEntrez) {
      skip <- FALSE
      decodeFun <- .decodeEntrez
    }

    if (!skip) {
      BiocParallel::bpstart(be)
      .bploadLibraryQuiet("IRanges", be)
      # message("Pacakge loaded!")
      for (region in remp_options(".default.genomicRegionColNames")) {
        InRegion.original <- code[, region]
        ind.missing <- which(is.na(InRegion.original))
        InRegion <- na.omit(InRegion.original)
        res <- strsplit(InRegion, "[|]")
        res <- lapply(res, as.numeric)
        res <- IRanges::NumericList(res)
        res <- unlist(bplapply(res, decodeFun, refgene = refgene, BPPARAM = be))
        InRegion.original[-ind.missing] <- res
        annot$newregionEncode <- InRegion.original
        annot <- .changeColNames(
          annot, "newregionEncode",
          paste(region, type, sep = ".")
        )
      }
      metadata(object)$REannotation <- annot
      BiocParallel::bpstop(be)
    }
    return(object)
  }
)

#' @rdname REMProduct-class
setMethod(
  "rempTrim", signature(object = "REMProduct"),
  function(object, threshold = 1.7, missingRate = 0.2) {
    method <- object@REMPInfo[["predictModel"]]

    if (grepl("trimmed", method)) {
      previous_thres <- gsub("[\\(\\)]", "", regmatches(method, gregexpr("\\(.*?\\)", method))[[1]])
      if (threshold >= as.numeric(previous_thres)) {
        message("More stringent or equal threshold (", previous_thres, ") has been applied. No changes made.")
        return(object)
      } else {
        object@REMPInfo[["predictModel"]] <- method <- "Random Forest"
      }
    }

    if (!grepl("Random Forest", method)) {
      message("rempTrim() is only applicable to prediction using Random Forest model. No changes made.")
      return(object)
    }

    REtype <- object@REMPInfo[["REtype"]]
    # beta <- as.matrix(rempB(object))
    M <- as.matrix(rempM(object))
    QC <- as.matrix(rempQC(object))
    badInd <- QC > threshold
    QC[badInd] <- NA
    # beta[badInd] <- NA
    M[badInd] <- NA

    removeCpG <- rowMeans(badInd) > missingRate

    RE_annotation <- rempAnnot(object)
    RE_CpG_ILMN <- metadata(object)$RECpG
    regionCode <- metadata(object)$regionCode
    refgene_main <- metadata(object)$refGene

    # beta <- beta[!removeCpG,,drop = FALSE]
    M <- M[!removeCpG, , drop = FALSE]
    QC <- QC[!removeCpG, , drop = FALSE]
    cpgRanges <- rowRanges(object)[!removeCpG, ]

    mcols(RE_annotation) <- cbind(mcols(RE_annotation), regionCode)
    RE_annotation <- subsetByOverlaps(RE_annotation, cpgRanges) # RE_annotation is updated here
    RE_annotation_name <- colnames(mcols(RE_annotation))
    regionCode <- mcols(RE_annotation)[remp_options(".default.genomicRegionColNames")]
    RE_annotation <- RE_annotation[, RE_annotation_name[!RE_annotation_name %in%
      remp_options(".default.genomicRegionColNames")]]

    ## Updated RE coverage
    RE_COVERAGE <- .coverageStats_RE(RE_annotation, regionCode, cpgRanges, RE_CpG_ILMN,
      REtype,
      indent = "    ", FALSE
    )

    # Updated Gene coverage
    GENE_COVERAGE <- .coverageStats_GENE(regionCode, refgene_main,
      REtype,
      indent = "    ", FALSE
    )


    ## Update object
    object_trim <- REMProduct(
      REtype = object@REMPInfo[["REtype"]],
      genome = object@REMPInfo[["genome"]],
      platform = object@REMPInfo[["platform"]],
      win = object@REMPInfo[["win"]],
      predictModel = paste0(object@REMPInfo[["predictModel"]], " - trimmed (", threshold, ")"),
      QCModel = object@REMPInfo[["QCModel"]],
      rempM = M, rempQC = QC,
      cpgRanges = cpgRanges, sampleInfo = colData(object),
      REannotation = RE_annotation,
      RECpG = RE_CpG_ILMN,
      regionCode = regionCode,
      refGene = refgene_main,
      varImp = rempImp(object),
      REStats = RE_COVERAGE, GeneStats = GENE_COVERAGE,
      Seed = metadata(object)$Seed
    )

    return(object_trim)
  }
)

#' @rdname REMProduct-class
setMethod(
  "rempAggregate", signature(object = "REMProduct"),
  function(object, NCpG = 2, ncore = 1, BPPARAM = NULL) {
    be <- getBackend(ncore, BPPARAM, TRUE)

    REtype_string <- object@REMPInfo[["REtype"]]
    REtype <- strsplit(REtype_string, "\\s+\\(")[[1]][1]
    
    method <- object@REMPInfo[["predictModel"]]

    if (REtype_string == paste0(REtype, " (aggregated by mean: min # of CpGs: ", NCpG, ")")) {
      message("The results are already aggregated. No changes made.")
      return(object)
    }

    # If it is random forest prediction, then it has to be trimmed
    # Otherwise it is fine.
    if (grepl("Random Forest", method)) {
      if (!grepl("trimmed", method)) {
        stop("Please first trim the results (using function 'rempTrim()').")
      }
    }

    # The NCpG has to be >=2
    if (NCpG < 2) {
      stop("The minimum number of CpGs must be >=2.")
    }

    M <- rempM(object)
    QC <- rempQC(object)
    RE_annotation <- rempAnnot(object)
    regionCode <- metadata(object)$regionCode
    mcols(RE_annotation) <- cbind(mcols(RE_annotation), regionCode)
    GR <- rowRanges(object)

    ## Remove RE with less than NCpG predicted
    M_df <- data.frame(M,
      Index = as.character(GR$RE.Index),
      check.names = FALSE, stringsAsFactors = FALSE
    )
    RE_annotation <- RE_annotation[base::match(
      unique(M_df$Index),
      runValue(RE_annotation$Index)
    )]
    # identical(runValue(RE_annotation$Index), unique(as.character(M_df$Index)))
    select_RE <- runValue(RE_annotation$Index)
    count_table <- table(M_df$Index)
    count_table <- count_table[select_RE]
    # identical(names(count_table), as.character(RE_annotation$Index))
    count <- as.numeric(count_table)
    RE_annotation <- RE_annotation[count >= NCpG]
    select_RE <- select_RE[count >= NCpG]
    # identical(select_RE, as.character(RE_annotation$Index))
    M_df <- M_df[M_df$Index %in% select_RE, ]

    ## Aggregate predicted M values
    if (ncore > 1) {
      BiocParallel::bpstart(be)
      ITER <- .iblkrow_dup(M_df, chunks = ncore, index_col_name = "Index")
      M_df_MEAN <- BiocParallel::bpiterate(ITER, .aggregateREMP,
        index_col_name = "Index",
        BPPARAM = be,
        REDUCE = rbind,
        reduce.in.order = TRUE
      )
      BiocParallel::bpstop(be)
    } else {
      M_df_MEAN <- aggregate(. ~ Index, M_df, mean, na.rm = TRUE, na.action = na.pass)
      M_df_MEAN <- M_df_MEAN[base::match(select_RE, M_df_MEAN$Index), ]
    }
    # identical(M_df_MEAN_par, M_df_MEAN)
    # identical(runValue(RE_annotation$Index), as.character(M_df_MEAN$Index))
    M <- as.matrix(M_df_MEAN[, -1, drop = FALSE])

    ## Aggregate QC only works for random forest
    if (grepl("Random Forest", method)) {
      QC_df <- data.frame(QC,
        Index = GR$RE.Index, check.names = FALSE,
        stringsAsFactors = FALSE
      )
      QC_df <- QC_df[QC_df$Index %in% select_RE, ]

      if (ncore > 1) {
        BiocParallel::bpstart(be)
        ITER <- .iblkrow_dup(QC_df, chunks = ncore, index_col_name = "Index")
        QC_df_MEAN <- BiocParallel::bpiterate(ITER, .aggregateREMP,
          index_col_name = "Index",
          BPPARAM = be,
          REDUCE = rbind,
          reduce.in.order = TRUE
        )
        BiocParallel::bpstop(be)
      } else {
        QC_df_MEAN <- aggregate(. ~ Index, QC_df, mean, na.rm = TRUE, na.action = na.pass)
        QC_df_MEAN <- QC_df_MEAN[base::match(select_RE, QC_df_MEAN$Index), ]
      }
      # identical(QC_df_MEAN_par, QC_df_MEAN)
      QC <- as.matrix(QC_df_MEAN[, -1, drop = FALSE])
    } else {
      QC <- NULL
    }
    # identical(QC_df_MEAN$Index, M_df_MEAN$Index)
    # identical(as.character(RE_annotation$Index), as.character(M_df_MEAN$Index))

    message(sum(count < NCpG), " ", REtype, " that have less than ", NCpG, " CpGs predicted are not aggretated.")

    ## Wrapping up
    RE_annotation_name <- colnames(mcols(RE_annotation))
    regionCode <- mcols(RE_annotation)[remp_options(".default.genomicRegionColNames")]
    RE_annotation <- RE_annotation[, RE_annotation_name[!RE_annotation_name %in%
      remp_options(".default.genomicRegionColNames")]]
    cpgRanges <- granges(RE_annotation) # cpgRanges is now identical to RE ranges
    cpgRanges$RE.Index <- RE_annotation$Index

    refgene_main <- metadata(object)$refGene
    RE_CpG_ILMN <- metadata(object)$RECpG

    ## Add rownames (RE index)
    RE_index <- runValue(cpgRanges$RE.Index)
    rownames(M) <- RE_index
    if (grepl("Random Forest", method)) {
      rownames(QC) <- RE_index
    }

    ## Updated RE coverage
    RE_COVERAGE <- .coverageStats_RE(RE_annotation, regionCode, cpgRanges, RE_CpG_ILMN,
      REtype,
      indent = "    ", FALSE
    )

    # Updated Gene coverage
    GENE_COVERAGE <- .coverageStats_GENE(regionCode, refgene_main,
      REtype,
      indent = "    ", FALSE
    )

    ## Update object
    object_aggregated <- REMProduct(
      REtype = paste0(REtype, " (aggregated by mean: min # of CpGs: ", NCpG, ")"),
      genome = object@REMPInfo[["genome"]],
      platform = object@REMPInfo[["platform"]],
      win = object@REMPInfo[["win"]],
      predictModel = method,
      QCModel = object@REMPInfo[["QCModel"]],
      rempM = M, rempQC = QC,
      cpgRanges = cpgRanges, sampleInfo = colData(object),
      REannotation = RE_annotation,
      RECpG = RE_CpG_ILMN,
      regionCode = regionCode,
      refGene = refgene_main,
      varImp = rempImp(object),
      REStats = RE_COVERAGE, GeneStats = GENE_COVERAGE,
      Seed = metadata(object)$Seed
    )
    return(object_aggregated)
  }
)

#' @rdname REMProduct-class
setMethod(
  "rempCombine", signature(object1 = "REMProduct", object2 = "REMProduct"),
  function(object1, object2) {
    REtype1 <- object1@REMPInfo[["REtype"]]
    genome1 <- object1@REMPInfo[["genome"]]
    method1 <- object1@REMPInfo[["predictModel"]]
    platform1 <- object1@REMPInfo[["platform"]]
    win1 <- object1@REMPInfo[["win"]]
    seed1 <- metadata(object1)$Seed
    cnames1 <- colnames(object1)

    REtype2 <- object2@REMPInfo[["REtype"]]
    genome2 <- object2@REMPInfo[["genome"]]
    method2 <- object2@REMPInfo[["predictModel"]]
    platform2 <- object2@REMPInfo[["platform"]]
    win2 <- object2@REMPInfo[["win"]]
    seed2 <- metadata(object2)$Seed
    cnames2 <- colnames(object2)

    duplicated_samples <- intersect(cnames1, cnames2)

    if (!identical(REtype1, REtype2)) stop(paste0("You cannot combine ", REtype1, " methylation data with ", REtype2, " methylation data."))
    if (!identical(genome1, genome2)) stop(paste0("You cannot combine genome build ", genome1, " methylation data with genome build ", genome2, " methylation data."))
    if (!identical(method1, method2)) stop(paste0("You cannot combine ", method1, " prediction with ", method2, " prediction."))
    if (!identical(platform1, platform2)) stop(paste0("You cannot combine ", platform1, " array with ", platform2, " array."))
    if (!identical(win1, win2)) stop(paste0("You cannot combine prediction with flanking window size = ", win1, " with size = ", win2, "."))
    if (!identical(seed1, seed2)) stop(paste0("You cannot combine prediction with using seed = ", win1, " with seed = ", win2, "."))
    if (length(duplicated_samples) > 0) stop(paste0("Duplicated samples detected: ", paste0(duplicated_samples, collapse = ", ")))

    M1 <- as.matrix(rempM(object1))
    M2 <- as.matrix(rempM(object2))
    nrow1 <- nrow(M1)
    nrow2 <- nrow(M2)
    if (nrow1 != nrow2) stop("Number of rows does mot match!")

    B1 <- as.matrix(rempB(object1))
    B2 <- as.matrix(rempB(object2))

    QC1 <- as.matrix(rempQC(object1))
    QC2 <- as.matrix(rempQC(object2))

    QCModel <- object1@REMPInfo[["QCModel"]]

    cpgranges1 <- rowRanges(object1)
    sampleinfo1 <- colData(object1)
    reannot1 <- rempAnnot(object1)
    regionCode1 <- metadata(object1)$regionCode
    mcols(reannot1) <- cbind(mcols(reannot1), regionCode1)
    RE_CpG_ILMN1 <- metadata(object1)$RECpG
    varimp1 <- rempImp(object1)

    cpgranges2 <- rowRanges(object2)
    sampleinfo2 <- colData(object2)
    reannot2 <- rempAnnot(object2)
    regionCode2 <- metadata(object2)$regionCode
    mcols(reannot2) <- cbind(mcols(reannot2), regionCode2)
    RE_CpG_ILMN2 <- metadata(object2)$RECpG
    varimp2 <- rempImp(object2)

    ## Combine:
    M <- cbind(M1, M2)
    B <- cbind(B1, B2)
    QC <- cbind(QC1, QC2)
    if (nrow(QC) == 0) QC <- NULL

    cpgRanges <- cpgranges1
    sampleinfo <- rbind(sampleinfo1, sampleinfo2)
    varimp <- DataFrame(varimp1, varimp2, check.names = FALSE)

    RE_annotation <- unique(c(reannot1, reannot2))
    RE_CpG_ILMN <- unique(c(RE_CpG_ILMN1, RE_CpG_ILMN2))

    ## Wrapping up
    RE_annotation_name <- colnames(mcols(RE_annotation))
    regionCode <- mcols(RE_annotation)[remp_options(".default.genomicRegionColNames")]
    RE_annotation <- RE_annotation[, RE_annotation_name[!RE_annotation_name %in%
      remp_options(".default.genomicRegionColNames")]]

    refgene_main <- metadata(object1)$refGene

    ## Updated RE coverage
    RE_COVERAGE <- .coverageStats_RE(RE_annotation, regionCode, cpgRanges, RE_CpG_ILMN,
      REtype1,
      indent = "    ", FALSE
    )

    # Updated Gene coverage
    GENE_COVERAGE <- .coverageStats_GENE(regionCode, refgene_main,
      REtype1,
      indent = "    ", FALSE
    )


    ## Update object
    object_combined <- REMProduct(
      REtype = REtype1,
      genome = genome1,
      platform = platform1,
      win = win1,
      predictModel = method1,
      QCModel = object1@REMPInfo[["QCModel"]],
      rempM = M, rempB = B, rempQC = QC,
      cpgRanges = cpgRanges, sampleInfo = sampleinfo,
      REannotation = RE_annotation,
      RECpG = RE_CpG_ILMN,
      regionCode = regionCode,
      refGene = refgene_main,
      varImp = varimp,
      REStats = RE_COVERAGE, GeneStats = GENE_COVERAGE,
      Seed = seed1
    )
    return(object_combined)
  }
)

#' @title Get BiocParallel back-end
#'
#' @description
#' \code{getBackend} is used to obtain \code{BiocParallel} Back-end to allow
#' parallel computing.
#'
#' @param ncore Number of cores used for parallel computing. By default max number
#' of cores available in the machine will be utilized. If \code{ncore = 1}, no
#' parallel computing is allowed.
#' @param BPPARAM An optional \code{BiocParallelParam} instance determining the
#' parallel back-end to be used during evaluation. If not specified, default
#' back-end in the machine will be used.
#' @param verbose Logical parameter. Should the function be verbose?
#'
#' @return A \code{\link{BiocParallel}} object that can be used for parallel
#' computing.
#'
#' @examples
#' # Non-parallel mode
#' be <- getBackend(ncore = 1, verbose = TRUE)
#' be
#' 
#' # parallel mode (2 workers)
#' be <- getBackend(ncore = 2, verbose = TRUE)
#' be
#' 
#' @export
getBackend <- function(ncore, 
                       BPPARAM = NULL, 
                       verbose = FALSE) {
  maxCore <- parallel::detectCores()
  if (ncore > maxCore) {
    if (!is.na(maxCore)) {
      ncore <- maxCore
    } else {
      
    }
  }

  if (ncore > 16) message("Note: Requesting more than 16 cores may not benifit much.")

  if (ncore == 1) {
    backend <- BiocParallel::SerialParam()
    if (verbose) {
      message("You have successfully set non-parallel mode (single worker).")
    }
  } else {
    if (is.null(BPPARAM)) bpparam <- BiocParallel::bpparam() else bpparam <- BPPARAM
    backend_class <- attributes(bpparam)$class
    if (!backend_class %in% names(BiocParallel::registered())) {
      stop(
        "Back-end '", backend_class, "' is not supported on ",
        Sys.info()["sysname"], "!"
      )
    } else {
      if ("MulticoreParam" %in% names(BiocParallel::registered())) {
        backend <- BiocParallel::MulticoreParam(workers = ncore)
      } else {
        backend <- BiocParallel::SnowParam(workers = ncore)
      }
      if (verbose) {
        message(
          "You have successfully set parallel mode with ",
          BiocParallel::bpnworkers(backend), " workers (", backend_class, ")."
        )
      }
    }
  }
  return(backend)
}




#' @title Get RE database from RepeatMasker
#'
#' @description
#' \code{fetchRMSK} is used to obtain specified RE database from RepeatMasker Database
#' provided by AnnotationHub.
#' 
#' @param REtype Type of RE. Currently \code{"Alu"}, \code{"L1"}, and \code{"ERV"} are supported.
#' @param annotation.source Character parameter. Specify the source of annotation databases, including
#' the RefSeq Gene annotation database and RepeatMasker annotation database. If \code{"AH"}, the database 
#' will be obtained from the AnnotationHub package. If \code{"UCSC"}, the database will be downloaded 
#' from the UCSC website http://hgdownload.cse.ucsc.edu/goldenpath. The corresponding build (\code{"hg19"} or 
#' \code{"hg38"}) will be specified in the parameter \code{genome}.
#' @param genome Character parameter. Specify the build of human genome. Can be either \code{"hg19"} or
#' \code{"hg38"}.
#' @param verbose Logical parameter. Should the function be verbose?
#'
#' @return A \code{\link{GRanges}} object containing RE database. 'repName' column
#' indicates the RE name; 'swScore' column indicates the SW score; 'Index' is an
#' internal index for RE to facilitate data referral, which is meaningless for external use.
#'
#' @examples
#' L1 <- fetchRMSK(REtype = "L1", 
#'                 annotation.source = "AH",
#'                 genome = "hg19", 
#'                 verbose = TRUE)
#' L1
#' @export
fetchRMSK <- function(REtype = c("Alu", "L1", "ERV"), 
                      annotation.source = c("AH", "UCSC"), 
                      genome = c("hg19", "hg38"), 
                      verbose = FALSE) {
  REtype = match.arg(REtype)
  annotation.source = match.arg(annotation.source)
  genome = match.arg(genome)
  
  if(genome == "hg19") {
    if(annotation.source == "UCSC") {
      message("UCSC RepeatMasker annotation database is not recommended. Switching to AnnotationHub instead...")
      annotation.source = "AH"
    }
    if(annotation.source == "AH") {
      if (verbose) message(
        "Loading ", REtype, " annotation data from RepeatMasker (hg19) database from AnnotationHub: ",
        remp_options(".default.AH.repeatmasker.hg19")
      )
      ah <- .initiateAH()
      if(is.null(ah)) {
        warning("AnnotationHub is currently not accessible, Please try again later.")
      } else {
        rmsk <- suppressMessages(ah[[remp_options(".default.AH.repeatmasker.hg19")]])
      }
    }
  }
  
  if(genome == "hg38") {
    if(annotation.source == "UCSC") {
      message("UCSC RepeatMasker annotation database is not recommended. Switching to AnnotationHub instead...")
      annotation.source = "AH"
    }
    if(annotation.source == "AH") {
      if (verbose) message(
        "Loading ", REtype, " annotation data from RepeatMasker (hg38) database from AnnotationHub: ",
        remp_options(".default.AH.repeatmasker.hg38")
      )
      ah <- .initiateAH()
      if(is.null(ah)) {
        warning("AnnotationHub is currently not accessible, Please try again later.")
      } else {
        rmsk <- suppressMessages(ah[[remp_options(".default.AH.repeatmasker.hg38")]])
      }
    }
  }

  if (REtype == "Alu") {
    REFamily_grep <- remp_options(".default.AluFamily.grep")
  }
  if (REtype == "L1") {
    REFamily_grep <- remp_options(".default.L1Family.grep")
  }
  if (REtype == "ERV") {
    REFamily_grep <- remp_options(".default.ERVFamily.grep")
  }
  
  RE <- rmsk[grep(REFamily_grep, rmsk$repFamily)]
  RE <- RE[as.character(seqnames(RE)) %in% remp_options(".default.chr")] # chr1 - 22, chrX, chrY
  seqlevels(RE) <- remp_options(".default.chr") # remove uncommon chr

  RE <- sort(RE) # sort so that AH and UCSC are identical
  
  mcols(RE)$Index <- Rle(paste(REtype, formatC(seq_len(length(RE)),
    width = 7,
    flag = "0"
  ), sep = "_")) # Add internal index, meaningless for external database
  
  if(verbose) message("Obtained ", length(RE), " ", REtype, " (", genome, ").")
  
  return(RE)
}




#' @title Get RefSeq gene database
#'
#' @description
#' \code{fetchRefSeqGene} is used to obtain refSeq gene database provided by AnnotationHub (hg19) 
#' or UCSC web database (hg19/hg38).
#'
#' @param annotation.source Character parameter. Specify the source of annotation databases, including
#' the RefSeq Gene annotation database and RepeatMasker annotation database. If \code{"AH"}, the database 
#' will be obtained from the AnnotationHub package. If \code{"UCSC"}, the database will be downloaded 
#' from the UCSC website http://hgdownload.cse.ucsc.edu/goldenpath. The corresponding build (\code{"hg19"} or 
#' \code{"hg38"}) will be specified in the parameter \code{genome}.
#' @param genome Character parameter. Specify the build of human genome. Can be either \code{"hg19"} or 
#' \code{"hg38"}. Note that if \code{annotation.source == "AH"}, only hg19 database is available.
#' @param mainOnly Logical parameter. See details.
#' @param verbose Logical parameter. Should the function be verbose?
#'
#' @details
#' When \code{mainOnly = FALSE}, only the transcript location information will be returned,
#' Otherwise, a \code{\link{GRangesList}} object containing gene regions
#' information will be added. Gene regions include: 2000 base pair upstream of the transcript
#' start site (\code{$tss})), 5'UTR (\code{$fiveUTR})), coding sequence (\code{$cds})),
#' exon (\code{$exon})), and 3'UTR (\code{$threeUTR})). The \code{index} column is an internal
#' index that is used to facilitate data referral, which is meaningless for external use.
#'
#' @return A single \code{\link{GRanges}} (for main refgene data) object or a list incorporating
#' both \code{\link{GRanges}} object (for main refgene data) and \code{\link{GRangesList}} object
#' (for gene regions data).
#'
#' @examples
#' if (!exists("refgene.hg19")) 
#' refgene.hg19 <- fetchRefSeqGene(annotation.source = "AH", 
#'                                 genome = "hg19", 
#'                                 verbose = TRUE)
#' refgene.hg19
#' 
#' @export
fetchRefSeqGene <- function(annotation.source = c("AH", "UCSC"), 
                            genome = c("hg19", "hg38"), 
                            mainOnly = FALSE, 
                            verbose = FALSE) {
  annotation.source = match.arg(annotation.source)
  genome = match.arg(genome)
  
  if(genome == "hg19") {
    if(annotation.source == "AH") {
      if(verbose) message(
        "Loading static refSeq gene (hg19) database from AnnotationHub: ",
        remp_options(".default.AH.refgene.hg19"))
      ah <- .initiateAH()
      if(is.null(ah)) {
        message("AnnotationHub is currently not accessible, Switching to 'UCSC' source instead...")
        annotation.source <- "UCSC"
      } else refgene <- suppressMessages(ah[[remp_options(".default.AH.refgene.hg19")]])
    }
    if(annotation.source == "UCSC") {
      if(verbose) message("Loading UCSC refSeq gene (hg19) database from ", remp_options(".default.refGene.hg19.URL"))
      refgene <- .UCSCrefGeneDownload(url = remp_options(".default.refGene.hg19.URL"), 
                                      tag = "refGene.hg19",
                                      verbose)
    }
  }
  
  if(genome == "hg38") {
    if(annotation.source == "AH") {
      message("RefSeq Gene database in hg38 build is not available in AnnotationHub. Switching to 'UCSC' source instead...")
      annotation.source <- "UCSC"
    }
    if(annotation.source == "UCSC") {
      if(verbose) message("Loading UCSC refSeq gene (hg38) database from ", remp_options(".default.refGene.hg38.URL"))
      refgene <- .UCSCrefGeneDownload(url = remp_options(".default.refGene.hg38.URL"), 
                                      tag = "refGene.hg38",
                                      verbose)
    }
  }
  
  refgene <- refgene[as.character(seqnames(refgene)) %in%
    paste0("chr", c(seq_len(22), "X", "Y"))] ## chr1 - 22, chrX, chrY
  refgene$type <- substring(refgene$name, 1, 2) ## Protein coding gene (NM) or noncoding RNA gene (NR)
  refgene$index <- as.character(seq_len(length(refgene))) ## This is internal index, meaningless for external database

  ## Master database with detailed gene information
  refgene_main <- GRanges(
    seqnames = seqnames(refgene), ranges = ranges(refgene),
    strand = strand(refgene), name = mcols(refgene)$name, type = mcols(refgene)$type,
    index = mcols(refgene)$index
  )

  refseq2EG <- unlist(BiocGenerics::mget(refgene_main$name, org.Hs.eg.db::org.Hs.egREFSEQ2EG,
    ifnotfound = NA
  ))
  if (verbose) {
    message("Note: There are ", sum(is.na(refseq2EG)), " refseq ID cannot be mapped to Entrez ID, which could be obsolete.")
  }
  # obsoleteID <- refgene_main$name[is.na(refseq2EG)]
  refseq2EG <- na.omit(refseq2EG)

  EG2Symbol <- unlist(BiocGenerics::mget(refseq2EG, org.Hs.eg.db::org.Hs.egSYMBOL)) ## If EG -> symbol not found, this is a fatel error, should stop here

  refseq2EG_Symbol.DF <- DataFrame(
    refseqID = names(refseq2EG), EntrezGeneID = refseq2EG,
    GeneSymbol = EG2Symbol
  )

  refgene_main$EntrezGene <- refseq2EG_Symbol.DF$EntrezGeneID[base::match(
    refgene_main$name,
    refseq2EG_Symbol.DF$refseqID
  )]
  refgene_main$GeneSymbol <- refseq2EG_Symbol.DF$GeneSymbol[base::match(
    refgene_main$name,
    refseq2EG_Symbol.DF$refseqID
  )]

  if (mainOnly) {
    return(refgene_main)
  }

  ## TSS2000
  refgene_tss <- promoters(refgene,
    upstream = remp_options(".default.TSS.upstream"),
    downstream = remp_options(".default.TSS.downstream")
  )[, "index"]

  ## CDS
  refgene_cds <- GRanges(
    seqnames = seqnames(refgene), ranges = refgene$thick,
    strand = strand(refgene), index = mcols(refgene)$index
  )

  ## EXON
  block_shift <- shift(refgene$blocks, start(refgene) - 1) # minus one
  refgene_exon <- makeGRangesListFromFeatureFragments(
    seqnames = seqnames(refgene_main),
    fragmentStarts = start(block_shift), fragmentEnds = end(block_shift),
    strand = strand(refgene_main)
  )
  names(refgene_exon) <- refgene_main$index
  refgene_exon <- unlist(refgene_exon)
  refgene_exon$index <- as.integer(names(refgene_exon))
  names(refgene_exon) <- NULL

  ## UTR (5'UTR/3'UTR concept only applied for protein-coding gene)
  refgene_NM <- refgene[refgene$type == "NM", c("thick", "index")]
  refgene_f <- refgene_NM[strand(refgene_NM) == "+"]
  refgene_r <- refgene_NM[strand(refgene_NM) == "-"]

  refgene_5UTR_f <- GRanges(
    seqnames = seqnames(refgene_f),
    ranges = IRanges(
      start = start(refgene_f),
      end = start(refgene_f$thick)
    ),
    strand = strand(refgene_f),
    index = refgene_f$index
  )

  refgene_5UTR_r <- GRanges(
    seqnames = seqnames(refgene_r),
    ranges = IRanges(
      start = end(refgene_r$thick),
      end = end(refgene_r)
    ),
    strand = strand(refgene_r),
    index = refgene_r$index
  )

  refgene_3UTR_f <- GRanges(
    seqnames = seqnames(refgene_f),
    ranges = IRanges(
      start = end(refgene_f$thick),
      end = end(refgene_f)
    ),
    strand = strand(refgene_f),
    index = refgene_f$index
  )

  refgene_3UTR_r <- GRanges(
    seqnames = seqnames(refgene_r),
    ranges = IRanges(
      start = start(refgene_r),
      end = start(refgene_r$thick)
    ),
    strand = strand(refgene_r),
    index = refgene_r$index
  )

  refgene_5UTR <- c(refgene_5UTR_f, refgene_5UTR_r)
  refgene_3UTR <- c(refgene_3UTR_f, refgene_3UTR_r)

  return(list(
    main = refgene_main,
    regions = GRangesList(
      tss = refgene_tss,
      cds = refgene_cds,
      exon = refgene_exon,
      fiveUTR = refgene_5UTR,
      threeUTR = refgene_3UTR
    )
  ))
}



#' @title Find RE-CpG genomic location given RE ranges information
#'
#' @description
#' \code{findRECpG} is used to obtain RE-CpG genomic location data.
#'
#' @param RE A \code{\link{GRanges}} object of RE genomic location database. This
#' can be obtained by \code{\link{fetchRMSK}}.
#' @param REtype Type of RE. Currently \code{"Alu"}, \code{"L1"}, and \code{"ERV"} are supported.
#' @param genome Character parameter. Specify the build of human genome. Can be either \code{"hg19"} or 
#' \code{"hg38"}. User should make sure the genome build of \code{RE} is consistent with this parameter.
#' @param be A \code{\link{BiocParallel}} object containing back-end information that is
#' ready for parallel computing. This can be obtained by \code{\link{getBackend}}. If not specified,
#' non-parallel mode is used.
#' @param verbose logical parameter. Should the function be verbose?
#'
#' @details
#' CpG site is defined as 5'-C-p-G-3'. It is reasonable to assume that the methylation
#' status across all CpG/CpG dyads are concordant. Maintenance methyltransferase exhibits
#' a preference for hemimethylated CpG/CpG dyads (methylated on one strand only).
#' As a result, methyaltion status of CpG sites in both forward and reverse strands are usually consistent.
#' Therefore, to accommodate the cytosine loci in both strands, the returned genomic
#' ranges cover the 'CG' sequence with width of 2. The 'strand' information indicates the strand of the RE.
#' Locating CpG sites in RE sequences can be computation intensive. It is recommanded to get more than
#' one work in the backend for a faster running speed.
#'
#' @return A \code{\link{GRanges}} object containing identified RE-CpG genomic
#' location data.
#'
#' @examples
#' data(Alu.hg19.demo)
#' RE.CpG <- findRECpG(RE = Alu.hg19.demo, 
#'                     REtype = "Alu", 
#'                     genome = "hg19", 
#'                     verbose = TRUE)
#' RE.CpG
#' 
#' @export
findRECpG <- function(RE, 
                      REtype = c("Alu", "L1", "ERV"), 
                      genome = c("hg19", "hg38"), 
                      be = NULL, 
                      verbose = FALSE) {
  if (is.null(be)) be <- getBackend(1, verbose)
  REtype <- match.arg(REtype)
  genome <- match.arg(genome)
  
  ## Flank RE by 1 base pair to cover the cases where CG is located in
  ## RE's head or tail
  RE <- .twoWayFlank(RE, 1)

  if (verbose) {
    message(
      "    Getting sequence of the total ", length(RE),
      " ", REtype, " ..."
    )
  }
  
  if(genome == "hg19") {
    if (requireNamespace("BSgenome.Hsapiens.UCSC.hg19", quietly = TRUE)) {
      if(isNamespaceLoaded("BSgenome.Hsapiens.UCSC.hg38")) unloadNamespace("BSgenome.Hsapiens.UCSC.hg38")
      if(!isNamespaceLoaded("BSgenome.Hsapiens.UCSC.hg19")) attachNamespace("BSgenome.Hsapiens.UCSC.hg19")
      if(verbose) message("    Using UCSC Human Reference Genome: hg19")
      hs <- BSgenome.Hsapiens.UCSC.hg19::Hsapiens
    } else stop("Please install missing package: BSgenome.Hsapiens.UCSC.hg19") 
  }
  
  if(genome == "hg38") {
    if (requireNamespace("BSgenome.Hsapiens.UCSC.hg38", quietly = TRUE)) {
      if(isNamespaceLoaded("BSgenome.Hsapiens.UCSC.hg19")) unloadNamespace("BSgenome.Hsapiens.UCSC.hg19")
      if(!isNamespaceLoaded("BSgenome.Hsapiens.UCSC.hg38")) attachNamespace("BSgenome.Hsapiens.UCSC.hg38")
      if(verbose) message("    Using UCSC Human Reference Genome: hg38")
      hs <- BSgenome.Hsapiens.UCSC.hg38::Hsapiens
    } else stop("Please install missing package: BSgenome.Hsapiens.UCSC.hg38") 
  }
  
  SEQ.RE <- BSgenome::getSeq(hs, 
                             GRanges(seqnames(RE), 
                                     IRanges(start = start(RE), end = end(RE)))) # ignore strand

  if (verbose) {
    message("    Identifying CpG sites in ", REtype, " sequence ...")
  }

  if (BiocParallel::bpnworkers(be) > 1) {
    BiocParallel::bpstart(be)
    .bploadLibraryQuiet("Biostrings", be)
    RE.CpG <- bpvec(SEQ.RE, .vRECpGPos, CpG = DNAString("CG"), BPPARAM = be)
    BiocParallel::bpstop(be)
  } else {
    RE.CpG <- .vRECpGPos(SEQ.RE, CpG = DNAString("CG"))
  }

  noCpGind <- which(vapply(RE.CpG, length, integer(1)) == 0) ## Remove sequence without CpG sites

  if (length(noCpGind) == 0) {
    if (verbose) {
      message("    All ", REtype, " sequences contain CpGs.")
    }
  } else {
    if (verbose) {
      message("    There are ", length(noCpGind), " ", REtype, " sequences contain no CpG site.")
    }
    RE <- RE[-noCpGind]
    RE.CpG <- RE.CpG[-noCpGind]
  }

  ## Extend 'C' to 'CG' AND Shift to the real genomic location
  RE.CpG.IRList <- IRangesList(start = RE.CpG - 1, end = RE.CpG)
  block_shift <- shift(RE.CpG.IRList, start(RE))

  RE.CpG <- makeGRangesListFromFeatureFragments(
    seqnames = seqnames(RE),
    fragmentStarts = start(block_shift),
    fragmentEnds = end(block_shift),
    strand = strand(RE)
  )
  elementLen <- elementNROWS(RE.CpG)

  RE.CpG <- unlist(RE.CpG)
  RE.CpG$Index <- Rle(runValue(RE$Index), elementLen)
  RE.CpG <- RE.CpG[order(RE.CpG$Index)]

  if (verbose) {
    message(
      "    Identified ", length(RE.CpG), " CpG/CpG dyads in ",
      nrun(RE.CpG$Index), " ", REtype
    )
  }

  return(RE.CpG)
}



#' @title Annotate genomic ranges data with gene region information.
#'
#' @description
#' \code{GRannot} is used to annotate a \code{\link{GRanges}} dataset with gene region
#' information using refseq gene database
#'
#' @param object.GR An \code{\link{GRanges}} object of a genomic location database.
#' @param refgene A complete refGene annotation database returned by
#' \code{\link{fetchRefSeqGene}} (with parameter \code{mainOnly = FALSE}).
#' @param symbol Logical parameter. Should the annotation return gene symbol?
#' @param verbose Logical parameter. Should the function be verbose?
#'
#' @details
#' The annotated gene region information includes: protein coding gene (InNM),
#' noncoding RNA gene (InNR), 2000 base pair upstream of the transcript start site (InTSS),
#' 5'UTR (In5UTR), coding sequence (InCDS), exon (InExon), and 3'UTR (In3UTR). The intergenic
#' and intron regions can then be represented by the combination of these region data.
#' The number shown in these columns represent the row number or 'index' column in the
#' main refgene database obtained by \code{\link{fetchRefSeqGene}}.
#'
#' @return A \code{\link{GRanges}} or a \code{\link{GRangesList}} object containing refSeq
#' Gene database.
#'
#' @examples
#' data(Alu.hg19.demo)
#' if (!exists("refgene.hg19")) 
#'   refgene.hg19 <- fetchRefSeqGene(annotation.source = "AH", 
#'                                   genome = "hg19",
#'                                   verbose = TRUE)
#' Alu.hg19.demo.refGene <- GRannot(Alu.hg19.demo, refgene.hg19, verbose = TRUE)
#' Alu.hg19.demo.refGene
#' 
#' @export
GRannot <- function(object.GR, 
                    refgene, 
                    symbol = FALSE, 
                    verbose = FALSE) {

  ## Check if the object.GR contains a metadata column called "Index"
  if (!"Index" %in% colnames(mcols(object.GR))) {
    object.GR$Index <- as.character(seq_len(length(object.GR)))
  }

  if (any(duplicated(as.character(object.GR$Index)))) {
    stop("The 'Index' column provided in the GRanges object must be unique.")
  }

  main <- refgene$main

  ####### InNM
  NM.GR <- main[main$type == "NM"]
  NM.GR <- .oneWayFlank(NM.GR, 2000, start = TRUE)
  object.GR <- .addannot(
    object.GR,
    NM.GR,
    main,
    symbol,
    "InNM",
    verbose
  )

  ####### InNR
  NR.GR <- main[main$type == "NR"]
  NR.GR <- .oneWayFlank(NR.GR, 2000, start = TRUE)
  object.GR <- .addannot(
    object.GR,
    NR.GR,
    main,
    symbol,
    "InNR",
    verbose
  )

  ####### InTSS2000
  object.GR <- .addannot(
    object.GR,
    refgene$regions$tss,
    main,
    symbol,
    "InTSS",
    verbose
  )

  ####### In5UTR
  object.GR <- .addannot(
    object.GR,
    refgene$regions$fiveUTR,
    main,
    symbol,
    "In5UTR",
    verbose
  )

  ####### InCDS
  object.GR <- .addannot(
    object.GR,
    refgene$regions$cds,
    main,
    symbol,
    "InCDS",
    verbose
  )

  ####### InExon
  object.GR <- .addannot(
    object.GR,
    refgene$regions$exon,
    main,
    symbol,
    "InExon",
    verbose
  )

  ####### In3UTR
  object.GR <- .addannot(
    object.GR,
    refgene$regions$threeUTR,
    main,
    symbol,
    "In3UTR",
    verbose
  )

  return(object.GR)
}



## --------------------------------------
## Internal functions

# Used by findRECpG (vetorized function)
.vRECpGPos <- function(seq, CpG) {
  start(vmatchPattern(CpG, seq))
}

# Used by addannot
.clps <- function(x) paste0(x, collapse = "|")

# Used by GRannot
.addannot <- function(object.GR, annot.GR, main, symbol, annotName, verbose) {
  if (verbose) {
    message("    ", annotName, " ... ", appendLF = FALSE)
  }

  object_annot_Hit <- findOverlaps(object.GR, annot.GR, ignore.strand = TRUE)
  object_raw <- DataFrame(InRegion = annot.GR[subjectHits(object_annot_Hit)]$index)
  InRegion <- rep(NA, length(object.GR))

  if (nrow(object_raw) > 0) {
    if (symbol) {
      object_raw$InRegion <- main$GeneSymbol[as.integer(object_raw$InRegion)]
    }

    object_agg <- aggregate(
      object_raw,
      list(Index = as.character(object.GR$Index[queryHits(object_annot_Hit)])),
      .clps
    )
    InRegion[base::match(object_agg$Index, 
                         runValue(object.GR$Index))] <- object_agg$InRegion
  }

  object.GR$newRegion <- InRegion
  mcols(object.GR) <- .changeColNames(
    mcols(object.GR),
    "newRegion",
    annotName
  )

  if (verbose) {
    message("  ", nrow(object_agg), " found!")
  }

  return(object.GR)
}

# Used by fetchRefSeqGene
.UCSCrefGeneDownload <- function(url, tag, verbose)
{
  refgene_raw <- .webDownload(url = url, 
                              tag = tag, 
                              col_types = readr::cols_only(X2 = readr::col_character(),
                                                           X3 = readr::col_character(),
                                                           X4 = readr::col_character(),
                                                           X5 = readr::col_integer(),
                                                           X6 = readr::col_integer(),
                                                           X7 = readr::col_integer(),
                                                           X8 = readr::col_integer(),
                                                           X10 = readr::col_character(),
                                                           X11 = readr::col_character()), 
                              verbose = verbose)
  refgene <- GRanges(seqnames = refgene_raw$X3, IRanges(start = refgene_raw$X5 + 1, end = refgene_raw$X6),
                     strand = refgene_raw$X4, name = vapply(strsplit(refgene_raw$X2, "\\."), function(x) x[1], character(1)), 
                     thick = IRanges(start = refgene_raw$X7 + 1, end = refgene_raw$X8),
                     blocks = IRangesList(start = mapply(function(x, y) x + 1 - y,  
                                                         lapply(strsplit(refgene_raw$X10, ","), as.integer), 
                                                         refgene_raw$X5), 
                                          end = mapply(function(x, y) x - y,  
                                                       lapply(strsplit(refgene_raw$X11, ","), as.integer), 
                                                       refgene_raw$X5)))
  return(refgene)
}

## Internal utility functions:

.timeTrace <- function(oldT, indent = "  ") {
  currenT <- Sys.time()
  return(list(
    t = currenT,
    t_text = paste0(indent, "(", round(as.numeric(currenT - oldT, units = "secs"), 0), " sec.)")
  ))
}

.forwardSlashPath <- function(path) {
  gsub("\\\\", "/", path)
}

# Assign the pacakge to workers and suppress the loading messsage!
.bploadLibraryQuiet <- function(libraryList, be) {
  noreturn <- BiocParallel::bplapply(seq_len(BiocParallel::bpnworkers(be)), function(i, libraryList) {
    suppressPackageStartupMessages(lapply(libraryList, library, character.only = TRUE))
  }, BPPARAM = be, libraryList = libraryList)
}

.iblkrow <- function(M, chunks) {
  i <- 1
  uni_row <- nrow(M)
  it <- idiv(uni_row, chunks = chunks)

  nextEl <- function() {
    if (i == uni_row + 1) {
      return(NULL)
    }
    n <- nextElem(it)
    r <- seq(i, length = n)
    i <<- i + n
    M[r, , drop = FALSE]
  }
  nextEl
}

.iblkrow_dup <- function(M, chunks, index_col_name) {
  i <- 1
  ind <- which(colnames(M) == index_col_name)
  index <- M[, ind]
  unique_index <- unique(index)
  unique_length <- length(unique_index)

  it <- idiv(unique_length, chunks = chunks)

  nextEl <- function() {
    if (i == unique_length + 1) {
      return(NULL)
    }
    n <- nextElem(it)
    r <- seq(i, length = n)
    i <<- i + n
    M[index %in% unique_index[r], , drop = FALSE]
  }
  nextEl
}

.aggregateREMP <- function(dat, index_col_name) {
  eval(parse(text = paste0("aggregate(.~", index_col_name, ", dat, mean, na.rm = TRUE, na.action = na.pass)")))
}

.guessArrayType <- function(methyDat) {
  nProbes <- nrow(methyDat)
  if (nProbes <= remp_options(".default.27k.total.probes")) {
    return("27k")
  }
  if (nProbes > remp_options(".default.27k.total.probes") &
    nProbes <= remp_options(".default.450k.total.probes")) {
    return("450k")
  }
  if (nProbes > remp_options(".default.450k.total.probes") &
    nProbes <= remp_options(".default.epic.total.probes")) {
    return("EPIC")
  }
  if (nProbes > remp_options(".default.epic.total.probes")) {
    return("Sequencing")
  }
}

.guessDataType <- function(methyDat) {
  methyDat_sample <- methyDat[sample(seq_len(nrow(methyDat)), min(5000, nrow(methyDat))),
    sample(seq_len(ncol(methyDat)), min(5, ncol(methyDat))),
    drop = FALSE
  ]
  rng <- range(methyDat_sample, na.rm = TRUE)
  if (rng[1] >= 0 & rng[2] <= 1 |
    sum(methyDat_sample < 0 | methyDat_sample > 1, na.rm = TRUE) <
      nrow(methyDat_sample) * ncol(methyDat_sample) * 0.01) {
    return("beta")
  } else if (rng[1] < 0) {
    return("M")
  } else if (rng[1] == 0 & rng[2] == 100) {
    return("percentage")
  } else {
    return("unknown")
  }
}

.twoWayFlank <- function(object.GR, width) {
  .isGROrStop(object.GR)
  start(object.GR) <- start(object.GR) - width
  end(object.GR) <- end(object.GR) + width
  return(object.GR)
}

.oneWayFlank <- function(object.GR, width, start = TRUE) {
  .isGROrStop(object.GR)
  if (any(runValue(strand(object.GR)) == "*")) {
    stop("Strand information must not be missing.")
  }
  forward_ind <- strand(object.GR) == "+"
  reverse_ind <- strand(object.GR) == "-"

  if (start) {
    start(object.GR[forward_ind]) <- start(object.GR[forward_ind]) - width
    end(object.GR[reverse_ind]) <- end(object.GR[reverse_ind]) + width
  } else {
    end(object.GR[forward_ind]) <- end(object.GR[forward_ind]) + width
    start(object.GR[reverse_ind]) <- start(object.GR[reverse_ind]) - width
  }
  return(object.GR)
}

.changeColNames <- function(DForGR, oldnames, newnames) {
  if ("DataFrame" %in% is(DForGR)) {
    cnames <- colnames(DForGR)
    colnames(DForGR)[cnames %in% oldnames] <- newnames
    return(DForGR)
  }
  if ("GRanges" %in% is(DForGR)) {
    cnames <- colnames(mcols(DForGR))
    colnames(mcols(DForGR))[cnames %in% oldnames] <- newnames
    return(DForGR)
  }
}

.webDownload <- function(url, tag, col_types = NULL, verbose = FALSE)
{
  tempfilenames <- list.files(tempdir())
  which.tag.file <- grep(tag, tempfilenames)
  if (length(which.tag.file) == 0) {
    temp <- tempfile(pattern = tag) # Make an identifiable ext to avoid repeated download
    if(verbose) message("    Downloading file...")
    download.file(url, temp, quiet = TRUE)
  } else {
    if(verbose) message("    File existed!")
    temp <- file.path(tempdir(), tempfilenames[which.tag.file])
  }
  if(verbose) message("    Unzipping and loading file...")
  downloadedFile <- readr::read_delim(gzfile(temp), col_names = FALSE, delim = "\t", col_types = col_types)
  return(downloadedFile)
}

.liftOver_Hg19toHg38 <- function(ILMN.GR, m = NULL, verbose = FALSE) {
  if(verbose) message(m)
  ah <- .initiateAH()
  if(is.null(ah)) stop("AnnotationHub is currently not accessible. Please try again later.")
  ch <- suppressMessages(ah[[remp_options(".default.AH.hg19ToHg38.over.chain")]])
  ILMN.GR <- rtracklayer::liftOver(ILMN.GR, ch)
  ILMN.GR <- unlist(ILMN.GR)
  return(ILMN.GR)
}

.liftOver_Hg38toHg19 <- function(ILMN.GR, m = NULL, verbose = FALSE) {
  if(verbose) message(m)
  ah <- .initiateAH()
  if(is.null(ah)) stop("AnnotationHub is currently not accessible. Please try again later.")
  ch <- suppressMessages(ah[[remp_options(".default.AH.hg38ToHg19.over.chain")]])
  ILMN.GR <- rtracklayer::liftOver(ILMN.GR, ch)
  ILMN.GR <- unlist(ILMN.GR)
  return(ILMN.GR)
}

.initiateAH <- function()
{
  permission <- file.access(AnnotationHub::getAnnotationHubOption("CACHE"), 2)
  
  if (permission != 0) {
      message(
        AnnotationHub::getAnnotationHubOption("CACHE"),
        " is not writable, using temporal directory ",
        .forwardSlashPath(tempdir()), " instead."
      )
    AnnotationHub::setAnnotationHubOption("CACHE", file.path(tempdir(), ".AnnotationHub"))
  }
  ah <- tryCatch({
    suppressMessages(AnnotationHub::AnnotationHub())
    # stop("")
  }, error = function(e) {
    return(NULL)
  })
  return(ah)
}

#### Message display functions

.printMsgArray <- function(msg, printType) {
  if (printType == "message") for (m in msg) message(m)
  if (printType == "cat") for (m in msg) cat(m, "\n")
}

.showREMParceInfo <- function(object) {
  info <- object@REMParcelInfo
  cat("REMParcel object\n")
  cat("RE type:", info[["REtype"]], "\n")
  cat("Genome build:", info[["genome"]], "\n")
  cat("Illumina platform:", info[["platform"]], "\n")
  cat("Valid (max)", paste0(info[["REtype"]], "-CpG"), "flanking window size:", info[["max.win"]], "\n")
  cat("Number of RE:", length(object@RE), "\n")
  cat("Number of", paste0(info[["REtype"]], "-CpG:"), length(object@RECpG), "\n")
}

.showTrainingStats <- function(REStats, REtype, printType, indent) {
  if (!isEmpty(REStats)) {
    p <- rep(NA, 2)

    p[1] <- paste0(
      indent, REStats[1, 1], " profiled ",
      REtype, " are used for model training."
    )
    p[2] <- paste0(
      indent, REStats[1, 2],
      " ", REtype, "-CpGs that have at least 2 neighboring profiled CpGs are used for model training."
    )
  } else {
    p <- "N/A"
  }
  .printMsgArray(p, printType)
}

.showREStats <- function(REStats, REtype, printType, indent, notAggregated) {
  if (!isEmpty(REStats)) {
    p <- paste0(indent, "The data cover ", REStats[1, 3], " ", REtype)
    if (notAggregated) p <- paste0(p, " (", REStats[1, 4], " ", REtype, "-CpG).")
  } else {
    p <- "N/A"
  }
  .printMsgArray(p, printType)
}

.showGeneStats <- function(GeneStats, REtype, printType, indent) {
  if (!isEmpty(GeneStats)) {
    p <- rep(NA, 4)

    p[1] <- paste0(indent, "Gene coverage by ", REtype, " (out of total # of RefSeq genes):")
    p[2] <- paste0(
      indent, indent, GeneStats[2, 3],
      " (", round(GeneStats[2, 3] / GeneStats[1, 3] * 100, 2), "%) total genes;"
    )
    p[3] <- paste0(
      indent, indent, GeneStats[2, 1],
      " (", round(GeneStats[2, 1] / GeneStats[1, 1] * 100, 2), "%) protein-coding genes;"
    )
    p[4] <- paste0(
      indent, indent, GeneStats[2, 2],
      " (", round(GeneStats[2, 2] / GeneStats[1, 2] * 100, 2), "%) non-coding RNA genes."
    )
  } else {
    p <- "N/A"
  }
  .printMsgArray(p, printType)
}

.showSampleID <- function(object, indent) {
  x <- object@sampleID
  cat("Number of sample(s):", length(x), "\n")
  cat("Sample ID:")
  if (length(x) > 5) {
    x <- x[seq_len(5)]
    cat(indent, paste0(x, collapse = ", "), " ...\n")
  } else {
    cat(indent, paste0(x, collapse = ", "), "\n")
  }
}

.showREMPinfo <- function(object) {
  info <- object@REMPInfo
  cat("RE type:", info[["REtype"]], "\n")
  cat("Genome build:", info[["genome"]], "\n")
  cat("Methylation profiling platform:", info[["platform"]], "\n")
  cat("Flanking window size:", info[["win"]], "\n")
  cat("Prediction model:", info[["predictModel"]], "\n")
  cat("QC model:", info[["QCModel"]], "\n")
  if (grepl("Random Forest", info[["predictModel"]])) {
    cat("Seed:", metadata(object)$Seed, "\n")
  }
  restats <- metadata(object)$REStats
  if (!isEmpty(restats)) {
    if (grepl("aggregated", info[["REtype"]])) {
      cat("Covered", restats[1, 3], info[["REtype"]], "\n")
    } else {
      cat(
        "Covered", restats[1, 4], "CpG sites in",
        restats[1, 3], info[["REtype"]], "\n"
      )
    }
  }
}

.showCpGcountbyChr <- function(object) {
  info <- object@REMPInfo
  if (grepl("aggregated", info[["REtype"]])) {
    cat("Number of", info[["REtype"]], "by chromosome:")
  } else {
    cat("Number of", paste0(info[["REtype"]], "-CpGs"), "by chromosome:")
  }
  chr <- table(as.character(seqnames(rowRanges(object))))
  chr.name <- names(chr)
  chr.name[chr.name == "chrX"] <- "chr23"
  chr.name[chr.name == "chrY"] <- "chr24"
  chr.num <- as.numeric(substring(chr.name, 4, 5))
  chr <- chr[order(chr.num)]

  for (i in seq_len(ceiling(length(chr) / 8)))
  {
    print(chr[((i - 1) * 8 + 1):min((8 * i), length(chr))])
  }
}

.showPredictionSummary <- function(object) {
  options(digits = 10)
  Bvalue <- assays(object)[["rempB"]]
  cat("Distribution of methylation value (beta value):", "\n")
  print(summary(as.numeric(Bvalue)))
  options(digits = 7)
}

.showQCSummary <- function(object) {
  options(digits = 10)
  QC <- assays(object)[["rempQC"]]
  cat("Distribution of reliability score (lower score = higher reliability):", "\n")
  print(summary(as.numeric(QC)))
  options(digits = 7)
}



## Convert refgene index to gene symbol or entrez id

.decodeEntrez <- function(x, refgene) {
  paste0(unique(refgene[x, "EntrezGene"]), collapse = "|")
}

.decodeSymbol <- function(x, refgene) {
  paste0(unique(refgene[x, "GeneSymbol"]), collapse = "|")
}



## Methylation beta <--> M value conversion

.toBeta <- function(M) {
  2^(M) / (1 + 2^(M))
}

.toM <- function(beta) {
  log2(beta / (1 - beta))
}



## Class check

.isGROrStop <- function(object) {
  if (!"GRanges" %in% is(object)) {
    stop(sprintf(
      "object is of class '%s', but needs to be of class 'GRanges' (see object definition in package 'GenomicRanges')",
      class(object)
    ))
  }
}

.isRSetOrStop <- function(object) {
  if (!is(object, "RatioSet") && !is(object, "GenomicRatioSet")) {
    stop(sprintf(
      "object is of class '%s', but needs to be of class '[Genomic]RatioSet' (see object definition in package 'minfi')",
      class(object)
    ))
  }
}

.isREMParcelOrStop <- function(object) {
  if (!is(object, "REMParcel")) {
    stop(sprintf(
      "object is of class '%s', but needs to be of class 'REMParcel'.",
      class(object)
    ))
  }
}

.isREMProductOrStop <- function(object) {
  if (!is(object, "REMProduct")) {
    stop(sprintf(
      "object is of class '%s', but needs to be of class 'REMProduct'.",
      class(object)
    ))
  }
}

#' @title REMProduct instances
#'
#' @description Class \code{REMProduct} is to maintain RE methylation prediction results.
#' \code{REMProduct} inherits Bioconductor's \code{RangedSummarizedExperiment} class.
#'
#' @name REMProduct-class
#'
#' @rdname REMProduct-class
#'
#' @param object A \code{REMProduct} object.
#' @param object1 A \code{REMProduct} object.
#' @param object2 A \code{REMProduct} object.
#' @param REtype Type of RE (\code{"Alu"}, \code{"L1"}, or \code{"ERV"}).
#' @param genome Specify the build of human genome. Can be either \code{"hg19"} or \code{"hg38"}.
#' @param platform Illumina methylation profiling platform (\code{"450k"} or \code{"EPIC"}).
#' @param win Flanking window size of the predicting RE-CpG.
#' @param predictModel Name of the model used for prediction.
#' @param QCModel Name of the model used for prediction quality evaluation.
#' @param rempM Predicted methylation level in M value.
#' @param rempB Predicted methylation level in beta value (optional).
#' @param rempQC Prediction quality scores, which is available only when Random Forest
#' model is used in \code{\link{remp}}.
#' @param cpgRanges Genomic ranges of the predicting RE-CpG.
#' @param sampleInfo Sample information.
#' @param REannotation Annotation data for the predicting RE.
#' @param RECpG Annotation data for the RE-CpG profiled by Illumina platform.
#' @param regionCode Internal index code defined in \code{refGene} for gene region indicators.
#' @param refGene refSeq gene annotation data, which can be obtained by \code{\link{fetchRefSeqGene}}.
#' @param varImp Importance of the predictors.
#' @param REStats RE coverage statistics, which is internally generated in \code{\link{remp}}.
#' @param GeneStats Gene coverage statistics, which is internally generated in \code{\link{remp}}.
#' @param Seed Random seed for Random Forest model for reproducible prediction results.
#' @param type For \code{plot} and \code{decodeAnnot}: see Utilities.
#' @param ncore For \code{decodeAnnot} and \code{rempAggregate}: number of cores used for parallel computing.
#' By default no parallel computing is allowed (\code{ncore = 1}).
#' @param BPPARAM For \code{decodeAnnot} and \code{rempAggregate}: an optional \code{\link{BiocParallelParam}}
#' instance determining the parallel back-end to be used during evaluation. If not specified, default
#' back-end in the machine will be used.
#' @param threshold For \code{rempTrim}: see Utilities.
#' @param missingRate For \code{rempTrim}: see Utilities.
#' @param NCpG For \code{rempAggregate}: see Utilities.
#' @param ... For \code{plot}: \code{\link{graphical parameters}} to be passed to the \code{plot} method.
#'
#' @return An object of class \code{REMProduct} for the constructor.
#'
#' @section Accessors:
#' \describe{
#'     \item{\code{rempM(object)}}{Return M value of the prediction.}
#'     \item{\code{rempB(object)}}{Return beta value of the prediction.}
#'     \item{\code{rempQC(object)}}{Return prediction quality scores.}
#'     \item{\code{rempImp(object)}}{Return relative importance of predictors.}
#'     \item{\code{rempStats(object)}}{Return RE and gene coverage statistics.}
#'     \item{\code{rempAnnot(object)}}{Return annotation data for the predicted RE.}
#'     }
#'
#' @section Utilities:
#' \describe{
#'     \item{\code{remplot(object, type = c("individual", "overall"), ...)}}{Make a density plot of predicted methylation
#'     (beta values) in the \code{REMProduct} object. If \code{type = "individual"}, density curves will be
#'     plotted for each of the samples; If \code{type = "overall"}, one density curve of the mean methylation level
#'     across the samples will be plotted. Default \code{type = "individual"}.}
#'     \item{\code{details(object)}}{Display detailed descriptive statistics of the predicion results.}
#'     \item{\code{decodeAnnot(object, type = c("symbol", "entrez")), ncore = NULL, BPPARAM = NULL}}{Decode the
#'     RE annotation data by Gene Symbol (when \code{type = "Symbol"}) or Entrez Gene
#'     (when \code{type = "Entrez"}).Default \code{type = "Symbol"}. Annotation data are provided by
#'     \code{\link{org.Hs.eg.db}}.}
#'     \item{\code{rempTrim(object, threshold = 1.7, missingRate = 0.2)}}{Any predicted CpG values with
#'     quality score < threshold (default = 1.7, specified by \code{threshold = 1.7}) will be replaced with NA.
#'     CpGs contain more than missingRate * 100% (default = 20%, specified by \code{missingRage = 0.2}) missing
#'     rate across samples will be discarded. Relavant summary statistics will be re-evaluated.}
#'     \item{\code{rempAggregate(object, NCpG = 2, ncore = NULL, BPPARAM = NULL)}}{Aggregate the predicted RE-CpG
#'     methylation by RE using mean. To ensure the reliability of the aggregation, by default only RE with at
#'     least 2 predicted CpG sites (specified by \code{NCpG = 2}) will be aggregated.}
#'     \item{\code{rempCombine(object1, object2)}}{Combine two \code{REMProduct} objects by column.}
#'     }
#'
#' @examples
#' showClass("REMProduct")
#' @exportClass REMProduct
REMProduct <- setClass("REMProduct",
  slots = c(REMPInfo = "CharacterList"),
  contains = "RangedSummarizedExperiment"
)

## Constructor function
REMProduct <- function(REtype = "Unknown", genome = "Unknown", 
                       platform = "Unknown", win = "Unknown",
                       predictModel = "Unknown", QCModel = "Unknown",
                       rempM = NULL, rempB = NULL, rempQC = NULL,
                       cpgRanges = GRanges(), sampleInfo = DataFrame(),
                       REannotation = GRanges(), RECpG = GRanges(),
                       regionCode = DataFrame(),
                       refGene = GRanges(),
                       varImp = DataFrame(),
                       REStats = DataFrame(), GeneStats = DataFrame(),
                       Seed = NULL) {
  rempInfo <- CharacterList(
    REtype = REtype,
    genome = genome,
    platform = platform,
    win = win,
    predictModel = predictModel,
    QCModel = QCModel
  )
  if (is.null(rempB)) rempB <- .toBeta(rempM)
  assays <- SimpleList(rempB = rempB, rempM = rempM, rempQC = rempQC)
  assays <- assays[!vapply(assays, is.null, logical(1))]

  stopifnot(identical(REannotation, unique(REannotation)))

  metadata <- list(
    REannotation = REannotation,
    RECpG = RECpG,
    regionCode = regionCode,
    refGene = refGene,
    varImp = varImp,
    REStats = REStats,
    GeneStats = GeneStats,
    Seed = Seed
  )
  new("REMProduct",
    REMPInfo = rempInfo,
    SummarizedExperiment(
      assays = assays,
      rowRanges = cpgRanges,
      colData = sampleInfo,
      metadata = metadata
    )
  )
}

## Welcome message when package is loaded

.onAttach <- function(libname, pkgname) {
  pkg_version <- packageVersion("REMP")
  dev = as.integer(pkg_version[1L, 2L]) %% 2L == 1L

  packageStartupMessage(rep("*", 80))
  
  packageStartupMessage("REMP version ", pkg_version, if(dev) paste0 (" (Devel)"), "\n",
                        "To access full functionality of REMP, please make sure this version is current.")
  
  citation <- paste0("\nCitation:\n",
                     "  Zheng Y, Joyce BT, Liu L, Zhang Z, Kibbe WA, Zhang W, Hou L.\n",
                     "  Prediction of genome-wide DNA methylation in repetitive elements.\n",
                     "  Nucleic Acids Research. 2017;45(15):8697-711.\n",
                     "  PubMed PMID: 28911103; PMCID: PMC5587781. http://dx.doi.org/10.1093/nar/gkx587")
  
  packageStartupMessage(citation)

  packageStartupMessage(rep("*", 80))
}

# REMParcel-methods

setMethod("show", signature(object = "REMParcel"), function(object) {
  .showREMParceInfo(object)
})

#' @rdname REMParcel-class
setMethod("saveParcel", signature(object = "REMParcel"), function(object, work.dir = tempdir(), verbose = FALSE, ...) {
  arrayType <- object@REMParcelInfo[["platform"]]
  REType <- object@REMParcelInfo[["REtype"]]
  win <- object@REMParcelInfo[["max.win"]]

  subDirName <- paste0("REMP.data.", arrayType)
  work.dir <- .forwardSlashPath(work.dir)
  data.dir <- .forwardSlashPath(file.path(work.dir, subDirName))

  if (dir.create(data.dir, showWarnings = FALSE)) {
    if (verbose) {
      message(
        "A new directory ", paste0("'", subDirName, "'"),
        " has been created under the directory:\n", data.dir
      )
    }
  } else {
    if (verbose) {
      message(
        "Directory ", paste0("'", subDirName, "'"),
        " already exists under the directory:\n", data.dir
      )
    }
  }

  path_to_parcel <- file.path(
    data.dir,
    paste0(
      "REMParcel_",
      REType, "_",
      arrayType, "_",
      win, ".rds"
    )
  )

  saveRDS(object, file = path_to_parcel, ...)

  message("REMParcel has been saved under the directory:\n", data.dir)
})

#' @rdname REMParcel-class
setMethod("getParcelInfo", signature(object = "REMParcel"), function(object) {
  return(object@REMParcelInfo)
})

#' @rdname REMParcel-class
setMethod("getRefGene", signature(object = "REMParcel"), function(object) {
  return(object@RefGene)
})

#' @rdname REMParcel-class
setMethod("getRE", signature(object = "REMParcel"), function(object) {
  return(object@RE)
})

#' @rdname REMParcel-class
setMethod("getRECpG", signature(object = "REMParcel"), function(object) {
  return(object@RECpG)
})

#' @rdname REMParcel-class
setMethod("getILMN", signature(object = "REMParcel"), function(object, REonly = FALSE) {
  ilmn <- object@ILMN
  if (REonly) {
    return(ilmn[!is.na(ilmn$RE.Index)])
  } else {
    return(ilmn)
  }
})

#' @title Prepare data template for REMP
#'
#' @description
#' \code{rempTemplate} is used to build a set of data templates for prediction. The data templates include
#' RE-CpGs and their methylation data for model training, neighboring CpGs of RE-CpGs and their
#' methylation data for model prediction, and other necessary information about the prediction.
#' This function is useful when one needs to experiment different tunning parameters so that these pre-built
#' data templates can be re-used and substaintially improve efficiency.
#'
#' @param methyDat A \code{\link{RatioSet}}, \code{\link{GenomicRatioSet}}, \code{\link{DataFrame}},
#' \code{data.table}, \code{data.frame}, or \code{matrix} of Illumina BeadChip methylation data
#' (450k or EPIC array) or Illumina methylation percentage estimates by sequencing.
#' @param Seq.GR A \code{\link{GRanges}} object containing genomic locations of the CpGs profiled by sequencing
#' platforms. This parameter should not be \code{NULL} if the input methylation data \code{methyDat} are
#' obtained by sequencing. Note that the genomic location can be in either hg19 or hg38 build, but must be consistent
#' with the build as \code{parcel}. See details in \code{\link{initREMP}}.
#' @param parcel An \code{\link{REMParcel}} object containing necessary data to carry out the
#' prediction. If \code{NULL}, the function will search the \code{.rds} data file in \code{work.dir}
#' exported by \code{\link{initREMP}} (with \code{export = TRUE}) or \code{\link{saveParcel}}.
#' @param win An integer specifying window size to confine the upstream and downstream flanking
#' region centered on the predicted CpG in RE for prediction. Default = \code{1000}.
#' @param verbose Logical parameter. Should the function be verbose?
#'
#' @return A \code{template} object containing a \code{\link{GRanges}} object of neifhboring CpGs of RE-CpGs
#' to be predicted (\code{$NBCpG_GR}) and their methylation dataset matrix (\code{$NBCpG_methyDat});
#' a \code{\link{GRanges}} object of RE-CpGs for model training (\code{$RECpG_GR}) and their methylation
#' dataset matrix (\code{$RECpG_methyDat}); \code{\link{GRanges}} objects of RefSeq Gene database (\code{$refgene})
#' and RE (\code{$RE}) that are extracted from the \code{parcel} input; a string of RE type (\code{$REtype})
#' and a string of methylation platform (\code{$arrayType}). Note: the subset operator \code{[]} is supported.
#'
#' @examples
#' if (!exists("GM12878_450k"))
#'   GM12878_450k <- getGM12878("450k")
#' if (!exists("remparcel")) {
#'   data(Alu.hg19.demo)
#'   remparcel <- initREMP(arrayType = "450k",
#'                         REtype = "Alu",
#'                         annotation.source = "AH",
#'                         genome = "hg19",
#'                         RE = Alu.hg19.demo,
#'                         ncore = 1,
#'                         verbose = TRUE)
#' }
#' 
#' template <- rempTemplate(GM12878_450k, 
#'                          parcel = remparcel, 
#'                          win = 1000, 
#'                          verbose = TRUE)
#' template
#' 
#' ## To make a subset
#' template[1]
#' 
#' @export
rempTemplate <- function(methyDat = NULL, 
                         Seq.GR = NULL, 
                         parcel = NULL, 
                         win = 1000, 
                         verbose = FALSE) {
  if (is.null(methyDat)) stop("Methylation dataset (methyDat) is missing.")
  if (!is.null(Seq.GR)) {
    .isGROrStop(Seq.GR)
    names(Seq.GR) <- NULL
  }
  
  if (is.null(parcel)) stop("REMParcel object is missing.")
  .isREMParcelOrStop(parcel)

  if (win > remp_options(".default.max.flankWindow")) {
    stop(
      "Flanking window size cannot be greater than ", remp_options(".default.max.flankWindow"),
      ". Please see ?remp_options for details."
    )
  }

  currenT <- Sys.time()

  ## Groom methylation data
  methyDat <- grooMethy(methyDat, Seq.GR, verbose = verbose)
  methyDat <- minfi::getM(methyDat)

  ## Get parcel information
  arrayType <- getParcelInfo(parcel)$platform
  REtype <- getParcelInfo(parcel)$REtype
  genome <- getParcelInfo(parcel)$genome
  RE_refGene.original <- getRE(parcel)
  RE_CpG <- getRECpG(parcel)
  ILMN <- getILMN(parcel)
  RE_CpG_ILMN <- getILMN(parcel, REonly = TRUE)

  ## Make indicator of genomic regions
  RE_refGene <- .toIndicator(RE_refGene.original)

  ## RE.CpG : Add CpG density and rename
  ## identical(runValue(RE_refGene$Index), runValue(RE_CpG$Index))
  RE_refGene$N <- runLength(RE_CpG$Index)
  RE_refGene$Length <- width(RE_refGene)
  RE_refGene$CpG.density <- RE_refGene$N / RE_refGene$Length # density of CpG within RE sequence

  merge_RE_CpG_meta <- mcols(RE_refGene)[base::match(
    as.character(RE_CpG$Index),
    as.character(RE_refGene$Index)), ]
  merge_RE_CpG_meta$CpG.ID <- as.character(granges(RE_CpG))
  mcols(RE_CpG) <- setNames(merge_RE_CpG_meta, paste0("RE.", colnames(merge_RE_CpG_meta)))
  # all(as.character(granges(RE_CpG)) == RE_CpG$RE.CpG.ID)

  ## Find valid CpG sites with methylation profiled
  commonCpGIndex <- intersect(rownames(methyDat), ILMN$Index)

  ## Trim down methyDat:
  methyDat <- methyDat[base::match(commonCpGIndex, rownames(methyDat)), , drop = FALSE]

  ## ILMN : create pointer and rename. Note: Different Methy data will
  ## result in different size of this data!
  ILMN <- ILMN[base::match(commonCpGIndex, ILMN$Index), ]
  mcols(ILMN) <- setNames(mcols(ILMN), paste0("ILMN.", colnames(mcols(ILMN))))
  ILMN$Methy.ptr <- seq_len(length(ILMN))
  # identical(ILMN$ILMN.Index, rownames(methyDat))

  ## RE_CpG_ILMN : Compute total number of RE covered by ILMN. Note:
  ## Different Methy data will result in different size of this data.
  RE_CpG_ILMN <- RE_CpG_ILMN[RE_CpG_ILMN$Index %in% rownames(methyDat), ]
  RE_CpG_ILMN <- unique(RE_CpG_ILMN)

  RE_CpG_ILMN_DATA <- methyDat[base::match(RE_CpG_ILMN$Index, rownames(methyDat)), , drop = FALSE]

  ############################################################# Find RE_CpG's neighboring CpG
  message("Processing ", REtype, " with window +/- ", win, " base pair ...")

  RE_CpG_flanking <- .twoWayFlank(granges(RE_CpG), win)
  HITS <- findOverlaps(RE_CpG_flanking, ILMN, ignore.strand = TRUE)

  ## Part 1: RE-CpG
  RE_NeibCpG <- RE_CpG[queryHits(HITS), ]
  ## Total RE that can be predicted (contains neighboring CpGs within given window)

  ## Part 2: Neighboring ILMN CpG
  RE_NeibCpG_ILMN <- ILMN[subjectHits(HITS), ]
  RE_NeibCpG_ILMN <- DataFrame(
    RE_NeibCpG_ILMN.GR = granges(RE_NeibCpG_ILMN),
    mcols(RE_NeibCpG_ILMN)
  )
  mcols(RE_NeibCpG) <- DataFrame(mcols(RE_NeibCpG), RE_NeibCpG_ILMN)
  # identical(as.character(granges(RE_NeibCpG)), RE_NeibCpG$RE.CpG.ID)

  ## Remove singleton (RE-CpGs with only one neighboring ILMN CpG)
  RE_NeibCpG$distance <- abs(start(RE_NeibCpG$RE_NeibCpG_ILMN.GR) - start(RE_NeibCpG))
  RE_NeibCpG <- RE_NeibCpG[RE_NeibCpG$distance > 1, ]

  ## Make distance bucket
  RE_NeibCpG$distance_cat <- cut(RE_NeibCpG$distance, seq(0, win, win / 4),
    labels = FALSE
  )
  RE_NeibCpG$distance_cat[is.na(RE_NeibCpG$distance_cat)] <- 4

  ## Add core predictors.
  RE_NeibCpG <- RE_NeibCpG[order(RE_NeibCpG$RE.CpG.ID, RE_NeibCpG$distance), ]
  RE_NeibCpG.DF <- mcols(RE_NeibCpG)[, c("RE.CpG.ID", "Methy.ptr", "distance")]

  distance_agg <- .aggregateNeib(
    distance ~ RE.CpG.ID, RE_NeibCpG.DF,
    function(x) c(length(x), mean(x), sd(x), x[1], x[2]),
    c(
      "RE.CpG.ID", "N.nbr", "distance.mean",
      "distance.std", "distance.min", "distance.min2"
    )
  )
  Methy_ptr_agg <- .aggregateNeib(
    Methy.ptr ~ RE.CpG.ID, RE_NeibCpG.DF,
    function(x) c(x[1], x[2]),
    c("RE.CpG.ID", "Methy.ptr.min", "Methy.ptr.min2")
  )

  Methy_ptr_distance_agg <- merge(distance_agg, Methy_ptr_agg, by = "RE.CpG.ID")

  RE_NeibCpG_meta <- mcols(RE_NeibCpG)
  mcols(RE_NeibCpG) <- cbind(
    RE_NeibCpG_meta,
    Methy_ptr_distance_agg[base::match(
      RE_NeibCpG_meta$RE.CpG.ID,
      Methy_ptr_distance_agg$RE.CpG.ID
    ), -1]
  )
  RE_NeibCpG <- RE_NeibCpG[order(RE_NeibCpG$RE.Index)]
  RE_NeibCpG <- RE_NeibCpG[RE_NeibCpG$N.nbr > 1, ] ## Remove singleton

  refgene_main <- getRefGene(parcel)
  RE_refGene.original <- getRE(parcel)

  res <- list(
    NBCpG_methyDat = DataFrame(methyDat), RECpG_methyDat = DataFrame(RE_CpG_ILMN_DATA),
    NBCpG_GR = RE_NeibCpG, RECpG_GR = RE_CpG_ILMN,
    refgene = getRefGene(parcel),
    RE = getRE(parcel),
    REtype = REtype,
    genome = genome,
    arrayType = arrayType
  )

  class(res) <- "template"

  message("Done.", .timeTrace(currenT)$t_text)

  return(res)
}

#' @export
"[.template" <- function(x, i) {
  x$NBCpG_methyDat <- x$NBCpG_methyDat[, i, drop = FALSE]
  x$RECpG_methyDat <- x$RECpG_methyDat[, i, drop = FALSE]
  x
}

#' @title RE Annotation Database Initialization
#'
#' @description
#' \code{initREMP} is used to initialize annotation database for RE methylation prediction.
#' Three RE types in human, Alu element (Alu), LINE-1 (L1), and endogenous retrovirus (ERV) are available.
#'
#' @param arrayType Illumina methylation array type. Currently \code{"450k"}, \code{"EPIC"},
#' and \code{"Sequencing"} are supported. Default = \code{"450k"}.
#' @param REtype Type of RE. Currently \code{"Alu"}, \code{"L1"}, and \code{"ERV"} are supported.
#' @param annotation.source Character parameter. Specify the source of annotation databases, including
#' the RefSeq Gene annotation database and RepeatMasker annotation database. If \code{"AH"}, the database 
#' will be obtained from the AnnotationHub package. If \code{"UCSC"}, the database will be downloaded 
#' from the UCSC website http://hgdownload.cse.ucsc.edu/goldenpath. The corresponding build (\code{"hg19"} or 
#' \code{"hg38"}) can be specified in the parameter \code{genome}.
#' @param genome Character parameter. Specify the build of human genome. Can be either \code{"hg19"} or 
#' \code{"hg38"}. Note that if \code{annotation.source == "AH"}, only hg19 database is available.
#' @param RE A \code{\link{GRanges}} object containing user-specified RE genomic location information.
#' If \code{NULL}, the function will retrive RepeatMasker RE database from \code{\link{AnnotationHub}}
#' (build hg19) or download the database from UCSC website (build hg19/hg38).
#' @param Seq.GR A \code{\link{GRanges}} object containing genomic locations of the CpGs profiled by sequencing
#' platforms. This parameter should not be \code{NULL} if \code{arrayType == 'Sequencing'}. Note that the genomic
#' location can be in either hg19 or hg38 build. See details.
#' @param ncore Number of cores used for parallel computing. By default max number of cores
#' available in the machine will be utilized. If \code{ncore = 1}, no parallel computing is allowed.
#' @param BPPARAM An optional \code{\link{BiocParallelParam}} instance determining the parallel back-end to
#' be used during evaluation. If not specified, default back-end in the machine will be used.
#' @param export Logical. Should the returned \code{\link{REMParcel}} object be saved to local machine?
#' See Details.
#' @param work.dir Path to the directory where the generated data will be saved. Valid when
#' \code{export = TRUE}. If not specified and \code{export = TRUE}, temporary directory \code{tempdir()}
#' will be used.
#' @param verbose Logical parameter. Should the function be verbose?
#'
#' @details
#' Currently, we support two major types of RE in the human genome, Alu and L1. The main purpose of
#' \code{initREMP} is to generate and annotate CpG/RE data using the refSeq Gene (hg19)
#' annotation database (provided by \code{\link{AnnotationHub}}). These annotation data are crucial to
#' RE methylation prediction in \code{\link{remp}}. Once generated, the data can be reused in the future
#' (data can be very large). Therefore, we recommend the user to save the output from
#' \code{initREMP} to the local machine, so that user only need to run this function once
#' as long as there is no change to the RE database. To minimize the size of the resulting data file, the generated
#' annotation data are only for REs that contain RE-CpGs with neighboring profiled CpGs. By default, the
#' neighboring CpGs are confined within 1200 bp flanking window. This window size can be modified using
#' \code{\link{remp_options}}. Note that the refSeq Gene database from UCSC is dynamic (updated periodically) 
#' and reflecting the latest knowledge of gene, whereas the database from AnnotationHub is static and classic. 
#' Using different sources will have a slight impact on the prediction results of RE methylation and gene annotation 
#' of final results. For sequencing methylation data, please specify the genomic location of CpGs
#' in a \code{GenomicRanges} object and specify it in \code{Seq.GR}. For an example of \code{Seq.GR}, Please
#' run \code{minfi::getLocations(IlluminaHumanMethylation450kanno.ilmn12.hg19)} (the row names of the CpGs in
#' \code{Seq.GR} can be \code{NULL}). The user should make sure the genome build of \code{Seq.GR} match the 
#' build specified in \code{genome} parameter (default is \code{"hg19"}).
#'
#' @return An \code{\link{REMParcel}} object containing data needed for RE methylation prediction.
#'
#' @seealso See \code{\link{remp}} for RE methylation prediction.
#'
#' @examples
#' if (!exists("remparcel")) {
#'   data(Alu.hg19.demo)
#'   remparcel <- initREMP(arrayType = "450k", 
#'                         REtype = "Alu", 
#'                         annotation.source = "AH",
#'                         genome = "hg19",
#'                         RE = Alu.hg19.demo, 
#'                         ncore = 1,
#'                         verbose = TRUE)
#' }
#' 
#' @export
initREMP <- function(arrayType = c("450k", "EPIC", "Sequencing"), 
                     REtype = c("Alu", "L1", "ERV"),
                     annotation.source = c("AH", "UCSC"), 
                     genome = c("hg19", "hg38"),
                     RE = NULL, 
                     Seq.GR = NULL,
                     ncore = NULL, 
                     BPPARAM = NULL,
                     export = FALSE, 
                     work.dir = tempdir(),
                     verbose = FALSE) {
  ## Initiate running time
  currenT <- Sys.time()

  arrayType <- match.arg(arrayType)
  REtype <- match.arg(REtype)
  annotation.source = match.arg(annotation.source)
  genome = match.arg(genome)
  
  if (!is.null(Seq.GR)) {
    .isGROrStop(Seq.GR)
    names(Seq.GR) <- NULL
  }
  
  message("Start ", REtype, " annotation data initialization (", genome, ")...")
  message("Gene annotation database: ", annotation.source)
  message("Illumina platform: ", arrayType)

  if (is.null(ncore)) ncore <- 1

  ## Setup backend for parallel computing
  be <- getBackend(ncore, BPPARAM, verbose)

  if (arrayType == "450k") {
    if (requireNamespace("IlluminaHumanMethylation450kanno.ilmn12.hg19", quietly = TRUE)) {
      suppressPackageStartupMessages(require("IlluminaHumanMethylation450kanno.ilmn12.hg19"))
      ILMN.GR <- minfi::getLocations(IlluminaHumanMethylation450kanno.ilmn12.hg19)
      if(genome == "hg38") {
        ILMN.GR <- .liftOver_Hg19toHg38(ILMN.GR, 
                                        "Lifting over Illumina CpG probe location from hg19 to hg38...",
                                        verbose)
      }
    } else stop("Please install missing package: IlluminaHumanMethylation450kanno.ilmn12.hg19")
  } else if (arrayType == "EPIC") {
    if (requireNamespace("IlluminaHumanMethylationEPICanno.ilm10b2.hg19", quietly = TRUE)) {
      suppressPackageStartupMessages(require("IlluminaHumanMethylationEPICanno.ilm10b2.hg19"))
      ILMN.GR <- minfi::getLocations(IlluminaHumanMethylationEPICanno.ilm10b2.hg19)
      if(genome == "hg38") {
        ILMN.GR <- .liftOver_Hg19toHg38(ILMN.GR, 
                                        "Lifting over Illumina CpG probe location from hg19 to hg38...",
                                        verbose)
      }
    } else stop("Please install missing package: IlluminaHumanMethylationEPICanno.ilm10b2.hg19")
  } else if (arrayType == "Sequencing") {
    if (!is.null(Seq.GR)) {
      ILMN.GR <- Seq.GR
    } else {
      stop("Seq.GR must be specified if arrayType == 'Sequencing'.")
    }
  }
  
  if (arrayType == "Sequencing") {
    ILMN.GR$Index <- paste0(seqnames(ILMN.GR), ":", start(ILMN.GR))
  } else {
    ILMN.GR <- ILMN.GR[substring(names(ILMN.GR), 1, 2) != "ch"] # remove ch probes
    ILMN.GR$Index <- names(ILMN.GR)
  }

  if (verbose) {
    message(
      "Valid RE-CpG flanking window size: ",
      remp_options(".default.max.flankWindow"), " bp"
    )
  }

  ##################################################
  ### ------ Part I - Resources preparation ------###
  ##################################################

  if (is.null(RE)) {
    ### Get RE annotation database
    RE.hg <- fetchRMSK(REtype = REtype, 
                       annotation.source = annotation.source,
                       genome = genome, 
                       verbose = verbose)
  } else {
    REtype_db <- .guessREtype(RE)
    if(REtype_db != REtype){
      message("Specified REtype (", REtype, ")", " does not match ", REtype_db, " database provided. REtype is set to '", REtype_db, "'.")
      REtype <- REtype_db
    }
    RE.hg <- RE
  }

  ### Get refSeq gene database
  refgene.hg <- fetchRefSeqGene(annotation.source, genome, mainOnly = FALSE, verbose)

  ### Get RE-CpG location database Narrow down RE.hg to RE sequence that
  ### overlaps with CpG sites flanking region For demo data, this will not
  ### change anything.
  ILMN.GR.flank <- .twoWayFlank(ILMN.GR, remp_options(".default.max.flankWindow"))
  RE.hg <- subsetByOverlaps(RE.hg, ILMN.GR.flank, ignore.strand = TRUE)

  ## Narrow down RE.hg to RE sequence that overlaps with gene sequence
  ## RE.hg <- subsetByOverlaps(RE.hg, refgene.hg$main, ignore.strand
  ## = TRUE)

  ## Locate RE-CpG
  RE.CpG <- findRECpG(RE.hg, REtype, genome, be, verbose)

  ## Make RE.hg and RE.CpG dataset consistent with RE
  RE.hg <- RE.hg[base::match(runValue(RE.CpG$Index), runValue(RE.hg$Index))]

  ########################################
  ### ------ Part II - Annotation ------###
  ########################################

  ### Annotate ILMN CpG
  ## Narrow down Illumina CpG Probe using RE.CpG flanking region
  RE.CpG.flanking <- .twoWayFlank(RE.CpG, remp_options(".default.max.flankWindow"))
  ILMN.GR <- subsetByOverlaps(ILMN.GR, RE.CpG.flanking, ignore.strand = TRUE)

  ### ----------------------------------------------------------------------------------------------------
  ### Annotate RE
  RE.refGene <- GRannot(RE.hg, refgene.hg, symbol = FALSE, verbose = verbose)

  ### ----------------------------------------------------------------------------------------------------
  ### RE-CpG covered by ILMN
  RECpG_Platform.hits <- findOverlaps(ILMN.GR, RE.CpG, ignore.strand = TRUE)
  mcols(ILMN.GR)$RE.Index <- Rle(NA)
  ILMN.GR$RE.Index[queryHits(RECpG_Platform.hits)] <- RE.CpG[subjectHits(RECpG_Platform.hits)]$Index

  remparcel <- REMParcel(
    REtype = REtype, 
    genome = genome,
    platform = arrayType,
    RefGene = refgene.hg$main,
    RE = RE.refGene, RECpG = RE.CpG,
    ILMN = ILMN.GR
  )

  message("Done.", .timeTrace(currenT)$t_text)

  if (export) {
    saveParcel(remparcel, work.dir, verbose)
  }

  return(remparcel)
} # End of intREMP


## --------------------------------------
## Internal functions

.guessREtype <- function(RE)
{
  .isGROrStop(RE)
  RE_family <- RE$repFamily
  if(is.null(RE_family)) stop("Please provide the 'repFamily' column specifying the RE family.")
  
  REtype <- c("Alu", "L1", "ERV")
  
  RE_count <- c(sum(grepl("Alu", RE_family)),
                sum(grepl("L1", RE_family)),
                sum(grepl("ERV", RE_family)))
  
  REtype <- REtype[which.max(RE_count)]
  return(REtype)
}

#' @title REMParcel instances
#'
#' @description \code{REMParcel} is a container class to organize required datasets for
#' RE methylation prediction generated from \code{\link{initREMP}} and used in \code{\link{remp}}.
#'
#' @name REMParcel-class
#'
#' @rdname REMParcel-class
#'
#' @param object A \code{REMParcel} object.
#' @param REtype Type of RE (\code{"Alu"}, \code{"L1"}, or \code{"ERV"}).
#' @param genome Specify the build of human genome. Can be either \code{"hg19"} or \code{"hg38"}.
#' @param platform Illumina methylation profiling platform (\code{"450k"} or \code{"EPIC"}).
#' @param RefGene refSeq gene annotation data, which can be obtained by \code{\link{fetchRefSeqGene}}.
#' @param RE Annotated RE genomic range data, which can be obtained by \code{\link{fetchRMSK}} and annotated
#' by \code{\link{GRannot}}.
#' @param RECpG Genomic range data of annotated CpG site identified in RE DNA sequence, which can
#' be obtained by \code{\link{findRECpG}} and annotated by \code{\link{GRannot}}.
#' @param ILMN Illumina CpG probe genomic range data.
#' @param REonly For \code{getILMN}: see Accessors.
#' @param work.dir For \code{saveParcel}: path to the directory where the generated data will be saved.
#' If not specified, temporary directory \code{tempdir()} will be used.
#' @param verbose For \code{saveParcel}: logical parameter. Should the function be verbose?
#' @param ... For \code{saveParcel}: other parameters to be passed to the \code{saveRDS} method. See
#' \code{\link{saveRDS}}.
#'
#' @return An object of class \code{REMParcel} for the constructor.
#'
#' @section Accessors:
#' \describe{
#'     \item{\code{getParcelInfo(object)}}{Return data type, RE type, and flanking window size information
#'     of the parcel.}
#'     \item{\code{getRefGene(object)}}{Return RefSeq gene annotation data.}
#'     \item{\code{getRE(object)}}{Return RE genomic location data for prediction
#'     (annotated by refSeq gene database).}
#'     \item{\code{getRECpG(object)}}{Return RE-CpG genomic location
#'     data for prediction.}
#'     \item{\code{getILMN(object, REonly = FALSE)}}{Return Illumina CpG probe genomic
#'     location data for prediction (annotated by refSeq gene database). If
#'     \code{REonly = TRUE}, only probes within RE region are returned.}
#'     }
#'
#' @section Utilities:
#' \describe{
#'     \item{\code{saveParcel(object, work.dir = tempdir(), verbose = FALSE, ...)}}{Save
#'     the object to local machine. }
#'     }
#'
#' @examples
#' showClass("REMParcel")
#' @exportClass REMParcel
REMParcel <- setClass("REMParcel",
  slots = c(
    REMParcelInfo = "CharacterList",
    RefGene = "GRanges",
    RE = "GRanges",
    RECpG = "GRanges",
    ILMN = "GRanges"
  ),
  validity = function(object) {
    if (!identical(
      runValue(object@RE$Index),
      runValue(object@RECpG$Index)
    )) {
      stop("RE data does not match RE-CpG data.")
    }
  }
)

## Constructor function
REMParcel <- function(REtype = "Unknown", 
                      genome = "Unknown", 
                      platform = "Unknown",
                      RefGene = GRanges(),
                      RE = GRanges(), RECpG = GRanges(), ILMN = GRanges()) {
  REMParcelInfo <- CharacterList(
    REtype = REtype,
    genome = genome,
    platform = platform,
    max.win = as.character(remp_options(".default.max.flankWindow"))
  )

  new("REMParcel",
    REMParcelInfo = REMParcelInfo,
    RefGene = RefGene,
    RE = RE,
    RECpG = RECpG,
    ILMN = ILMN
  )
}

#' @title Extract DNA methylation data profiled in RE
#'
#' @description
#' \code{remprofile} is used to extract profiled methylation of CpG sites in RE.
#'
#' @param methyDat A \code{\link{RatioSet}}, \code{\link{GenomicRatioSet}}, \code{\link{DataFrame}},
#' \code{data.table}, \code{data.frame}, or \code{matrix} of Illumina BeadChip methylation data
#' (450k or EPIC array) or Illumina methylation percentage estimates by sequencing.
#' @param genome Character parameter. Specify the build of human genome. Can be either \code{"hg19"} or 
#' \code{"hg38"}. For 450k/EPIC array, \code{"hg19"} is used more often while specifying \code{"hg38"} will 
#' lift over the Illumina CpG probe location to build \code{"hg38"}. For sequencing data, please make
#' sure the specified genome build is consistent with the actual genome build of \code{Seq.GR}.
#' @param REtype Type of RE. Currently \code{"Alu"}, \code{"L1"}, and \code{"ERV"} are supported.
#' @param annotation.source Character parameter. Specify the source of annotation databases, including
#' the RefSeq Gene annotation database and RepeatMasker annotation database. If \code{"AH"}, the database 
#' will be obtained from the AnnotationHub package. If \code{"UCSC"}, the database will be downloaded 
#' from the UCSC website http://hgdownload.cse.ucsc.edu/goldenpath. The corresponding build (\code{"hg19"} or 
#' \code{"hg38"}) will be specified in the parameter \code{genome}.
#' @param Seq.GR A \code{\link{GRanges}} object containing genomic locations of the CpGs profiled by sequencing
#' platforms. This parameter should not be \code{NULL} if the input methylation data \code{methyDat} are
#' obtained by sequencing. Note that the genomic location can be in either hg19 or hg38 build. The user
#' should make sure the parameter \code{genome} is correctly specified. 
#' @param RE A \code{\link{GRanges}} object containing user-specified RE genomic location information.
#' If \code{NULL}, the function will retrive RepeatMasker RE database from \code{\link{AnnotationHub}}
#' (build hg19) or download the database from UCSC website (build hg19/hg38).
#' @param impute Parameter used by \code{\link{grooMethy}}. If \code{TRUE}, K-Nearest Neighbouring
#' imputation will be applied to fill the missing values. Default = \code{FALSE}.
#' @param imputebyrow Parameter used by \code{\link{grooMethy}}. If \code{TRUE}, missing values will
#' be imputed using similar values in row (i.e., across samples); if \code{FALSE}, missing values
#' will be imputed using similar values in column (i.e., across CpGs). Default is \code{TRUE}.
#' @param verbose Logical parameter. Should the function be verbose?
#'
#' @return A \code{\link{REMProduct}} object containing profiled RE methylation results.
#'
#' @examples
#' data(Alu.hg19.demo)
#' if (!exists("GM12878_450k")) GM12878_450k <- getGM12878("450k")
#' remprofile.res <- remprofile(GM12878_450k,
#'                              REtype = "Alu",
#'                              annotation.source = "AH",
#'                              genome = "hg19",
#'                              RE = Alu.hg19.demo,
#'                              verbose = TRUE)
#' details(remprofile.res)
#' rempB(remprofile.res) # Methylation data (beta value)
#' 
#' remprofile.res <- rempAggregate(remprofile.res)
#' details(remprofile.res)
#' rempB(remprofile.res) # Methylation data (beta value)
#' 
#' @export
remprofile <- function(methyDat, 
                       REtype = c("Alu", "L1", "ERV"), 
                       annotation.source = c("AH", "UCSC"), 
                       genome = c("hg19", "hg38"), 
                       Seq.GR = NULL,
                       RE = NULL, 
                       impute = FALSE, 
                       imputebyrow = TRUE, 
                       verbose = FALSE) {
  if (is.null(methyDat)) stop("Methylation dataset (methyDat) is missing.")
  if (!is.null(Seq.GR)) {
    .isGROrStop(Seq.GR)
    names(Seq.GR) <- NULL
  }
  
  genome <- match.arg(genome)
  REtype <- match.arg(REtype)
  annotation.source = match.arg(annotation.source)
  
  ## Groom methylation data
  methyDat <- grooMethy(methyDat, Seq.GR, impute, verbose = verbose)
  arrayType <- gsub("IlluminaHumanMethylation", "", methyDat@annotation["array"])
  
  methyDat <- minfi::getM(methyDat)
  probeNames <- rownames(methyDat)
  
  if (arrayType == "450k") {
    if (requireNamespace("IlluminaHumanMethylation450kanno.ilmn12.hg19", quietly = TRUE)) {
      suppressPackageStartupMessages(require("IlluminaHumanMethylation450kanno.ilmn12.hg19"))
      ILMN.GR <- minfi::getLocations(IlluminaHumanMethylation450kanno.ilmn12.hg19)
      if(genome == "hg38") {
        ILMN.GR <- .liftOver_Hg19toHg38(ILMN.GR, 
                                        "Lifting over Illumina CpG probe location from hg19 to hg38...",
                                        verbose)
      }
    } else stop("Please install missing package: IlluminaHumanMethylation450kanno.ilmn12.hg19")
  } else if (arrayType == "EPIC") {
    if (requireNamespace("IlluminaHumanMethylationEPICanno.ilm10b2.hg19", quietly = TRUE)) {
      suppressPackageStartupMessages(require("IlluminaHumanMethylationEPICanno.ilm10b2.hg19"))
      ILMN.GR <- minfi::getLocations(IlluminaHumanMethylationEPICanno.ilm10b2.hg19)
      if(genome == "hg38") {
        ILMN.GR <- .liftOver_Hg19toHg38(ILMN.GR, 
                                        "Lifting over Illumina CpG probe location from hg19 to hg38...",
                                        verbose)
      }
    } else stop("Please install missing package: IlluminaHumanMethylationEPICanno.ilm10b2.hg19")
  } else if (arrayType == "Sequencing") {
    if (!is.null(Seq.GR)) {
      ILMN.GR <- Seq.GR
    } else {
      stop("Seq.GR must be specified if arrayType == 'Sequencing'.")
    }
  }
  
  if (arrayType == "Sequencing") {
    ILMN.GR$Index <- paste0(seqnames(ILMN.GR), ":", start(ILMN.GR))
  } else {
    ILMN.GR <- ILMN.GR[substring(names(ILMN.GR), 1, 2) != "ch"] # remove ch probes
    ILMN.GR$Index <- names(ILMN.GR)
  }

  probeNames <- intersect(probeNames, ILMN.GR$Index)

  methyDat <- methyDat[probeNames, , drop = FALSE]
  ILMN.GR <- ILMN.GR[base::match(probeNames, ILMN.GR$Index)]
  # identical(names(ILMN.GR), rownames(methyDat))

  if (is.null(RE)) {
    ### Get RE annotation database
    RE.hg <- fetchRMSK(REtype = REtype, 
                       annotation.source = annotation.source,
                       genome = genome, 
                       verbose = verbose)
  } else {
    .isGROrStop(RE)
    RE.hg <- RE
  }

  ### RE-CpG covered by ILMN
  RECpG_Platform.hits <- findOverlaps(RE.hg, ILMN.GR, ignore.strand = TRUE)

  # Update RE ranges
  RE.hg <- RE.hg[queryHits(RECpG_Platform.hits)]

  # Update CpG ranges
  mcols(ILMN.GR)$RE.Index <- Rle(NA)
  ILMN.GR$RE.Index[subjectHits(RECpG_Platform.hits)] <- RE.hg$Index
  RE_CpG_ILMN <- ILMN.GR[!is.na(ILMN.GR$RE.Index)]
  RE_CpG_ILMN <- RE_CpG_ILMN[order(RE_CpG_ILMN$RE.Index)]

  cpgRanges <- RE_CpG_ILMN

  RE_annotation <- unique(RE.hg)
  refgene.hg <- fetchRefSeqGene(annotation.source = annotation.source, genome = genome, mainOnly = FALSE, verbose)
  refgene_main <- refgene.hg$main

  RE_annotation <- GRannot(RE_annotation, refgene.hg, symbol = FALSE, verbose = verbose)

  RE_annotation_name <- colnames(mcols(RE_annotation))
  regionCode <- mcols(RE_annotation)[remp_options(".default.genomicRegionColNames")]
  RE_annotation <- RE_annotation[, RE_annotation_name[!RE_annotation_name %in%
    remp_options(".default.genomicRegionColNames")]]

  samplenames <- colnames(methyDat)
  sampleN <- length(samplenames)

  sampleinfo <- DataFrame(matrix(NA, nrow = sampleN, ncol = 1))
  colnames(sampleinfo) <- "Not_Applicable"
  rownames(sampleinfo) <- samplenames

  Profiled_RECpG_M <- methyDat[cpgRanges$Index, , drop = FALSE]

  ## RE coverage
  RE_COVERAGE <- .coverageStats_RE(RE_annotation, regionCode, cpgRanges, RE_CpG_ILMN,
    REtype,
    indent = "    ", verbose
  )

  # Gene coverage
  GENE_COVERAGE <- .coverageStats_GENE(regionCode, refgene_main,
    REtype,
    indent = "    ", verbose
  )

  remproduct <- REMProduct(
    REtype = REtype, genome = genome,
    platform = arrayType, win = "N/A",
    predictModel = "Profiled", QCModel = "N/A",
    rempM = Profiled_RECpG_M, rempQC = NULL,
    cpgRanges = cpgRanges, sampleInfo = sampleinfo,
    REannotation = RE_annotation,
    RECpG = ILMN.GR,
    regionCode = regionCode,
    refGene = refgene_main,
    varImp = DataFrame(),
    REStats = RE_COVERAGE, GeneStats = GENE_COVERAGE,
    Seed = NA
  )
  return(remproduct)
}

#' Subset of Alu genomic location dataset (hg19)
#'
#' A \code{\link{GRanges}} dataset containing 500 Alu sequences that have CpGs profiled
#' by both Illumina 450k and EPIC array. The variables are as follows:
#'
#' \itemize{
#'   \item seqnames: chromosome number
#'   \item ranges: hg19 genomic position
#'   \item strand: DNA strand
#'   \item swScore: Smith Waterman (SW) alignment score
#'   \item repName: Alu name
#'   \item repClass: Alu class
#'   \item repFamily: Alu family
#'   \item Index: internal index (meaningless for external use; 
#'   not communicable between genome builds)
#' }
#'
#' \code{Alu.hg19.demo} has the same format as the data object 
#' returned by \code{\link{fetchRMSK}}.
#'
#' @format A \code{\link{GRanges}} object.
#' 
#' @return A GRanges object with 500 ranges and 3 metadata columns.
#' 
#' @seealso See \code{\link{fetchRMSK}} to obtain the complete Alu/L1 dataset.

#' @source RepeatMasker database provided by package \code{AnnotationHub}
"Alu.hg19.demo"

#' Subset of Alu genomic location dataset (hg38)
#'
#' A \code{\link{GRanges}} dataset containing 500 Alu sequences that have CpGs profiled
#' by both Illumina 450k and EPIC array. The variables are as follows:
#'
#' \itemize{
#'   \item seqnames: chromosome number
#'   \item ranges: hg38 genomic position
#'   \item strand: DNA strand
#'   \item swScore: Smith Waterman (SW) alignment score
#'   \item repName: Alu name
#'   \item repClass: Alu class
#'   \item repFamily: Alu family
#'   \item Index: internal index (meaningless for external use; 
#'   not communicable between genome builds)
#' }
#'
#' \code{Alu.hg38.demo} has the same format as the data object 
#' returned by \code{\link{fetchRMSK}}.
#'
#' @format A \code{\link{GRanges}} object.
#' 
#' @return A GRanges object with 500 ranges and 3 metadata columns.
#' 
#' @seealso See \code{\link{fetchRMSK}} to obtain the complete Alu/L1 dataset.

#' @source RepeatMasker database (hg38) is provided by UCSC database downloaded 
#' from http://hgdownload.cse.ucsc.edu/goldenpath.
"Alu.hg38.demo"

#' @title Repetitive Element Methylation Prediction
#'
#' @description
#' Machine learning-based tools to predict DNA methylation of locus-specific repetitive elements (RE)
#' by learning surrounding genetic and epigenetic information. These tools provide genomewide and
#' single-base resolution of DNA methylation prediction on RE that are difficult to measure directly using
#' array-based or sequencing-based platforms, which enables epigenome-wide association study (EWAS)
#' and differentially methylated region (DMR) analysis on RE.
#'
#' @name REMP-package
#'
#' @aliases REMP-package REMP
#'
#' @docType package
#'
#' @section Overview - standard procedure:
#' \describe{
#'     \item{Step 1}{Start out generating required dataset for prediction using \code{\link{initREMP}}.
#'     The datasets include RE information, RE-CpG (i.e. CpGs located in RE region) information,
#'     and gene annotation, which are maintained in a \code{\link{REMParcel}} object.
#'     It is recommended to save these generated data to the working directory so they can be used in the future.}
#'     \item{Step 2}{Clean Illumina methylation dataset using \code{\link{grooMethy}}. This function
#'     can help identify and fix abnormal values and automatically impute missing values, which are
#'     essential for downstream prediction.}
#'     \item{Step 3}{Run \code{\link{remp}} to predict genome-wide locus specific RE methylation.}
#'     \item{Step 4}{Use the built-in accessors and utilities in \code{\link{REMProduct}} object to get or
#'     refine the prediction results.}
#'     }
#'
#' @author
#' Yinan Zheng \email{y-zheng@@northwestern.edu},
#' Lei Liu \email{lei.liu@@northwestern.edu},
#' Wei Zhang \email{wei.zhang1@@northwestern.edu},
#' Warren Kibbe \email{warren.kibbe@@nih.gov},
#' Lifang Hou \email{l-hou@@northwestern.edu}
#'
#' Maintainer: Yinan Zheng \email{y-zheng@@northwestern.edu}
#'
#' @references Zheng Y, Joyce BT, Liu L, Zhang Z, Kibbe WA, Zhang W, Hou L.
#' Prediction of genome-wide DNA methylation in repetitive elements.
#' Nucleic Acids Res. 2017;45(15):8697-711.
#' PubMed PMID: 28911103; PMCID: PMC5587781.
#' http://dx.doi.org/10.1093/nar/gkx587.
#'
#' @keywords package
#'
#' @import methods
#' @import S4Vectors
#' @import BiocParallel
#' @import caret
#' @importFrom SummarizedExperiment SummarizedExperiment rowRanges assays rowData colData
#' @importFrom minfi getAnnotation getBeta getM MethylSet RatioSet GenomicRatioSet IlluminaMethylationAnnotation
#' @importFrom readr read_delim cols_only col_character col_integer
#' @importFrom rtracklayer liftOver
#' @importFrom impute impute.knn
#' @importFrom kernlab predict
#' @importFrom ranger ranger importance
#' @importFrom settings stop_if_reserved reset options_manager
#' @importFrom stats setNames density na.pass
#' @importFrom utils download.file read.table data packageVersion
#' @importFrom graphics lines plot
#' @importFrom AnnotationHub AnnotationHub getAnnotationHubOption setAnnotationHubOption
#' @importFrom iterators nextElem idiv
#' @importFrom org.Hs.eg.db org.Hs.egREFSEQ2EG org.Hs.egSYMBOL
#' @importFrom GenomicRanges GRanges GRangesList makeGRangesListFromFeatureFragments findOverlaps
#' @importFrom GenomicRanges granges seqnames start end strand ranges promoters shift start<- end<-
#' @importFrom IRanges IRanges IRangesList CharacterList NumericList subsetByOverlaps
#' @importFrom Seqinfo seqlevels<-
#' @importFrom parallel detectCores
#' @importFrom doParallel registerDoParallel
#' @importFrom foreach registerDoSEQ
#' @importFrom BiocGenerics mget
#' @importFrom BSgenome getSeq
#' @importFrom Biostrings DNAString vmatchPattern
NULL

#' @title Groom methylation data to fix potential data issues
#'
#' @description
#' \code{grooMethy} is used to automatically detect and fix data issues including zero beta
#' value, missing value, and infinite value.
#'
#' @param methyDat A \code{\link{RatioSet}}, \code{\link{GenomicRatioSet}}, \code{\link{DataFrame}},
#' \code{data.table}, \code{data.frame}, or \code{matrix} of Illumina BeadChip methylation data
#' (450k or EPIC array) or Illumina methylation percentage estimates by sequencing. If the data are prepared
#' as a \code{data.frame} or alike format, for Illumina array data, please make sure there is a column or row names 
#' are available to indicate the Illumina probe names (i.e. cg00000029); for sequencing methylation data, please provide 
#' the corresponding CpG location information in \code{Seq.GR}.
#' @param Seq.GR A \code{\link{GRanges}} object containing genomic locations of the CpGs profiled by sequencing
#' platforms. This parameter should not be \code{NULL} if the input methylation data \code{methyDat} are
#' obtained by sequencing platform. The order of \code{Seq.GR} should match the order of \code{methyDat}.
#' Note that the genomic location can be in either hg19 or hg38 build. See details.
#' @param impute If \code{TRUE}, K-Nearest Neighbouring imputation will be applied to fill
#' the missing values. Default = \code{TRUE}. See Details.
#' @param imputebyrow If \code{TRUE}, missing values will be imputed using similar values in row
#' (i.e., across samples); if \code{FALSE}, missing values will be imputed using similar values
#' in column (i.e., across CpGs). Default is \code{TRUE}.
#' @param mapGenome Logical parameter. If \code{TRUE}, function will return a \code{\link{GenomicRatioSet}}
#' object instead of a \code{\link{RatioSet}}. This function is not applicable for sequencing data.
#' @param verbose Logical parameter. Should the function be verbose?
#'
#' @details
#' For methylation data in beta value, if zero/one value exists, the logit transformation
#' from beta to M value will produce infinite value. Therefore, zero/one beta value
#' will be replaced with the smallest non-zero beta/largest non-one beta value found in the dataset.
#' \code{grooMethy} can also handle missing value (i.e. \code{NA} or \code{NaN}) using KNN imputation (see
#' \code{\link{impute.knn}}). The infinite value will be also treated as missing value for imputation.
#' If the original dataset is in beta value, \code{grooMethy} will first transform it to M value
#' before imputation is carried out. If the imputed value is out of the original range (which is possible when
#' \code{imputebyrow = FALSE}), mean value will be used instead. Warning: imputed
#' values for multimodal distributed CpGs (across samples) may not be correct. Please check package \code{ENmix} to
#' identify the CpGs with multimodal distribution. Please note that \code{grooMethy} is
#' also embedded in \code{\link{remp}} so the user can run \code{\link{remp}} directly without
#' explicitly running \code{grooMethy}. For sequencing methylation data, please specify the genomic location of CpGs
#' in a \code{GenomicRanges} object and specify it in \code{Seq.GR}. For an example of \code{Seq.GR}, Please
#' run \code{minfi::getLocations(IlluminaHumanMethylation450kanno.ilmn12.hg19)} (the row names of the CpGs in \code{Seq.GR}
#' can be \code{NULL}). The user should make sure the genome build of \code{Seq.GR} match the build specified 
#' in \code{genome} parameter of function \code{\link{initREMP}} and \code{\link{remprofile}} (default is \code{"hg19"}).
#'
#' @return A \code{\link{RatioSet}} or \code{\link{GenomicRatioSet}} containing beta value and
#' M value of the methylation data.
#'
#' @examples
#' # Get GM12878 methylation data (450k array)
#' if (!exists("GM12878_450k")) GM12878_450k <- getGM12878("450k")
#' GM12878_450k <- grooMethy(GM12878_450k, verbose = TRUE)
#' 
#' # Also works if data input is a matrix
#' grooMethy(minfi::getBeta(GM12878_450k), verbose = TRUE)
#' @export
grooMethy <- function(methyDat, 
                      Seq.GR = NULL, 
                      impute = TRUE, 
                      imputebyrow = TRUE, 
                      mapGenome = FALSE, 
                      verbose = FALSE) {
  currenT <- Sys.time()

  if (is.null(methyDat)) stop("Methylation dataset (methyDat) is missing.")
  if (!is.null(Seq.GR)) {
    .isGROrStop(Seq.GR)
    names(Seq.GR) <- NULL
  }
  
  methyDat_work <- methyDat
  
  if (is(methyDat_work, "RatioSet") || is(methyDat_work, "GenomicRatioSet")) {
    if(minfi::preprocessMethod(methyDat_work) == "grooMethy(REMP)") {
      if (verbose) message("Methylation data are already groomed.")
      return(methyDat_work)
    }  
    methyDat_work <- minfi::getBeta(methyDat_work)
  }

  methyDat_work <- .methyMatrix(methyDat_work, Seq.GR)

  type <- .guessDataType(methyDat_work)
  arrayType <- .guessArrayType(methyDat_work)

  if (arrayType == "27k") {
    stop("Illumina 27k array is not supported.")
  }
  if (arrayType == "450k") {
    annotationInfo <- c(array = "IlluminaHumanMethylation450k", annotation = remp_options(".default.450k.annotation"))
  }
  if (arrayType == "EPIC") {
    annotationInfo <- c(array = "IlluminaHumanMethylationEPIC", annotation = remp_options(".default.epic.annotation"))
  }
  if (arrayType == "Sequencing" | !is.null(Seq.GR)) {
    annotationInfo <- c(array = "IlluminaHumanMethylationSequencing", annotation = "Custom")
  }

  if (verbose) {
    message(
      "Illumina ", arrayType, " Methylation data in ", type,
      " value detected."
    )
  }

  if (type == "percentage") {
    methyDat_work <- methyDat_work / 100
    if (verbose) message("Percentage data have been divided by 100 and scaled to range 0-1.")
  }

  dc <- .dataCheck(methyDat_work, type)

  if (is.null(dc$code)) {
    if (verbose) {
      message("    No issue found in the methylation dataset.")
    }
    impute <- FALSE
  } else {
    ## If beta and any beta value is out of range
    if (1 %in% dc$code) {
      stop("Negative beta or beta > 1 values detected! Please check your data and data preprocessing.")
    }

    ## If any beta value = 0 or 1 is found
    if (2 %in% dc$code) {
      nzero <- sum(methyDat_work == 0, na.rm = TRUE)
      none <- sum(methyDat_work == 1, na.rm = TRUE)

      if (nzero > 0) {
        if (verbose) message("    A total of ", nzero, " zero beta values are found.")
        smallBeta <- min(methyDat_work[methyDat_work > 0], na.rm = TRUE)
        if (verbose) message("    Fixing 'zero' beta values with the smallest non-0 beta value = ", smallBeta, " ...")
        methyDat_work[methyDat_work == 0] <- smallBeta
      }

      if (none > 0) {
        if (verbose) message("    A total of ", none, " one beta values are found.")
        bigBeta <- max(methyDat_work[methyDat_work < 1], na.rm = TRUE)
        if (verbose) message("    Fixing 'one' beta values with the largest non-1 beta value = ", bigBeta, " ...")
        methyDat_work[methyDat_work == 1] <- bigBeta
      }
    }

    ## If any beta/M value is NaN or infinite
    if (any(c(3, 4, 5) %in% dc$code)) {
      methyDat_work[is.nan(methyDat_work)] <- NA
      methyDat_work[is.infinite(methyDat_work)] <- NA

      if (verbose) {
        message("    A total of ", sum(is.na(methyDat_work)), " NA/NaN/Inf values are found.")
      }
    } else {
      impute <- FALSE
    }
  }

  ## Prepare both beta and M value using corrected data
  if (type == "M") {
    if (verbose) {
      message("    Converting M value to beta value ...")
    }
    betadata <- .toBeta(methyDat_work)
    mdata <- methyDat_work
  } else {
    if (verbose) {
      message("    Converting beta/percentage value to M value ...")
    }
    betadata <- methyDat_work
    mdata <- .toM(methyDat_work)
  }

  ## Impute the data!
  if (impute) {
    if (verbose) {
      message("    Imputing missing values using KNN method ...")
    }
    if (ncol(mdata) == 1) {
      stop("KNN-imputation cannot be applied to single sample.")
    }

    if (imputebyrow) resu <- impute::impute.knn(t(mdata)) else resu <- impute::impute.knn(mdata)

    ## Error checking and imperfect fix (use mean to replace the
    ## out-of-range imputation)
    rg <- apply(mdata, 1, function(x) range(x, na.rm = TRUE)) # Calculate the range of the data before imputation

    if (imputebyrow) imputed_mdata <- t(resu$data) else imputed_mdata <- resu$data

    ## Check if any imputed data is out of the original range
    idx <- which(imputed_mdata < rg[1, ] | imputed_mdata > rg[2, ],
      arr.ind = TRUE
    )

    ## If so then use mean to fix them
    if (nrow(idx) > 0) {
      if (verbose) {
        message("    Fixing ", nrow(idx), " imputed probes that are out of the original data range...")
      }
      m <- apply(mdata[idx[, 1], ], 1, function(x) mean(x, na.rm = TRUE)) ## calculate the mean of the original data
      for (i in seq_len(nrow(idx))) {
        imputed_mdata[idx[i, 1], idx[i, 2]] <- m[i]
      }
    } else {
      if (verbose) {
        message("    All imputed probes are within the original data range.")
      }
    }

    ## Update with imputed data
    mdata <- imputed_mdata
    betadata <- .toBeta(mdata)
  }

  rset <- minfi::RatioSet(Beta = betadata, M = mdata, 
                          annotation = annotationInfo, 
                          preprocessMethod = "grooMethy(REMP)")
  if (arrayType != "Sequencing" & mapGenome) {
    rset <- minfi::mapToGenome(rset)
  }

  if (verbose) message("Methylation data grooming is completed.", .timeTrace(currenT)$t_text)

  return(rset)
} ## End of grooMethy

## Internal functions
.methyMatrix <- function(methyDat, Seq.GR = NULL) {
  methyDat <- data.frame(methyDat, check.names = FALSE)

  probeNameIndicator <- which(vapply(methyDat, class, character(1)) %in% c("factor", "character"))

  if (!is.null(Seq.GR)) {
    if (length(probeNameIndicator) > 0)
      stop("Factor or Character columns detected! The input methylation data from sequencing platform should all be numeric.")
    if (nrow(methyDat) != length(Seq.GR)) 
      stop("The number of rows in methylation data (", nrow(methyDat), ") does not match the length of Seq.GR provided (", length(Seq.GR), ").")
    methyDat.matrix <- as.matrix(methyDat)
    rownames(methyDat.matrix) <- paste0(seqnames(Seq.GR), ":", start(Seq.GR))
  } else {
    if (length(probeNameIndicator) > 1) {
      stop(paste(
        "For array methylation data, please only keep one column or just use row names to indicate Illumina probe names (i.e. cg00000029).",
        "For sequencing methylation data, the parameter Seq.GR cannot be missing. Please provide it."
      ))
    }

    containILMN <- "cg" %in% unique(substring(methyDat[seq_len(10), probeNameIndicator], 1, 2))
    containRownames <- "cg" %in% unique(substring(rownames(methyDat)[seq_len(10)], 1, 2))
    if (!containILMN & !containRownames) {
      stop(paste(
        "For array methylation data, a column or row names that indicates Illumina probe names (i.e. cg00000029) is missing.",
        "Please fix it. For sequencing methylation data, the parameter Seq.GR cannot be missing. Please provide it."
      ))
    }

    if (containILMN) {
      methyDat.matrix <- as.matrix(methyDat[, -probeNameIndicator],
        ncol = ncol(methyDat) - 1
      )
      rownames(methyDat.matrix) <- methyDat[, probeNameIndicator]
      colnames(methyDat.matrix) <- colnames(methyDat)[-probeNameIndicator]
    } else if (containRownames) {
      methyDat.matrix <- as.matrix(methyDat)
    }
  }
  return(methyDat.matrix)
}

.dataCheck <- function(methyDat, type) {
  errorString <- NULL
  code <- NULL
  if (type %in% c("[Genomic]RatioSet", "beta", "percentage") & any(methyDat < 0 | methyDat >
    1, na.rm = TRUE)) {
    errorString <- c(errorString, "out-of-range")
    code <- c(code, 1)
  }
  if (type %in% c("[Genomic]RatioSet", "beta", "percentage") & any(methyDat == 0 | methyDat == 1, na.rm = TRUE)) {
    errorString <- c(errorString, "0/1")
    code <- c(code, 2)
  }
  if (any(is.nan(methyDat))) {
    errorString <- c(errorString, "NaN")
    code <- c(code, 3)
  }
  if (any(is.na(methyDat))) {
    errorString <- c(errorString, "NA")
    code <- c(code, 4)
  }
  if (any(is.infinite(methyDat))) {
    errorString <- c(errorString, "Inf")
    code <- c(code, 5)
  }
  return(list(code = code, errorString = errorString))
}

### REMParcel-class

#' @rdname REMParcel-class
#' @exportMethod getParcelInfo
setGeneric("getParcelInfo", function(object) standardGeneric("getParcelInfo"))

#' @rdname REMParcel-class
#' @exportMethod getRefGene
setGeneric("getRefGene", function(object) standardGeneric("getRefGene"))

#' @rdname REMParcel-class
#' @exportMethod getRE
setGeneric("getRE", function(object) standardGeneric("getRE"))

#' @rdname REMParcel-class
#' @exportMethod getRECpG
setGeneric("getRECpG", function(object) standardGeneric("getRECpG"))

#' @rdname REMParcel-class
#' @exportMethod getILMN
setGeneric("getILMN", function(object, ...) standardGeneric("getILMN"))

#' @rdname REMParcel-class
#' @exportMethod saveParcel
setGeneric("saveParcel", function(object, ...) standardGeneric("saveParcel"))


### REMProduct-class

#' @rdname REMProduct-class
#' @exportMethod rempM
setGeneric("rempM", function(object) standardGeneric("rempM"))

#' @rdname REMProduct-class
#' @exportMethod rempB
setGeneric("rempB", function(object) standardGeneric("rempB"))

#' @rdname REMProduct-class
#' @exportMethod rempQC
setGeneric("rempQC", function(object) standardGeneric("rempQC"))

#' @rdname REMProduct-class
#' @exportMethod rempAnnot
setGeneric("rempAnnot", function(object) standardGeneric("rempAnnot"))

#' @rdname REMProduct-class
#' @exportMethod rempImp
setGeneric("rempImp", function(object) standardGeneric("rempImp"))

#' @rdname REMProduct-class
#' @exportMethod rempStats
setGeneric("rempStats", function(object) standardGeneric("rempStats"))

#' @rdname REMProduct-class
#' @exportMethod remplot
setGeneric("remplot", function(object, ...) standardGeneric("remplot"))

#' @rdname REMProduct-class
#' @exportMethod details
setGeneric("details", function(object) standardGeneric("details"))

#' @rdname REMProduct-class
#' @exportMethod decodeAnnot
setGeneric("decodeAnnot", function(object, ...) standardGeneric("decodeAnnot"))

#' @rdname REMProduct-class
#' @exportMethod rempTrim
setGeneric("rempTrim", function(object, ...) standardGeneric("rempTrim"))

#' @rdname REMProduct-class
#' @exportMethod rempAggregate
setGeneric("rempAggregate", function(object, ...) standardGeneric("rempAggregate"))

#' @rdname REMProduct-class
#' @exportMethod rempCombine
setGeneric("rempCombine", function(object1, object2) standardGeneric("rempCombine"))

%\VignetteIndexEntry{An Introduction to the REMP Package}
%\VignetteKeywords{DNA Methylation, repetitive element, prediction}
%\VignettePackage{REMP}

\documentclass{article}

\usepackage{listings}
\usepackage{times}
\usepackage{hyperref}
\usepackage{biblatex}

\textwidth=6.5in
\textheight=9in
\parskip=.1cm
\oddsidemargin=0in
\evensidemargin=0in
\headheight=-.6in

\newcommand{\Rfunction}[1]{{\texttt{#1}}}
\newcommand{\Robject}[1]{{\texttt{#1}}}
\newcommand{\Rpkg}[1]{{\textit{#1}}}
\newcommand{\Rmethod}[1]{{\texttt{#1}}}
\newcommand{\Rfunarg}[1]{{\texttt{#1}}}
\newcommand{\Rclass}[1]{{\textit{#1}}}
\newcommand{\Rcode}[1]{{\texttt{#1}}}

<<echo=FALSE>>=
library(knitr)
options(width=72)

listing <- function(x, options) {
  paste("\\begin{lstlisting}[basicstyle=\\ttfamily,breaklines=true]\n",
    x, "\\end{lstlisting}\n", sep = "")
}
knit_hooks$set(source=listing, output=listing)
@

\title{An Introduction to the \Rpkg{REMP} Package}
\author{Yinan Zheng}
\date{\today}

\begin{document}

\maketitle

\tableofcontents

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\newpage

\section{Introduction}

\Rpkg{REMP} predicts DNA methylation of locus-specific repetitive elements (RE) by learning surrounding genetic and epigenetic information. \Rpkg{REMP} provides genomewide single-base resolution of DNA methylation on RE that is difficult to measure directly using array-based or sequencing-based platforms, which enables epigenome-wide association study (EWAS) and differentially methylated region (DMR) analysis on RE. \Rpkg{REMP} also provides handy tool to extract methylation data of CpGs that are located within RE sequences.

\Rpkg{REMP} supports both Illumina methylation BeadChip array platforms (450k and EPIC) and sequencing platforms (e.g. TruSeq Methyl Capture EPIC). Both genome build hg19 and hg38 are supported.

\section{Installation}
Install \Rpkg{REMP} (release version):

<<bioconductorREMPrelease, eval=FALSE>>=
if (!requireNamespace("BiocManager", quietly=TRUE))
    install.packages("BiocManager")
BiocManager::install("REMP")
@

To install devel version:

<<bioconductorREMPdev, eval=FALSE>>=
library(devtools)
install_github("YinanZheng/REMP")
@

Load \Rpkg{REMP} into the workspace:

<<loadREMP, message=FALSE, eval=TRUE>>=
library(REMP)
@

\section{REMP: Repetitive Element Methylation Prediction}
Currently \Rpkg{REMP} supports Human (hg19/hg38) Alu, LINE-1 (L1), and Long Terminal Repeat (LTR) (including endogenous retroviruses, ERV) repetitive element (RE) methylation prediction using Illumina 450k/EPIC array or sequencing platform.

\subsection{Groom methylation data}
Appropriate data preprocessing including quality control and normalization of methylation data are recommended before running \Rpkg{REMP}. Many packages are available to carry out these data preprocessing steps, for example, \Rpkg{minfi}, \Rpkg{wateRmelon}, and \Rpkg{methylumi}.

\Rpkg{REMP} is trying to minimize the requirement of the methylation data format. Users can maintain the methylation data in \Rclass{RatioSet} or \Rclass{GenomicRatioSet} object offered by \Rpkg{minfi}, \Rclass{data.table}, \Rclass{data.frame}, \Rclass{DataFrame}, or \Rclass{matrix}. Users can input either beta value or M-value. There are only two basic requirements of the methylation array data (450k/EPIC):
\begin{enumerate}
  \item Each row should represent CpG probe and each column should represent sample.
  \item The row names should indicate Illumina probe ID (i.e. cg00000029).
\end{enumerate}

However, there are some other common data issues that may prevent \Rpkg{REMP} from running correctly. For example, if the methylation data are in beta value and contain zero methylation values, logit transformation (to create M-value) will create negative infinite value; or the methylation data contain \Rcode{NA}, \Rcode{Inf}, or \Rcode{NaN} data. To tackle these potential issues, \Rpkg{REMP} includes a handy function \Rfunction{grooMethy} which can help detect and fix these issues. We highly recommend to take advantage of this function:

<<grooMethy, eval=TRUE>>=
# Get GM12878 methylation data (450k array)
GM12878_450k <- getGM12878('450k') 
GM12878_450k <- grooMethy(GM12878_450k)
GM12878_450k
@

For zero beta values, \Rfunction{grooMethy} will replace them with smallest non-zero beta value. For one beta values, \Rfunction{grooMethy} will replace them with largest non-one beta value. For \Rcode{NA}/\Rcode{NaN}/\Rcode{Inf} values, \Rfunction{grooMethy} will treat them as missing values and then apply KNN-imputation to complete the dataset. If the imputed value is out of the original range (which is possible when \Rcode{imputebyrow = FALSE}), mean value will be used instead. Warning: imputed values for multimodal distributed CpGs (across samples) may not be correct. Please check package \Rpkg{ENmix} to identify the CpGs with multimodal distribution. 

For sequencing data, the users only need to prepare a methylation data matrix (row = CpGs, column = samples). The corresponding CpG location information (either in hg19 or hg38) should be prepared in a separate \Rclass{GRanges} object and provide it to the \Rfunarg{Seq.GR} argument in \Rfunction{grooMethy}. For an example of \Rfunarg{Seq.GR}, please run:

<<SeqGR, eval=TRUE>>=
library(IlluminaHumanMethylation450kanno.ilmn12.hg19)
getLocations(IlluminaHumanMethylation450kanno.ilmn12.hg19)
@

Note that the row names of the CpGs in \Rfunarg{Seq.GR} can be \Rcode{NULL}. 

\subsection{Prepare annotation data}
To run \Rpkg{REMP} for RE methylation prediction, users first need to prepare some annotation datasets. The function \Rfunction{initREMP} is designed to do the job.

Suppose users will predict Alu methylation using Illumina 450k array data:

<<remparcel, eval=TRUE>>=
data(Alu.hg19.demo)
remparcel <- initREMP(arrayType = "450k",
                     REtype = "Alu",
                     annotation.source = "AH",
                     genome = "hg19",
                     RE = Alu.hg19.demo,
                     ncore = 1)
remparcel
@

For demonstration, we only use 500 selected Alu sequence dataset which comes along with the package (\Robject{Alu.hg19.demo}). We specify \Rfunarg{RE = Alu.hg19.demo}, so that the annotation dataset will be generated for the 500 selected Alu sequences. Most of the time, specifying \Rfunarg{RE} is not necessary, as the function will fetch the complete RE sequence dataset from package \Rpkg{AnnotationHub} using \Rfunction{fetchRMSK}. Users can also use this argument \Rfunarg{RE} to provide customized RE dataset.

\Rfunarg{annotation.source} allows the users to switch the source of the annotation databases, including the RefSeq Gene annotation database and RepeatMasker annotation database. If \Rfunarg{annotation.source = "AH"}, the database will be obtained from the AnnotationHub package. If \Rfunarg{annotation.source = "UCSC"}, the database will be downloaded from the UCSC website http://hgdownload.cse.ucsc.edu/goldenpath. The corresponding build (\Rfunarg{"hg19"} or \Rfunarg{"hg38"}) can be specified in the argument \Rfunarg{genome}. Most of the time \Rcode{"hg19"} is used for array data. But if \Rcode{"hg38"} is specified, the function will liftover the CpG probe location information to \Rcode{"hg38"} and obtain annotation databases in \Rcode{"hg38"}.

If \Rfunarg{arrayType = "Sequencing"}, users should provide the genomic location information of the CpGs in a \Rclass{GRanges} object to \Rfunarg{Seq.GR}. Note that the genome build of \Rfunarg{Seq.GR} provided should match the genome build specified in \Rfunarg{genome}.

All data are stored in the \Rclass{REMParcel} object:

<<saveParcel, eval=TRUE>>=
saveParcel(remparcel)
@

It is recommended to specify a working directory using argument \Rfunarg{work.dir} in \Rfunction{initREMP} so that the annnotation data generated can be re-used. Without specifying working directory, the annotation dataset will be created under the temporal directory \Rcode{tempdir()} by default. Users can also turn on the \Rfunarg{export} argument in \Rfunction{initREMP} to save the data automatically.

\subsection{Run prediction}

Once the annotation data are ready, users can pass the annotation data parcel to \Rfunction{remp} for prediction:

<<rempredict, eval=TRUE>>=
remp.res <- remp(GM12878_450k, 
                 REtype = 'Alu', 
                 parcel = remparcel, 
                 ncore = 1, 
                 seed = 777)
@

If \Rfunarg{parcel} is missing, \Rfunction{remp} will then try to search the \Rclass{REMParcel} data file in the directory indicated by \Rfunarg{work.dir}. If \Rfunarg{work.dir} is also missing, \Rfunction{remp} will try to search the REMParcel data file in the temporal directory \Rcode{tempdir()}.

By default, \Rfunction{remp} uses Random Forest (\Rfunarg{method = 'rf'}) model (package \Rpkg{ranger} for fast implementation) for prediction. Random Forest model is recommended because it offers more accurate prediction results and it automatically enables Quantile Regression Forest (Nicolai Meinshausen, 2006) for prediction reliability evaluation. \Rfunction{remp} constructs predictors to carry out the prediction. For Random Forest model, the tuning parameter \Rfunarg{param = 6} (i.e. \Rfunarg{mtry} in \Rpkg{ranger} or \Rpkg{randomForest}) indicates how many predictors will be randomly selected for building the individual trees. The performance of random forest model is often relatively insensitive to the choice of \Rfunarg{mtry}. Therefore, auto-tune will be turned off using random forest and \Rfunarg{mtry} will be set to one third of the total number of predictors. It is recommended to specify a seed for reproducible prediction results.

Besides random forest, \Rfunction{remp} provides other machine learning engines for users to explore, including Extreme Gradient Boosting, SVM with linear kernel, and SVM with radial kernel).

\Rfunction{remp} will return a \Rclass{REMPset} object, which inherits Bioconductor's \Rclass{RangedSummarizedExperiment} class:

<<rempprint, eval=TRUE>>=
remp.res

# Display more detailed information
details(remp.res)
@

Prediction results can be obtained by accessors: 

<<rempaccessors, eval=TRUE>>=
# Predicted RE-CpG methylation value (Beta value)
rempB(remp.res)

# Predicted RE-CpG methylation value (M value)
rempM(remp.res)

# Genomic location information of the predicted RE-CpG
# Function inherit from class 'RangedSummarizedExperiment'
rowRanges(remp.res)

# Standard error-scaled permutation importance of predictors
rempImp(remp.res)

# Retrive seed number used for the reesults
metadata(remp.res)$Seed
@

Trim off less reliable predicted results:

<<remptrim, eval=TRUE>>=
# Any predicted CpG values with quality score less than
# threshold (default = 1.7) will be replaced with NA. 
# CpGs contain more than missingRate * 100% (default = 20%) 
# missing rate across samples will be discarded. 
remp.res <- rempTrim(remp.res, threshold = 1.7, missingRate = 0.2)
details(remp.res)
@

(Optional) Aggregate the predicted methylation of CpGs in RE by averaging them to obtain the RE-specific methylation level:

<<rempaggregate, eval=TRUE>>=
remp.res <- rempAggregate(remp.res, NCpG = 2)
details(remp.res)
@

Aggregating CpGs in the same RE for RE-level methylation data is beneficial because 1) it greatly reduces the data dimension for downstream analysis and 2) it may produce more robust RE methylation estimation. Note that by default, RE with 2 or more predicted CpG sites will be aggregated. Therefore, the downside of doing this is the reduced coverage of RE. The assumption of doing this is the CpG methylation level within each RE are similar.

To add genomic regions annotation of the predicted REs:

<<rempdecodeAnnot, eval=TRUE>>=
# By default gene symbol annotation will be added 
remp.res <- decodeAnnot(remp.res)
rempAnnot(remp.res)
@

Seven genomic region indicators will be added to the annotation data in the input \Rclass{REMProduct} object:

\begin{itemize}
\item InNM: in protein-coding genes (overlap with refSeq gene's "NM" transcripts + 2000 bp upstream of the transcription start site (TSS))
\item InNR: in noncoding RNA genes (overlap with refSeq gene's "NR" transcripts + 2000 bp upstream of the TSS)
\item InTSS: in flanking region of 2000 bp upstream of the TSS. Default upstream limit is 2000 bp, which can be modified globally using \Rfunction{remp\_options}
\item In5UTR: in 5'untranslated regions (UTRs)
\item InCDS: in coding DNA sequence regions
\item InExon: in exon regions
\item In3UTR: in 3'UTRs
\end{itemize}

Note that intron region and intergenic region information can be derived from the above genomic region indicators: if "InNM" and/or "InNR" is not missing but "InTSS", "In5UTR", "InExon", and "In3UTR" are missing, then the RE is strictly located within intron region; if all indicators are missing, then the RE is strictly located in intergenic region.

\subsection{Plot prediction}

Make a density plot of the predicted methylation (beta values):

\begin{figure}[h!]
\centering
<<rempplot, eval=TRUE, fig.width=4.5, fig.height=4>>=
remplot(remp.res, main = "Alu methylation (GM12878)", col = "blue")
@
\end{figure}

\pagebreak

\section{Extract RE-CpG methylation profiled by Illumina BeadChip array}

\Rpkg{REMP} offers a handy tool to extract methylation data of CpGs that are located in RE. Similar as \Rfunction{remp}, users can choose the source of annotation database (\Rcode{AH}: AnnotationHub or \Rcode{UCSC}: UCSC website) and genome build (\Rcode{hg19} or \Rcode{hg38}). 

<<remprofile, eval=TRUE>>=
# Use Alu.hg19.demo for demonstration
remp.res <- remprofile(GM12878_450k, 
                       REtype = "Alu", 
                       annotation.source = "AH", 
                       genome = "hg19", 
                       RE = Alu.hg19.demo)
details(remp.res)

# All accessors and utilites for REMProduct are applicable
remp.res <- rempAggregate(remp.res)
details(remp.res)
@

\end{document}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/REMPtools.R
\name{getBackend}
\alias{getBackend}
\title{Get BiocParallel back-end}
\usage{
getBackend(ncore, BPPARAM = NULL, verbose = FALSE)
}
\arguments{
\item{ncore}{Number of cores used for parallel computing. By default max number
of cores available in the machine will be utilized. If \code{ncore = 1}, no
parallel computing is allowed.}

\item{BPPARAM}{An optional \code{BiocParallelParam} instance determining the
parallel back-end to be used during evaluation. If not specified, default
back-end in the machine will be used.}

\item{verbose}{Logical parameter. Should the function be verbose?}
}
\value{
A \code{\link{BiocParallel}} object that can be used for parallel
computing.
}
\description{
\code{getBackend} is used to obtain \code{BiocParallel} Back-end to allow
parallel computing.
}
\examples{
# Non-parallel mode
be <- getBackend(ncore = 1, verbose = TRUE)
be

# parallel mode (2 workers)
be <- getBackend(ncore = 2, verbose = TRUE)
be

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getGM12878.R
\name{getGM12878}
\alias{getGM12878}
\title{Get methylation data of HapMap LCL sample GM12878 profiled by
Illumina 450k array or EPIC array}
\usage{
getGM12878(arrayType = c("450k", "EPIC"), mapGenome = FALSE)
}
\arguments{
\item{arrayType}{Illumina methylation array type. Currently \code{"450k"}
and \code{"EPIC"} are supported. Default = \code{"450k"}.}

\item{mapGenome}{Logical parameter. If \code{TRUE}, function will return
a \code{\link{GenomicRatioSet}} object instead of a \code{link{RatioSet}}
object.}
}
\value{
A \code{\link{RatioSet}} or \code{\link{GenomicRatioSet}} containing
beta value and M value of the methylation data.
}
\description{
\code{getGM12878} is used to obtain public available methylation profiling
data of HapMap LCL sample GM12878.
}
\details{
Illumina 450k data were sourced and curated from ENCODE
\url{http://hgdownload.cse.ucsc.edu/goldenPath/hg19/encodeDCC/wgEncodeHaibMethyl450/wgEncodeHaibMethyl450Gm12878SitesRep1.bed.gz}.
Illumina EPIC data were obtained from data package \code{minfiDataEPIC}.
}
\examples{
\dontrun{
# Get GM12878 methylation data (450k array)
if (!exists("GM12878_450k")) GM12878_450k <- getGM12878("450k")
GM12878_450k
}

# Get GM12878 methylation data (EPIC array)
if (!exists("GM12878_EPIC")) GM12878_EPIC <- getGM12878("EPIC")
GM12878_EPIC
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/REMPtools.R
\name{fetchRefSeqGene}
\alias{fetchRefSeqGene}
\title{Get RefSeq gene database}
\usage{
fetchRefSeqGene(
  annotation.source = c("AH", "UCSC"),
  genome = c("hg19", "hg38"),
  mainOnly = FALSE,
  verbose = FALSE
)
}
\arguments{
\item{annotation.source}{Character parameter. Specify the source of annotation databases, including
the RefSeq Gene annotation database and RepeatMasker annotation database. If \code{"AH"}, the database 
will be obtained from the AnnotationHub package. If \code{"UCSC"}, the database will be downloaded 
from the UCSC website http://hgdownload.cse.ucsc.edu/goldenpath. The corresponding build (\code{"hg19"} or 
\code{"hg38"}) will be specified in the parameter \code{genome}.}

\item{genome}{Character parameter. Specify the build of human genome. Can be either \code{"hg19"} or 
\code{"hg38"}. Note that if \code{annotation.source == "AH"}, only hg19 database is available.}

\item{mainOnly}{Logical parameter. See details.}

\item{verbose}{Logical parameter. Should the function be verbose?}
}
\value{
A single \code{\link{GRanges}} (for main refgene data) object or a list incorporating
both \code{\link{GRanges}} object (for main refgene data) and \code{\link{GRangesList}} object
(for gene regions data).
}
\description{
\code{fetchRefSeqGene} is used to obtain refSeq gene database provided by AnnotationHub (hg19) 
or UCSC web database (hg19/hg38).
}
\details{
When \code{mainOnly = FALSE}, only the transcript location information will be returned,
Otherwise, a \code{\link{GRangesList}} object containing gene regions
information will be added. Gene regions include: 2000 base pair upstream of the transcript
start site (\code{$tss})), 5'UTR (\code{$fiveUTR})), coding sequence (\code{$cds})),
exon (\code{$exon})), and 3'UTR (\code{$threeUTR})). The \code{index} column is an internal
index that is used to facilitate data referral, which is meaningless for external use.
}
\examples{
if (!exists("refgene.hg19")) 
refgene.hg19 <- fetchRefSeqGene(annotation.source = "AH", 
                                genome = "hg19", 
                                verbose = TRUE)
refgene.hg19

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/REMPtools.R
\name{findRECpG}
\alias{findRECpG}
\title{Find RE-CpG genomic location given RE ranges information}
\usage{
findRECpG(
  RE,
  REtype = c("Alu", "L1", "ERV"),
  genome = c("hg19", "hg38"),
  be = NULL,
  verbose = FALSE
)
}
\arguments{
\item{RE}{A \code{\link{GRanges}} object of RE genomic location database. This
can be obtained by \code{\link{fetchRMSK}}.}

\item{REtype}{Type of RE. Currently \code{"Alu"}, \code{"L1"}, and \code{"ERV"} are supported.}

\item{genome}{Character parameter. Specify the build of human genome. Can be either \code{"hg19"} or 
\code{"hg38"}. User should make sure the genome build of \code{RE} is consistent with this parameter.}

\item{be}{A \code{\link{BiocParallel}} object containing back-end information that is
ready for parallel computing. This can be obtained by \code{\link{getBackend}}. If not specified,
non-parallel mode is used.}

\item{verbose}{logical parameter. Should the function be verbose?}
}
\value{
A \code{\link{GRanges}} object containing identified RE-CpG genomic
location data.
}
\description{
\code{findRECpG} is used to obtain RE-CpG genomic location data.
}
\details{
CpG site is defined as 5'-C-p-G-3'. It is reasonable to assume that the methylation
status across all CpG/CpG dyads are concordant. Maintenance methyltransferase exhibits
a preference for hemimethylated CpG/CpG dyads (methylated on one strand only).
As a result, methyaltion status of CpG sites in both forward and reverse strands are usually consistent.
Therefore, to accommodate the cytosine loci in both strands, the returned genomic
ranges cover the 'CG' sequence with width of 2. The 'strand' information indicates the strand of the RE.
Locating CpG sites in RE sequences can be computation intensive. It is recommanded to get more than
one work in the backend for a faster running speed.
}
\examples{
data(Alu.hg19.demo)
RE.CpG <- findRECpG(RE = Alu.hg19.demo, 
                    REtype = "Alu", 
                    genome = "hg19", 
                    verbose = TRUE)
RE.CpG

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/remprofile.R
\name{remprofile}
\alias{remprofile}
\title{Extract DNA methylation data profiled in RE}
\usage{
remprofile(
  methyDat,
  REtype = c("Alu", "L1", "ERV"),
  annotation.source = c("AH", "UCSC"),
  genome = c("hg19", "hg38"),
  Seq.GR = NULL,
  RE = NULL,
  impute = FALSE,
  imputebyrow = TRUE,
  verbose = FALSE
)
}
\arguments{
\item{methyDat}{A \code{\link{RatioSet}}, \code{\link{GenomicRatioSet}}, \code{\link{DataFrame}},
\code{data.table}, \code{data.frame}, or \code{matrix} of Illumina BeadChip methylation data
(450k or EPIC array) or Illumina methylation percentage estimates by sequencing.}

\item{REtype}{Type of RE. Currently \code{"Alu"}, \code{"L1"}, and \code{"ERV"} are supported.}

\item{annotation.source}{Character parameter. Specify the source of annotation databases, including
the RefSeq Gene annotation database and RepeatMasker annotation database. If \code{"AH"}, the database 
will be obtained from the AnnotationHub package. If \code{"UCSC"}, the database will be downloaded 
from the UCSC website http://hgdownload.cse.ucsc.edu/goldenpath. The corresponding build (\code{"hg19"} or 
\code{"hg38"}) will be specified in the parameter \code{genome}.}

\item{genome}{Character parameter. Specify the build of human genome. Can be either \code{"hg19"} or 
\code{"hg38"}. For 450k/EPIC array, \code{"hg19"} is used more often while specifying \code{"hg38"} will 
lift over the Illumina CpG probe location to build \code{"hg38"}. For sequencing data, please make
sure the specified genome build is consistent with the actual genome build of \code{Seq.GR}.}

\item{Seq.GR}{A \code{\link{GRanges}} object containing genomic locations of the CpGs profiled by sequencing
platforms. This parameter should not be \code{NULL} if the input methylation data \code{methyDat} are
obtained by sequencing. Note that the genomic location can be in either hg19 or hg38 build. The user
should make sure the parameter \code{genome} is correctly specified.}

\item{RE}{A \code{\link{GRanges}} object containing user-specified RE genomic location information.
If \code{NULL}, the function will retrive RepeatMasker RE database from \code{\link{AnnotationHub}}
(build hg19) or download the database from UCSC website (build hg19/hg38).}

\item{impute}{Parameter used by \code{\link{grooMethy}}. If \code{TRUE}, K-Nearest Neighbouring
imputation will be applied to fill the missing values. Default = \code{FALSE}.}

\item{imputebyrow}{Parameter used by \code{\link{grooMethy}}. If \code{TRUE}, missing values will
be imputed using similar values in row (i.e., across samples); if \code{FALSE}, missing values
will be imputed using similar values in column (i.e., across CpGs). Default is \code{TRUE}.}

\item{verbose}{Logical parameter. Should the function be verbose?}
}
\value{
A \code{\link{REMProduct}} object containing profiled RE methylation results.
}
\description{
\code{remprofile} is used to extract profiled methylation of CpG sites in RE.
}
\examples{
data(Alu.hg19.demo)
if (!exists("GM12878_450k")) GM12878_450k <- getGM12878("450k")
remprofile.res <- remprofile(GM12878_450k,
                             REtype = "Alu",
                             annotation.source = "AH",
                             genome = "hg19",
                             RE = Alu.hg19.demo,
                             verbose = TRUE)
details(remprofile.res)
rempB(remprofile.res) # Methylation data (beta value)

remprofile.res <- rempAggregate(remprofile.res)
details(remprofile.res)
rempB(remprofile.res) # Methylation data (beta value)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/REMPtools.R
\name{GRannot}
\alias{GRannot}
\title{Annotate genomic ranges data with gene region information.}
\usage{
GRannot(object.GR, refgene, symbol = FALSE, verbose = FALSE)
}
\arguments{
\item{object.GR}{An \code{\link{GRanges}} object of a genomic location database.}

\item{refgene}{A complete refGene annotation database returned by
\code{\link{fetchRefSeqGene}} (with parameter \code{mainOnly = FALSE}).}

\item{symbol}{Logical parameter. Should the annotation return gene symbol?}

\item{verbose}{Logical parameter. Should the function be verbose?}
}
\value{
A \code{\link{GRanges}} or a \code{\link{GRangesList}} object containing refSeq
Gene database.
}
\description{
\code{GRannot} is used to annotate a \code{\link{GRanges}} dataset with gene region
information using refseq gene database
}
\details{
The annotated gene region information includes: protein coding gene (InNM),
noncoding RNA gene (InNR), 2000 base pair upstream of the transcript start site (InTSS),
5'UTR (In5UTR), coding sequence (InCDS), exon (InExon), and 3'UTR (In3UTR). The intergenic
and intron regions can then be represented by the combination of these region data.
The number shown in these columns represent the row number or 'index' column in the
main refgene database obtained by \code{\link{fetchRefSeqGene}}.
}
\examples{
data(Alu.hg19.demo)
if (!exists("refgene.hg19")) 
  refgene.hg19 <- fetchRefSeqGene(annotation.source = "AH", 
                                  genome = "hg19",
                                  verbose = TRUE)
Alu.hg19.demo.refGene <- GRannot(Alu.hg19.demo, refgene.hg19, verbose = TRUE)
Alu.hg19.demo.refGene

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/REMP-package.R
\docType{package}
\name{REMP-package}
\alias{REMP-package}
\alias{REMP}
\title{Repetitive Element Methylation Prediction}
\description{
Machine learning-based tools to predict DNA methylation of locus-specific repetitive elements (RE)
by learning surrounding genetic and epigenetic information. These tools provide genomewide and
single-base resolution of DNA methylation prediction on RE that are difficult to measure directly using
array-based or sequencing-based platforms, which enables epigenome-wide association study (EWAS)
and differentially methylated region (DMR) analysis on RE.
}
\section{Overview - standard procedure}{

\describe{
    \item{Step 1}{Start out generating required dataset for prediction using \code{\link{initREMP}}.
    The datasets include RE information, RE-CpG (i.e. CpGs located in RE region) information,
    and gene annotation, which are maintained in a \code{\link{REMParcel}} object.
    It is recommended to save these generated data to the working directory so they can be used in the future.}
    \item{Step 2}{Clean Illumina methylation dataset using \code{\link{grooMethy}}. This function
    can help identify and fix abnormal values and automatically impute missing values, which are
    essential for downstream prediction.}
    \item{Step 3}{Run \code{\link{remp}} to predict genome-wide locus specific RE methylation.}
    \item{Step 4}{Use the built-in accessors and utilities in \code{\link{REMProduct}} object to get or
    refine the prediction results.}
    }
}

\references{
Zheng Y, Joyce BT, Liu L, Zhang Z, Kibbe WA, Zhang W, Hou L.
Prediction of genome-wide DNA methylation in repetitive elements.
Nucleic Acids Res. 2017;45(15):8697-711.
PubMed PMID: 28911103; PMCID: PMC5587781.
http://dx.doi.org/10.1093/nar/gkx587.
}
\author{
Yinan Zheng \email{y-zheng@northwestern.edu},
Lei Liu \email{lei.liu@northwestern.edu},
Wei Zhang \email{wei.zhang1@northwestern.edu},
Warren Kibbe \email{warren.kibbe@nih.gov},
Lifang Hou \email{l-hou@northwestern.edu}

Maintainer: Yinan Zheng \email{y-zheng@northwestern.edu}
}
\keyword{package}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/REMParcel-class.R, R/generics.R,
%   R/REMParcel-methods.R
\name{REMParcel-class}
\alias{REMParcel-class}
\alias{REMParcel}
\alias{getParcelInfo}
\alias{getRefGene}
\alias{getRE}
\alias{getRECpG}
\alias{getILMN}
\alias{saveParcel}
\alias{saveParcel,REMParcel-method}
\alias{getParcelInfo,REMParcel-method}
\alias{getRefGene,REMParcel-method}
\alias{getRE,REMParcel-method}
\alias{getRECpG,REMParcel-method}
\alias{getILMN,REMParcel-method}
\title{REMParcel instances}
\usage{
REMParcel(
  REtype = "Unknown",
  genome = "Unknown",
  platform = "Unknown",
  RefGene = GRanges(),
  RE = GRanges(),
  RECpG = GRanges(),
  ILMN = GRanges()
)

getParcelInfo(object)

getRefGene(object)

getRE(object)

getRECpG(object)

getILMN(object, ...)

saveParcel(object, ...)

\S4method{saveParcel}{REMParcel}(object, work.dir = tempdir(), verbose = FALSE, ...)

\S4method{getParcelInfo}{REMParcel}(object)

\S4method{getRefGene}{REMParcel}(object)

\S4method{getRE}{REMParcel}(object)

\S4method{getRECpG}{REMParcel}(object)

\S4method{getILMN}{REMParcel}(object, REonly = FALSE)
}
\arguments{
\item{REtype}{Type of RE (\code{"Alu"}, \code{"L1"}, or \code{"ERV"}).}

\item{genome}{Specify the build of human genome. Can be either \code{"hg19"} or \code{"hg38"}.}

\item{platform}{Illumina methylation profiling platform (\code{"450k"} or \code{"EPIC"}).}

\item{RefGene}{refSeq gene annotation data, which can be obtained by \code{\link{fetchRefSeqGene}}.}

\item{RE}{Annotated RE genomic range data, which can be obtained by \code{\link{fetchRMSK}} and annotated
by \code{\link{GRannot}}.}

\item{RECpG}{Genomic range data of annotated CpG site identified in RE DNA sequence, which can
be obtained by \code{\link{findRECpG}} and annotated by \code{\link{GRannot}}.}

\item{ILMN}{Illumina CpG probe genomic range data.}

\item{object}{A \code{REMParcel} object.}

\item{...}{For \code{saveParcel}: other parameters to be passed to the \code{saveRDS} method. See
\code{\link{saveRDS}}.}

\item{work.dir}{For \code{saveParcel}: path to the directory where the generated data will be saved.
If not specified, temporary directory \code{tempdir()} will be used.}

\item{verbose}{For \code{saveParcel}: logical parameter. Should the function be verbose?}

\item{REonly}{For \code{getILMN}: see Accessors.}
}
\value{
An object of class \code{REMParcel} for the constructor.
}
\description{
\code{REMParcel} is a container class to organize required datasets for
RE methylation prediction generated from \code{\link{initREMP}} and used in \code{\link{remp}}.
}
\section{Accessors}{

\describe{
    \item{\code{getParcelInfo(object)}}{Return data type, RE type, and flanking window size information
    of the parcel.}
    \item{\code{getRefGene(object)}}{Return RefSeq gene annotation data.}
    \item{\code{getRE(object)}}{Return RE genomic location data for prediction
    (annotated by refSeq gene database).}
    \item{\code{getRECpG(object)}}{Return RE-CpG genomic location
    data for prediction.}
    \item{\code{getILMN(object, REonly = FALSE)}}{Return Illumina CpG probe genomic
    location data for prediction (annotated by refSeq gene database). If
    \code{REonly = TRUE}, only probes within RE region are returned.}
    }
}

\section{Utilities}{

\describe{
    \item{\code{saveParcel(object, work.dir = tempdir(), verbose = FALSE, ...)}}{Save
    the object to local machine. }
    }
}

\examples{
showClass("REMParcel")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grooMethy.R
\name{grooMethy}
\alias{grooMethy}
\title{Groom methylation data to fix potential data issues}
\usage{
grooMethy(
  methyDat,
  Seq.GR = NULL,
  impute = TRUE,
  imputebyrow = TRUE,
  mapGenome = FALSE,
  verbose = FALSE
)
}
\arguments{
\item{methyDat}{A \code{\link{RatioSet}}, \code{\link{GenomicRatioSet}}, \code{\link{DataFrame}},
\code{data.table}, \code{data.frame}, or \code{matrix} of Illumina BeadChip methylation data
(450k or EPIC array) or Illumina methylation percentage estimates by sequencing. If the data are prepared
as a \code{data.frame} or alike format, for Illumina array data, please make sure there is a column or row names 
are available to indicate the Illumina probe names (i.e. cg00000029); for sequencing methylation data, please provide 
the corresponding CpG location information in \code{Seq.GR}.}

\item{Seq.GR}{A \code{\link{GRanges}} object containing genomic locations of the CpGs profiled by sequencing
platforms. This parameter should not be \code{NULL} if the input methylation data \code{methyDat} are
obtained by sequencing platform. The order of \code{Seq.GR} should match the order of \code{methyDat}.
Note that the genomic location can be in either hg19 or hg38 build. See details.}

\item{impute}{If \code{TRUE}, K-Nearest Neighbouring imputation will be applied to fill
the missing values. Default = \code{TRUE}. See Details.}

\item{imputebyrow}{If \code{TRUE}, missing values will be imputed using similar values in row
(i.e., across samples); if \code{FALSE}, missing values will be imputed using similar values
in column (i.e., across CpGs). Default is \code{TRUE}.}

\item{mapGenome}{Logical parameter. If \code{TRUE}, function will return a \code{\link{GenomicRatioSet}}
object instead of a \code{\link{RatioSet}}. This function is not applicable for sequencing data.}

\item{verbose}{Logical parameter. Should the function be verbose?}
}
\value{
A \code{\link{RatioSet}} or \code{\link{GenomicRatioSet}} containing beta value and
M value of the methylation data.
}
\description{
\code{grooMethy} is used to automatically detect and fix data issues including zero beta
value, missing value, and infinite value.
}
\details{
For methylation data in beta value, if zero/one value exists, the logit transformation
from beta to M value will produce infinite value. Therefore, zero/one beta value
will be replaced with the smallest non-zero beta/largest non-one beta value found in the dataset.
\code{grooMethy} can also handle missing value (i.e. \code{NA} or \code{NaN}) using KNN imputation (see
\code{\link{impute.knn}}). The infinite value will be also treated as missing value for imputation.
If the original dataset is in beta value, \code{grooMethy} will first transform it to M value
before imputation is carried out. If the imputed value is out of the original range (which is possible when
\code{imputebyrow = FALSE}), mean value will be used instead. Warning: imputed
values for multimodal distributed CpGs (across samples) may not be correct. Please check package \code{ENmix} to
identify the CpGs with multimodal distribution. Please note that \code{grooMethy} is
also embedded in \code{\link{remp}} so the user can run \code{\link{remp}} directly without
explicitly running \code{grooMethy}. For sequencing methylation data, please specify the genomic location of CpGs
in a \code{GenomicRanges} object and specify it in \code{Seq.GR}. For an example of \code{Seq.GR}, Please
run \code{minfi::getLocations(IlluminaHumanMethylation450kanno.ilmn12.hg19)} (the row names of the CpGs in \code{Seq.GR}
can be \code{NULL}). The user should make sure the genome build of \code{Seq.GR} match the build specified 
in \code{genome} parameter of function \code{\link{initREMP}} and \code{\link{remprofile}} (default is \code{"hg19"}).
}
\examples{
# Get GM12878 methylation data (450k array)
if (!exists("GM12878_450k")) GM12878_450k <- getGM12878("450k")
GM12878_450k <- grooMethy(GM12878_450k, verbose = TRUE)

# Also works if data input is a matrix
grooMethy(minfi::getBeta(GM12878_450k), verbose = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/initREMP.R
\name{initREMP}
\alias{initREMP}
\title{RE Annotation Database Initialization}
\usage{
initREMP(
  arrayType = c("450k", "EPIC", "Sequencing"),
  REtype = c("Alu", "L1", "ERV"),
  annotation.source = c("AH", "UCSC"),
  genome = c("hg19", "hg38"),
  RE = NULL,
  Seq.GR = NULL,
  ncore = NULL,
  BPPARAM = NULL,
  export = FALSE,
  work.dir = tempdir(),
  verbose = FALSE
)
}
\arguments{
\item{arrayType}{Illumina methylation array type. Currently \code{"450k"}, \code{"EPIC"},
and \code{"Sequencing"} are supported. Default = \code{"450k"}.}

\item{REtype}{Type of RE. Currently \code{"Alu"}, \code{"L1"}, and \code{"ERV"} are supported.}

\item{annotation.source}{Character parameter. Specify the source of annotation databases, including
the RefSeq Gene annotation database and RepeatMasker annotation database. If \code{"AH"}, the database 
will be obtained from the AnnotationHub package. If \code{"UCSC"}, the database will be downloaded 
from the UCSC website http://hgdownload.cse.ucsc.edu/goldenpath. The corresponding build (\code{"hg19"} or 
\code{"hg38"}) can be specified in the parameter \code{genome}.}

\item{genome}{Character parameter. Specify the build of human genome. Can be either \code{"hg19"} or 
\code{"hg38"}. Note that if \code{annotation.source == "AH"}, only hg19 database is available.}

\item{RE}{A \code{\link{GRanges}} object containing user-specified RE genomic location information.
If \code{NULL}, the function will retrive RepeatMasker RE database from \code{\link{AnnotationHub}}
(build hg19) or download the database from UCSC website (build hg19/hg38).}

\item{Seq.GR}{A \code{\link{GRanges}} object containing genomic locations of the CpGs profiled by sequencing
platforms. This parameter should not be \code{NULL} if \code{arrayType == 'Sequencing'}. Note that the genomic
location can be in either hg19 or hg38 build. See details.}

\item{ncore}{Number of cores used for parallel computing. By default max number of cores
available in the machine will be utilized. If \code{ncore = 1}, no parallel computing is allowed.}

\item{BPPARAM}{An optional \code{\link{BiocParallelParam}} instance determining the parallel back-end to
be used during evaluation. If not specified, default back-end in the machine will be used.}

\item{export}{Logical. Should the returned \code{\link{REMParcel}} object be saved to local machine?
See Details.}

\item{work.dir}{Path to the directory where the generated data will be saved. Valid when
\code{export = TRUE}. If not specified and \code{export = TRUE}, temporary directory \code{tempdir()}
will be used.}

\item{verbose}{Logical parameter. Should the function be verbose?}
}
\value{
An \code{\link{REMParcel}} object containing data needed for RE methylation prediction.
}
\description{
\code{initREMP} is used to initialize annotation database for RE methylation prediction.
Three RE types in human, Alu element (Alu), LINE-1 (L1), and endogenous retrovirus (ERV) are available.
}
\details{
Currently, we support two major types of RE in the human genome, Alu and L1. The main purpose of
\code{initREMP} is to generate and annotate CpG/RE data using the refSeq Gene (hg19)
annotation database (provided by \code{\link{AnnotationHub}}). These annotation data are crucial to
RE methylation prediction in \code{\link{remp}}. Once generated, the data can be reused in the future
(data can be very large). Therefore, we recommend the user to save the output from
\code{initREMP} to the local machine, so that user only need to run this function once
as long as there is no change to the RE database. To minimize the size of the resulting data file, the generated
annotation data are only for REs that contain RE-CpGs with neighboring profiled CpGs. By default, the
neighboring CpGs are confined within 1200 bp flanking window. This window size can be modified using
\code{\link{remp_options}}. Note that the refSeq Gene database from UCSC is dynamic (updated periodically) 
and reflecting the latest knowledge of gene, whereas the database from AnnotationHub is static and classic. 
Using different sources will have a slight impact on the prediction results of RE methylation and gene annotation 
of final results. For sequencing methylation data, please specify the genomic location of CpGs
in a \code{GenomicRanges} object and specify it in \code{Seq.GR}. For an example of \code{Seq.GR}, Please
run \code{minfi::getLocations(IlluminaHumanMethylation450kanno.ilmn12.hg19)} (the row names of the CpGs in
\code{Seq.GR} can be \code{NULL}). The user should make sure the genome build of \code{Seq.GR} match the 
build specified in \code{genome} parameter (default is \code{"hg19"}).
}
\examples{
if (!exists("remparcel")) {
  data(Alu.hg19.demo)
  remparcel <- initREMP(arrayType = "450k", 
                        REtype = "Alu", 
                        annotation.source = "AH",
                        genome = "hg19",
                        RE = Alu.hg19.demo, 
                        ncore = 1,
                        verbose = TRUE)
}

}
\seealso{
See \code{\link{remp}} for RE methylation prediction.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/REMPtools.R
\name{fetchRMSK}
\alias{fetchRMSK}
\title{Get RE database from RepeatMasker}
\usage{
fetchRMSK(
  REtype = c("Alu", "L1", "ERV"),
  annotation.source = c("AH", "UCSC"),
  genome = c("hg19", "hg38"),
  verbose = FALSE
)
}
\arguments{
\item{REtype}{Type of RE. Currently \code{"Alu"}, \code{"L1"}, and \code{"ERV"} are supported.}

\item{annotation.source}{Character parameter. Specify the source of annotation databases, including
the RefSeq Gene annotation database and RepeatMasker annotation database. If \code{"AH"}, the database 
will be obtained from the AnnotationHub package. If \code{"UCSC"}, the database will be downloaded 
from the UCSC website http://hgdownload.cse.ucsc.edu/goldenpath. The corresponding build (\code{"hg19"} or 
\code{"hg38"}) will be specified in the parameter \code{genome}.}

\item{genome}{Character parameter. Specify the build of human genome. Can be either \code{"hg19"} or
\code{"hg38"}.}

\item{verbose}{Logical parameter. Should the function be verbose?}
}
\value{
A \code{\link{GRanges}} object containing RE database. 'repName' column
indicates the RE name; 'swScore' column indicates the SW score; 'Index' is an
internal index for RE to facilitate data referral, which is meaningless for external use.
}
\description{
\code{fetchRMSK} is used to obtain specified RE database from RepeatMasker Database
provided by AnnotationHub.
}
\examples{
L1 <- fetchRMSK(REtype = "L1", 
                annotation.source = "AH",
                genome = "hg19", 
                verbose = TRUE)
L1
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rempTemplate.R
\name{rempTemplate}
\alias{rempTemplate}
\title{Prepare data template for REMP}
\usage{
rempTemplate(
  methyDat = NULL,
  Seq.GR = NULL,
  parcel = NULL,
  win = 1000,
  verbose = FALSE
)
}
\arguments{
\item{methyDat}{A \code{\link{RatioSet}}, \code{\link{GenomicRatioSet}}, \code{\link{DataFrame}},
\code{data.table}, \code{data.frame}, or \code{matrix} of Illumina BeadChip methylation data
(450k or EPIC array) or Illumina methylation percentage estimates by sequencing.}

\item{Seq.GR}{A \code{\link{GRanges}} object containing genomic locations of the CpGs profiled by sequencing
platforms. This parameter should not be \code{NULL} if the input methylation data \code{methyDat} are
obtained by sequencing. Note that the genomic location can be in either hg19 or hg38 build, but must be consistent
with the build as \code{parcel}. See details in \code{\link{initREMP}}.}

\item{parcel}{An \code{\link{REMParcel}} object containing necessary data to carry out the
prediction. If \code{NULL}, the function will search the \code{.rds} data file in \code{work.dir}
exported by \code{\link{initREMP}} (with \code{export = TRUE}) or \code{\link{saveParcel}}.}

\item{win}{An integer specifying window size to confine the upstream and downstream flanking
region centered on the predicted CpG in RE for prediction. Default = \code{1000}.}

\item{verbose}{Logical parameter. Should the function be verbose?}
}
\value{
A \code{template} object containing a \code{\link{GRanges}} object of neifhboring CpGs of RE-CpGs
to be predicted (\code{$NBCpG_GR}) and their methylation dataset matrix (\code{$NBCpG_methyDat});
a \code{\link{GRanges}} object of RE-CpGs for model training (\code{$RECpG_GR}) and their methylation
dataset matrix (\code{$RECpG_methyDat}); \code{\link{GRanges}} objects of RefSeq Gene database (\code{$refgene})
and RE (\code{$RE}) that are extracted from the \code{parcel} input; a string of RE type (\code{$REtype})
and a string of methylation platform (\code{$arrayType}). Note: the subset operator \code{[]} is supported.
}
\description{
\code{rempTemplate} is used to build a set of data templates for prediction. The data templates include
RE-CpGs and their methylation data for model training, neighboring CpGs of RE-CpGs and their
methylation data for model prediction, and other necessary information about the prediction.
This function is useful when one needs to experiment different tunning parameters so that these pre-built
data templates can be re-used and substaintially improve efficiency.
}
\examples{
if (!exists("GM12878_450k"))
  GM12878_450k <- getGM12878("450k")
if (!exists("remparcel")) {
  data(Alu.hg19.demo)
  remparcel <- initREMP(arrayType = "450k",
                        REtype = "Alu",
                        annotation.source = "AH",
                        genome = "hg19",
                        RE = Alu.hg19.demo,
                        ncore = 1,
                        verbose = TRUE)
}

template <- rempTemplate(GM12878_450k, 
                         parcel = remparcel, 
                         win = 1000, 
                         verbose = TRUE)
template

## To make a subset
template[1]

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/remp.R
\name{remp}
\alias{remp}
\title{Repetitive element methylation prediction}
\usage{
remp(
  methyDat = NULL,
  REtype = c("Alu", "L1", "ERV"),
  Seq.GR = NULL,
  parcel = NULL,
  work.dir = tempdir(),
  win = 1000,
  method = c("rf", "xgbTree", "svmLinear", "svmRadial", "naive"),
  autoTune = TRUE,
  param = NULL,
  seed = NULL,
  ncore = NULL,
  BPPARAM = NULL,
  verbose = FALSE
)
}
\arguments{
\item{methyDat}{A \code{\link{RatioSet}}, \code{\link{GenomicRatioSet}}, \code{\link{DataFrame}},
\code{data.table}, \code{data.frame}, or \code{matrix} of Illumina BeadChip methylation data
(450k or EPIC array) or Illumina methylation percentage estimates by sequencing. See Details. 
Alternatively, user can also specify a pre-built data template (see \code{\link{rempTemplate}}).
\code{remp} to carry out the prediction. See \code{\link{rempTemplate}}. With template specified, \code{methyDat},
\code{REtype}, \code{parcel}, and \code{work.dir} can be skipped.}

\item{REtype}{Type of RE. Currently \code{"Alu"}, \code{"L1"}, and \code{"ERV"} are supported. If \code{NULL}, 
the type of RE will be extracted from \code{parcel}.}

\item{Seq.GR}{A \code{\link{GRanges}} object containing genomic locations of the CpGs profiled by sequencing
platforms. This parameter should not be \code{NULL} if the input methylation data \code{methyDat} are
obtained by sequencing. Note that the genomic location can be in either hg19 or hg38 build. 
See details in \code{\link{initREMP}}.}

\item{parcel}{An \code{\link{REMParcel}} object containing necessary data to carry out the
prediction. If \code{NULL}, \code{REtype} must specify a type of RE so that the function can search the
\code{.rds} data file in \code{work.dir} exported by \code{\link{initREMP}} (with \code{export = TRUE})
or \code{\link{saveParcel}}.}

\item{work.dir}{Path to the directory where the annotation data generated by \code{\link{initREMP}}
are saved. Valid when the argument \code{parcel} is missing. If not specified, temporary directory
\code{tempdir()} will be used. If specified, the directory path has to be the same as the
one specified in \code{\link{initREMP}} or in \code{\link{saveParcel}}.}

\item{win}{An integer specifying window size to confine the upstream and downstream flanking
region centered on the predicted CpG in RE for prediction. Default = \code{1000}. See Details.}

\item{method}{Name of model/approach for prediction. Currently \code{"rf"} (Random Forest),
\code{"xgbTree"} (Extreme Gradient Boosting), \code{"svmLinear"} (SVM with linear kernel), \code{"svmRadial"}
(SVM with radial kernel), and \code{"naive"} (carrying over methylation values of the closest
CpG site) are available. Default = \code{"rf"} (Random Forest). See Details.}

\item{autoTune}{Logical parameter. If \code{TRUE}, a 3-time repeated 5-fold cross validation
will be performed to determine the best model parameter. If \code{FALSE}, the \code{param} option
(see below) must be specified. Default = \code{TRUE}. Auto-tune will be disabled using Random Forest.
See Details.}

\item{param}{A list specifying fixed model tuning parameter(s) (not applicable for Random Forest, see Details).
For Extreme Gradient Boosting, \code{param} list must contain '\code{$nrounds}', '\code{$max_depth}',
'\code{$eta}', '\code{$gamma}', '\code{$colsample_bytree}', '\code{$min_child_weight}', and '\code{$subsample}'.
See \code{xgbTree} in package \code{caret}. For SVM, \code{param} list must contain '\code{$C}' (cost) for linear kernel
or '\code{$sigma}' and '\code{$C}' for radial basis function kernel. This parameter is valid only
when \code{autoTune = FALSE}.}

\item{seed}{Random seed for Random Forest model for reproducible prediction results.
Default is \code{NULL}, which generates a seed.}

\item{ncore}{Number of cores used for parallel computing. By default, max number of cores available
in the machine will be utilized. If \code{ncore = 1}, no parallel computing is allowed.}

\item{BPPARAM}{An optional \code{\link{BiocParallelParam}} instance determining the parallel back-end to
be used during evaluation. If not specified, default back-end in the machine will be used.}

\item{verbose}{Logical parameter. Should the function be verbose?}
}
\value{
A \code{\link{REMProduct}} object containing predicted RE methylation results.
}
\description{
\code{remp} is used to predict genomewide methylation levels of locus-specific repetitive elements (RE).
Two major RE types in human, Alu element (Alu) and LINE-1 (L1) are available.
}
\details{
Before running \code{remp}, user should make sure the methylation data have gone through
proper quality control, background correction, and normalization procedures. Both beta value
and M value are allowed. Rows represents probes and columns represents samples. For array data,
please make sure to have row names that specify the Illumina probe ID (i.e. cg00000029). For sequencing
data, please provide the genomic location of CpGs in a \code{\link{GRanges}} obejct and
specify it using \code{Seq.GR} parameter. \code{win = 1000} is based on previous findings showing that
neighboring CpGs are more likely to be co-modified within 1000 bp. User can specify narrower window size
for slight improvement of prediction accuracy at the cost of less predicted RE. Window size greater than 1000 is not
recommended as the machine learning models would not be able to learn much userful information
for prediction but introduce noise. Random Forest model (\code{method = "rf"}) is recommented
as it offers more accurate prediction and it also enables prediction reliability functionality.
Prediction reliability is estimated by conditional standard deviation using Quantile Regression Forest.
Please note that if parallel computing is allowed, parallel Random Forest
(powered by package \code{\link{ranger}}) will be used automatically. The performance of
Random Forest model is often relatively insensitive to the choice of \code{mtry}.
Therefore, auto-tune will be turned off using Random Forest and \code{mtry} will be set to one third
of the total number of predictors. For SVM, if \code{autoTune = TRUE}, preset tuning parameter
search grid can be access and modified using \code{\link{remp_options}}.
}
\examples{
# Obtain example Illumina example data (450k)
if (!exists("GM12878_450k")) 
  GM12878_450k <- getGM12878("450k")

# Make sure you have run 'initREMP' first. See ?initREMP.
if (!exists("remparcel")) {
  data(Alu.hg19.demo)
  remparcel <- initREMP(arrayType = "450k",
                        REtype = "Alu",
                        annotation.source = "AH",
                        genome = "hg19",
                        RE = Alu.hg19.demo,
                        ncore = 1,
                        verbose = TRUE)
}

# With data template pre-built. See ?rempTemplate.
if (!exists("template")) 
  template <- rempTemplate(GM12878_450k, 
                           parcel = remparcel, 
                           win = 1000, 
                           verbose = TRUE)

# Run remp with pre-built template:
remp.res <- remp(template, ncore = 1)

# Or run remp without pre-built template (identical results):
\dontrun{
  remp.res <- remp(GM12878_450k, 
                   REtype = "Alu", 
                   parcel = remparcel, 
                   ncore = 1,
                   verbose = TRUE)
}

remp.res
details(remp.res)
rempB(remp.res) # Methylation data (beta value)

# Extract CpG location information. 
# This accessor is inherit from class 'RangedSummarizedExperiment')
rowRanges(remp.res)

# RE annotation information
rempAnnot(remp.res)

# Add gene annotation
remp.res <- decodeAnnot(remp.res, type = "symbol")
rempAnnot(remp.res)

# (Recommended) Trim off less reliable prediction
remp.res <- rempTrim(remp.res)

# Obtain RE-level methylation (aggregate by mean)
remp.res <- rempAggregate(remp.res)
rempB(remp.res) # Methylation data (beta value)

# Extract RE location information
rowRanges(remp.res)

# Density plot across predicted RE
remplot(remp.res)

}
\seealso{
See \code{\link{initREMP}} to prepare necessary annotation database before running \code{remp}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DemoData.R
\docType{data}
\name{Alu.hg19.demo}
\alias{Alu.hg19.demo}
\title{Subset of Alu genomic location dataset (hg19)}
\format{
A \code{\link{GRanges}} object.
}
\source{
RepeatMasker database provided by package \code{AnnotationHub}
}
\usage{
Alu.hg19.demo
}
\value{
A GRanges object with 500 ranges and 3 metadata columns.
}
\description{
A \code{\link{GRanges}} dataset containing 500 Alu sequences that have CpGs profiled
by both Illumina 450k and EPIC array. The variables are as follows:
}
\details{
\itemize{
  \item seqnames: chromosome number
  \item ranges: hg19 genomic position
  \item strand: DNA strand
  \item swScore: Smith Waterman (SW) alignment score
  \item repName: Alu name
  \item repClass: Alu class
  \item repFamily: Alu family
  \item Index: internal index (meaningless for external use; 
  not communicable between genome builds)
}

\code{Alu.hg19.demo} has the same format as the data object 
returned by \code{\link{fetchRMSK}}.
}
\seealso{
See \code{\link{fetchRMSK}} to obtain the complete Alu/L1 dataset.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/REMProduct-class.R, R/generics.R,
%   R/REMProduct-methods.R
\name{REMProduct-class}
\alias{REMProduct-class}
\alias{REMProduct}
\alias{rempM}
\alias{rempB}
\alias{rempQC}
\alias{rempAnnot}
\alias{rempImp}
\alias{rempStats}
\alias{remplot}
\alias{details}
\alias{decodeAnnot}
\alias{rempTrim}
\alias{rempAggregate}
\alias{rempCombine}
\alias{rempM,REMProduct-method}
\alias{rempB,REMProduct-method}
\alias{rempQC,REMProduct-method}
\alias{rempImp,REMProduct-method}
\alias{rempAnnot,REMProduct-method}
\alias{rempStats,REMProduct-method}
\alias{remplot,REMProduct-method}
\alias{details,REMProduct-method}
\alias{decodeAnnot,REMProduct-method}
\alias{rempTrim,REMProduct-method}
\alias{rempAggregate,REMProduct-method}
\alias{rempCombine,REMProduct,REMProduct-method}
\title{REMProduct instances}
\usage{
REMProduct(
  REtype = "Unknown",
  genome = "Unknown",
  platform = "Unknown",
  win = "Unknown",
  predictModel = "Unknown",
  QCModel = "Unknown",
  rempM = NULL,
  rempB = NULL,
  rempQC = NULL,
  cpgRanges = GRanges(),
  sampleInfo = DataFrame(),
  REannotation = GRanges(),
  RECpG = GRanges(),
  regionCode = DataFrame(),
  refGene = GRanges(),
  varImp = DataFrame(),
  REStats = DataFrame(),
  GeneStats = DataFrame(),
  Seed = NULL
)

rempM(object)

rempB(object)

rempQC(object)

rempAnnot(object)

rempImp(object)

rempStats(object)

remplot(object, ...)

details(object)

decodeAnnot(object, ...)

rempTrim(object, ...)

rempAggregate(object, ...)

rempCombine(object1, object2)

\S4method{rempM}{REMProduct}(object)

\S4method{rempB}{REMProduct}(object)

\S4method{rempQC}{REMProduct}(object)

\S4method{rempImp}{REMProduct}(object)

\S4method{rempAnnot}{REMProduct}(object)

\S4method{rempStats}{REMProduct}(object)

\S4method{remplot}{REMProduct}(object, type = c("individual", "overall"), ...)

\S4method{details}{REMProduct}(object)

\S4method{decodeAnnot}{REMProduct}(object, type = c("symbol", "entrez"), ncore = 1, BPPARAM = NULL)

\S4method{rempTrim}{REMProduct}(object, threshold = 1.7, missingRate = 0.2)

\S4method{rempAggregate}{REMProduct}(object, NCpG = 2, ncore = 1, BPPARAM = NULL)

\S4method{rempCombine}{REMProduct,REMProduct}(object1, object2)
}
\arguments{
\item{REtype}{Type of RE (\code{"Alu"}, \code{"L1"}, or \code{"ERV"}).}

\item{genome}{Specify the build of human genome. Can be either \code{"hg19"} or \code{"hg38"}.}

\item{platform}{Illumina methylation profiling platform (\code{"450k"} or \code{"EPIC"}).}

\item{win}{Flanking window size of the predicting RE-CpG.}

\item{predictModel}{Name of the model used for prediction.}

\item{QCModel}{Name of the model used for prediction quality evaluation.}

\item{rempM}{Predicted methylation level in M value.}

\item{rempB}{Predicted methylation level in beta value (optional).}

\item{rempQC}{Prediction quality scores, which is available only when Random Forest
model is used in \code{\link{remp}}.}

\item{cpgRanges}{Genomic ranges of the predicting RE-CpG.}

\item{sampleInfo}{Sample information.}

\item{REannotation}{Annotation data for the predicting RE.}

\item{RECpG}{Annotation data for the RE-CpG profiled by Illumina platform.}

\item{regionCode}{Internal index code defined in \code{refGene} for gene region indicators.}

\item{refGene}{refSeq gene annotation data, which can be obtained by \code{\link{fetchRefSeqGene}}.}

\item{varImp}{Importance of the predictors.}

\item{REStats}{RE coverage statistics, which is internally generated in \code{\link{remp}}.}

\item{GeneStats}{Gene coverage statistics, which is internally generated in \code{\link{remp}}.}

\item{Seed}{Random seed for Random Forest model for reproducible prediction results.}

\item{object}{A \code{REMProduct} object.}

\item{...}{For \code{plot}: \code{\link{graphical parameters}} to be passed to the \code{plot} method.}

\item{object1}{A \code{REMProduct} object.}

\item{object2}{A \code{REMProduct} object.}

\item{type}{For \code{plot} and \code{decodeAnnot}: see Utilities.}

\item{ncore}{For \code{decodeAnnot} and \code{rempAggregate}: number of cores used for parallel computing.
By default no parallel computing is allowed (\code{ncore = 1}).}

\item{BPPARAM}{For \code{decodeAnnot} and \code{rempAggregate}: an optional \code{\link{BiocParallelParam}}
instance determining the parallel back-end to be used during evaluation. If not specified, default
back-end in the machine will be used.}

\item{threshold}{For \code{rempTrim}: see Utilities.}

\item{missingRate}{For \code{rempTrim}: see Utilities.}

\item{NCpG}{For \code{rempAggregate}: see Utilities.}
}
\value{
An object of class \code{REMProduct} for the constructor.
}
\description{
Class \code{REMProduct} is to maintain RE methylation prediction results.
\code{REMProduct} inherits Bioconductor's \code{RangedSummarizedExperiment} class.
}
\section{Accessors}{

\describe{
    \item{\code{rempM(object)}}{Return M value of the prediction.}
    \item{\code{rempB(object)}}{Return beta value of the prediction.}
    \item{\code{rempQC(object)}}{Return prediction quality scores.}
    \item{\code{rempImp(object)}}{Return relative importance of predictors.}
    \item{\code{rempStats(object)}}{Return RE and gene coverage statistics.}
    \item{\code{rempAnnot(object)}}{Return annotation data for the predicted RE.}
    }
}

\section{Utilities}{

\describe{
    \item{\code{remplot(object, type = c("individual", "overall"), ...)}}{Make a density plot of predicted methylation
    (beta values) in the \code{REMProduct} object. If \code{type = "individual"}, density curves will be
    plotted for each of the samples; If \code{type = "overall"}, one density curve of the mean methylation level
    across the samples will be plotted. Default \code{type = "individual"}.}
    \item{\code{details(object)}}{Display detailed descriptive statistics of the predicion results.}
    \item{\code{decodeAnnot(object, type = c("symbol", "entrez")), ncore = NULL, BPPARAM = NULL}}{Decode the
    RE annotation data by Gene Symbol (when \code{type = "Symbol"}) or Entrez Gene
    (when \code{type = "Entrez"}).Default \code{type = "Symbol"}. Annotation data are provided by
    \code{\link{org.Hs.eg.db}}.}
    \item{\code{rempTrim(object, threshold = 1.7, missingRate = 0.2)}}{Any predicted CpG values with
    quality score < threshold (default = 1.7, specified by \code{threshold = 1.7}) will be replaced with NA.
    CpGs contain more than missingRate * 100% (default = 20%, specified by \code{missingRage = 0.2}) missing
    rate across samples will be discarded. Relavant summary statistics will be re-evaluated.}
    \item{\code{rempAggregate(object, NCpG = 2, ncore = NULL, BPPARAM = NULL)}}{Aggregate the predicted RE-CpG
    methylation by RE using mean. To ensure the reliability of the aggregation, by default only RE with at
    least 2 predicted CpG sites (specified by \code{NCpG = 2}) will be aggregated.}
    \item{\code{rempCombine(object1, object2)}}{Combine two \code{REMProduct} objects by column.}
    }
}

\examples{
showClass("REMProduct")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/options.R
\name{remp_options}
\alias{remp_options}
\alias{remp_reset}
\title{Set or get options for REMP package}
\usage{
remp_options(...)

remp_reset()
}
\arguments{
\item{...}{Option names to retrieve option values or \code{[key]=[value]} pairs to set options.}
}
\value{
\code{NULL}
}
\description{
Tools to manage global setting options for REMP package.
}
\section{Supported options}{

The following options are supported
\describe{
 \item{\code{.default.AluFamily.grep}}{Regular expression for 'grep' to extract Alu family to be included in the prediction.}
 \item{\code{.default.L1Family.grep}}{Regular expression for 'grep' to extract L1 family to be included in the prediction.}
 \item{\code{.default.ERVFamily.grep}}{Regular expression for 'grep' to extract ERV family to be included in the prediction.}
 \item{\code{.default.chr}}{List of human chromosome.}
 \item{\code{.default.GM12878.450k.URL}}{URL to download GM12878 450k methylation profiling data.}
 \item{\code{.default.RMSK.hg19.URL}}{URL to download RepeatMasker database in hg19 genome.}
 \item{\code{.default.RMSK.hg38.URL}}{URL to download RepeatMasker database in hg38 genome.}
 \item{\code{.default.refGene.hg19.URL}}{URL to download refSeq gene database in hg19 genome.}
 \item{\code{.default.refGene.hg38.URL}}{URL to download refSeq gene database in hg38 genome.}
 \item{\code{.default.AH.repeatmasker.hg19}}{\code{AnnotationHub} data ID linked to RepeatMasker
 annotation database (Mar 2020, build hg19).}
 \item{\code{.default.AH.repeatmasker.hg38}}{\code{AnnotationHub} data ID linked to RepeatMasker
 annotation database (Sep 2021, build hg38).}
 \item{\code{.default.AH.refgene.hg19}}{\code{AnnotationHub} data ID linked to refSeq gene database
 (build hg19)}
 \item{\code{.default.AH.hg38ToHg19.over.chain}}{\code{AnnotationHub} hg38 to hg19 liftover chain data ID.}
 \item{\code{.default.AH.hg19ToHg38.over.chain}}{\code{AnnotationHub} hg19 to hg38 liftover chain data ID.}
 \item{\code{.default.TSS.upstream}}{Define the upstream range of transcription start site region.}
 \item{\code{.default.TSS.downstream}}{Define the downstream range of transcription start site region.}
 \item{\code{.default.max.flankWindow}}{Define the max size of the flanking window surrounding the
 predicted RE-CpG.}
 \item{\code{.default.27k.total.probes}}{Total number of probes designed in Illumina 27k array.}
 \item{\code{.default.450k.total.probes}}{Total number of probes designed in Illumina 450k array.}
 \item{\code{.default.epic.total.probes}}{Total number of probes designed in Illumina EPIC array.}
 \item{\code{.default.450k.annotation}}{A character string associated with the Illumina 450k
 array annotation dataset.}
 \item{\code{.default.epic.annotation}}{A character string associated with the Illumina EPIC
 array annotation dataset.}
 \item{\code{.default.genomicRegionColNames}}{Define the names of the genomic regions for prediction.}
 \item{\code{.default.predictors}}{Define the names of predictors for RE methylation prediction.}
 \item{\code{.default.svmLinear.tune}}{Define the default \code{C} (Cost) parameter for Support
 Vector Machine (SVM) using linear kernel.}
 \item{\code{.default.svmRadial.tune}}{Define the default parameters (\code{C} and \code{sigma}) for SVM
 using Radial basis function kernel.}
 \item{\code{.default.xgbTree.tune}}{Define the default parameters (\code{nrounds}, \code{eta}, \code{max_depth},
 \code{gamma}, \code{colsample_bytree}, \code{min_child_weight}, and \code{subsample}) for Extreme Gradient
 Boosting.}
}
}

\examples{
# Display all default settings
remp_options()

# Display a specified setting
remp_options(".default.max.flankWindow")

# Change default maximum flanking window size to 2000
remp_options(.default.max.flankWindow = 2000)

# Reset all options
remp_reset()
}
